/*
 * Code to handle R39XX IRQs plus some generic interrupt stuff.
 *
 * Copyright (C) 1992 Linus Torvalds
 * Copyright (C) 1999 Harald Koerfgen
 * Copyright (C) 2000 Pavel Machek
 *
 * $Id: irq.c,v 1.18 2000/09/05 18:14:46 pavel Exp $
 */
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/kernel_stat.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/timex.h>
#include <linux/malloc.h>
#include <linux/random.h>

#include <asm/bitops.h>
#include <asm/bootinfo.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/mipsregs.h>
#include <asm/system.h>
#include <asm/r39xx.h>

unsigned long spurious_count = 0;

irq_cpustat_t irq_stat [NR_CPUS];

#define I_NOMASK 1

static inline void mask_irq(unsigned int irq_nr)
{
	switch (irq_nr) {
	case 0:  /* Periodic Timer Interrupt */ 
		IntClear5 = INT5_PERIODICINT;
/* FIXME: what is this? It looks like irq 0 lives in two different locations?!?! */
#ifdef CONFIG_VTECH_HELIO_EMULATOR
		/* 
		 * The Helio emulator doesn't support INT6 high
		 * priority interrupts.	 Use PERINT from INT5
		 * instead. 
		 */
		IntEnable5 &= ~INT5_PERIODICINT;
#else
		IntClear6 = INT6_PERIODICINT;
		IntEnable6 &= ~INT6_PERIODICINT;
#endif
		break;
	case 1:  /* IDE Interrupt */
		/*
		 * Hack alert!
		 *
		 * The TC6345 seems to be register compatible with the
		 * ITE IT8368E PCMCIA/GPIO Buffer Chip. The following
		 * four lines simply acknowledge all pending interrupts.
		 * Eventually a it8368 driver has to take care of this.
		 * This hack is neccessary to make the ide driver work in
 		 * this early stage. HK
		 */
		*(volatile unsigned short *)0xb0400018 = *(volatile unsigned short *)0xb0400018;
		*(volatile unsigned short *)0xb040001a = *(volatile unsigned short *)0xb040001a;
		*(volatile unsigned short *)0xb040001c = *(volatile unsigned short *)0xb040001c;
		*(volatile unsigned short *)0xb040001e = *(volatile unsigned short *)0xb040001e;
		IntClear3 = 0x4;
		IntEnable3 &= ~0x4;
		break;
	case 3:  /* Serial port receive interrupt */
		IntClear2 = INT2_UARTARXINT;
		IntEnable2 &= ~INT2_UARTARXINT;
		break;
/* FIXME: clean this up! */
#ifdef CONFIG_VTECH_HELIO
	case 10: /* Helio Buttons Interrupt */
	case 11:
        case 12:
        case 13:
		IntEnable5 &= ~0x00003fff;
		IntEnable3 &= ~MFIO_PIN_IO_KEY8;
/*		IntClear5 = 0x00003fff;
		IntClear3 = MFIO_PIN_IO_KEY8; */
		break;
#endif
#undef INTERRUPT
#define INTERRUPT(num, X, flags, bit) \
	case num: \
		if (!(flags & I_NOMASK)) { \
		  IntEnable##X &= ~bit; \
		  IntClear##X = bit; \
		} \
		break;
#include "irq_table.c"
	default:
		printk( "Attempt to mask unknown IRQ %d?\n", irq_nr );
	}
}

static inline void unmask_irq(unsigned int irq_nr)
{
	switch (irq_nr) {
	case 0:
#ifdef CONFIG_VTECH_HELIO_EMULATOR
		IntEnable5 |= INT5_PERIODICINT;
		break;
#else
		IntEnable6 |= INT6_PERIODICINT;
#endif
		break;
	case 1:  /* IDE Interrupt */
		IntEnable3 |= 0x4;
		break;
	case 3:  /* Serial port receive interrupt */
		IntEnable2 |= INT2_UARTARXINT;
		break;
#ifdef CONFIG_VTECH_HELIO
	case 10: /* Helio Buttons Interrupt */
	case 11:
        case 12:
        case 13:
		IntEnable5 |= 0x00003fff;
		IntEnable3 |= MFIO_PIN_IO_KEY8;
		break;
#endif
#undef INTERRUPT
#define INTERRUPT(num, X, flags, bit) \
	case num: \
		if (!(flags & I_NOMASK)) IntEnable##X |= bit; \
		break;
#include "irq_table.c"

	default:
		printk( "Attempt to unmask unknown IRQ %d?\n", irq_nr );
	}
}

void disable_irq(unsigned int irq_nr)
{
    unsigned long flags;

    save_and_cli(flags);
    mask_irq(irq_nr);
    restore_flags(flags);
}

void enable_irq(unsigned int irq_nr)
{
    unsigned long flags;

    save_and_cli(flags);
    unmask_irq(irq_nr);
    restore_flags(flags);
}

/*
 * Pointers to the low-level handlers: first the general ones, then the
 * fast ones, then the bad ones.
 */
extern void interrupt(void);

/* NR_IRQS = 64 on MIPS */
#if NR_IRQS != 64
#error "NR_IRQS != 64"
#endif

static struct irqaction *irq_action[NR_IRQS] =
{
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};

int get_irq_list(char *buf)
{
    int i, len = 0;
    struct irqaction *action;

    for (i = 0; i < NR_IRQS; i++) {
	action = irq_action[i];
	if (!action)
	    continue;
	len += sprintf(buf + len, "%2d: %8d %c %s",
		       i, kstat.irqs[0][i],
		       (action->flags & SA_INTERRUPT) ? '+' : ' ',
		       action->name);
	for (action = action->next; action; action = action->next) {
	    len += sprintf(buf + len, ",%s %s",
			   (action->flags & SA_INTERRUPT) ? " +" : "",
			   action->name);
	}
	len += sprintf(buf + len, "\n");
    }
    return len;
}

atomic_t __mips_bh_counter;

/*
 * do_IRQ handles IRQ's that have been installed without the
 * SA_INTERRUPT flag: it uses the full signal-handling return
 * and runs with other interrupts enabled. All relatively slow
 * IRQ's should use this format: notably the keyboard/timer
 * routines.
 */
asmlinkage void do_IRQ(int irq, struct pt_regs *regs)
{
    struct irqaction *action;
    int do_random, cpu;

    if (irq == 20) {
#undef INTERRUPT
#define INTERRUPT(num, X, flags, bit) \
	if (IntStatus##X & bit) \
		return do_IRQ(num, regs);
#include "irq_table.c"
    }

    cpu = smp_processor_id();
    irq_enter(cpu);
    kstat.irqs[cpu][irq]++;

    if (irq == 20) {
	    printk("20 %08x %08x\n   %08x %08x\n   %08x\n",
		   IntStatus1, IntStatus2, IntStatus3,
		   IntStatus4, IntStatus5 );
    }

#ifdef CONFIG_VTECH_HELIO
    if ((irq >= 10) && (irq != 40) && (irq != 41)) {
	    printk("%d %01d %01d:%01d %02x:%02x P%01d%01d %01d\n", irq,
		   MFIOInput & MFIO_PIN_HOTSYNC_KEY ? 1 : 0,
		   IntStatus3 & MFIO_PIN_IO_KEY8 ? 1 : 0,
		   IntStatus4 & MFIO_PIN_IO_KEY8 ? 1 : 0,
		   (IntStatus5 & (0x7f << INT5_POSINT_SHIFT)) >> INT5_POSINT_SHIFT,
		   IntStatus5 & 0x7f,
		   IntStatus5 & INT5_POSONBUTINT ? 1 : 0,
		   IntStatus5 & INT5_NEGONBUTINT ? 1 : 0,
		   SIBControl & SIB_SIBIRQ ? 1 : 0);
    }
#endif

    mask_irq(irq);
    action = *(irq + irq_action);

    if ((irq >= 10) && (irq != 40) && (irq != 41)) {
#ifdef CONFIG_VTECH_HELIO
	    if ((irq == 10) && (IntStatus5 & (0x01 << INT5_POSINT_SHIFT))) {
		    if (MFIOOutput & MFIO_PIN_BACKLIGHT)
			    MFIOOutput &= ~MFIO_PIN_BACKLIGHT;
		    else
			    MFIOOutput |= MFIO_PIN_BACKLIGHT;
		    poke_blanked_console();
	    }
	    if (irq == 16) {
		    if (MFIOOutput & MFIO_PIN_LCD_POWER) {
			    pmtemp_max_doze();

			    /* calibrate_delay(); */
		    } else {
			    pmtemp_resume();
			    /* calibrate_delay(); */
		    }
		    poke_blanked_console();
	    }
    
#endif

       	    IntClear3 = 0xffffffff;
	    IntClear4 = 0xffffffff;
	    IntClear5 = 0xffffffff; 
//	    unmask_irq(irq);
    }

    if (action) {
	if (!(action->flags & SA_INTERRUPT))
	    __sti();
	action = *(irq + irq_action);
	do_random = 0;
	do {
	    do_random |= action->flags;
	    action->handler(irq, action->dev_id, regs);
	    action = action->next;
	} while (action);
	if (do_random & SA_SAMPLE_RANDOM)
	    add_interrupt_randomness(irq);
	IntClear3 = 0xffffffff;
	IntClear4 = 0xffffffff;
	IntClear5 = 0xffffffff;
	unmask_irq(irq);
	__cli();
    } else {
	    printk( "Unhandled interrupt %d came\n", irq );
	    IntClear3 = ~0;
	    IntClear4 = ~0;
	    IntClear5 = ~0;
	    unmask_irq(irq);
    }
    irq_exit(cpu);

    /* unmasking and bottom half handling is done magically for us. */
}

/*
 * Idea is to put all interrupts
 * in a single table and differenciate them just by number.
 */
int setup_r39xx_irq(int irq, struct irqaction *new)
{
    int shared = 0;
    struct irqaction *old, **p;
    unsigned long flags;

    p = irq_action + irq;
    if ((old = *p) != NULL) {
	/* Can't share interrupts unless both agree to */
	if (!(old->flags & new->flags & SA_SHIRQ))
	    return -EBUSY;

	/* Can't share interrupts unless both are same type */
	if ((old->flags ^ new->flags) & SA_INTERRUPT)
	    return -EBUSY;

	/* add new interrupt at end of irq queue */
	do {
	    p = &old->next;
	    old = *p;
	} while (old);
	shared = 1;
    }
    if (new->flags & SA_SAMPLE_RANDOM)
	rand_initialize_irq(irq);

    save_and_cli(flags);
    *p = new;

    if (!shared) {
	unmask_irq(irq);
    }
    restore_flags(flags);
    return 0;
}

int request_irq(unsigned int irq,
		void (*handler) (int, void *, struct pt_regs *),
		unsigned long irqflags,
		const char *devname,
		void *dev_id)
{
    int retval;
    struct irqaction *action;

    if (irq >= NR_IRQS)
	return -EINVAL;
    if (!handler)
	return -EINVAL;

    action = (struct irqaction *) kmalloc(sizeof(struct irqaction), GFP_KERNEL);
    if (!action)
	return -ENOMEM;

    action->handler = handler;
    action->flags = irqflags;
    action->mask = 0;
    action->name = devname;
    action->next = NULL;
    action->dev_id = dev_id;

    retval = setup_r39xx_irq(irq, action);

    if (retval)
	kfree(action);
    return retval;
}

void free_irq(unsigned int irq, void *dev_id)
{
    struct irqaction *action, **p;
    unsigned long flags;

    if (irq >= NR_IRQS) {
	printk("Trying to free IRQ%d\n", irq);
	return;
    }
    for (p = irq + irq_action; (action = *p) != NULL; p = &action->next) {
	if (action->dev_id != dev_id)
	    continue;

	/* Found it - now free it */
	save_and_cli(flags);
	*p = action->next;
	if (!irq[irq_action])
	    mask_irq(irq);
	restore_flags(flags);
	kfree(action);
	return;
    }
    printk("Trying to free free IRQ%d\n", irq);
}

unsigned long probe_irq_on(void)
{
    /* TODO */
    return 0;
}

int probe_irq_off(unsigned long irqs)
{
    /* TODO */
    return 0;
}

void __init init_IRQ(void)
{
    irq_setup();
}
