/*
 * Copyright (C) 2000 Yoichi Yuasa <yuasa@users.sourceforge.net>
 * All rights reserved.
 */

#include <linux/init.h>
#include <linux/malloc.h>
#include <linux/kernel.h>

#include <asm/pgtable.h>

#define KSEG2_PAGE_SHIFT	20
#define KSEG2_TABLE_SIZE	(KSEG2_MM_SIZE >> KSEG2_PAGE_SHIFT)
#define KSEG2_PAGE_SIZE		(1UL << KSEG2_PAGE_SHIFT)
#define KSEG2_PAGE_MASK		(~(KSEG2_PAGE_SIZE-1))

#define KSEG2_PFN_START(x)	(x->phys_addr & KSEG2_PAGE_MASK)
#define KSEG2_PFN_SIZE(x)	((x->size + KSEG2_PAGE_SIZE -1) & KSEG2_PAGE_MASK)
#define KSEG2_PFN_END(x)	(KSEG2_PFN_START(x) + KSEG2_PFN_SIZE(x))


#define KSEG2_VPN_START(x)	(x->kseg2_addr & KSEG2_PAGE_MASK & 0x3fffffff)
#define KSEG2_VPN_SIZE(x)	((x->size + KSEG2_PAGE_SIZE -1) & KSEG2_PAGE_MASK)
#define KSEG2_VPN_END(x)	(KSEG2_VPN_START(x) + KSEG2_VPN_SIZE(x))


#define KSEG2_OFFSET_SIZE(a, b)	((a & KSEG2_PAGE_MASK) - (b & KSEG2_PAGE_MASK))

#define KSEG2_FLAGS_MASK	0x00000f00

#undef KSEG2_DEBUG


typedef struct kseg2_map_t {
	unsigned long       phys_addr;
	unsigned long       size;
	unsigned long       kseg2_addr;
	struct kseg2_map_t *next;
} kseg2_map_t;
	
typedef struct kseg2_table_t {
	unsigned long pfn;
	unsigned long count;
} kseg2_table_t;

static kseg2_map_t kseg2_map = {0, 0, 0, NULL};
static kseg2_table_t kseg2_table[KSEG2_TABLE_SIZE];
static unsigned long kseg2_index = 0;

int kseg2_page_fault(unsigned long kseg2_addr)
{
	kseg2_table_t *table;
	unsigned long flags, entrylo, entryhi;
        unsigned long old_pagemask, ctx;
	int page;

	save_and_cli(flags);
        ctx = (get_entryhi() & 0xfc0);
        old_pagemask = get_pagemask();

	page = (kseg2_addr >> KSEG2_PAGE_SHIFT) & 0x3ff;
	table = kseg2_table + page;
	if (!table->count) {
		restore_flags(flags);
		return -1;
	}

	entrylo = table->pfn;
	entrylo &= (KSEG2_PAGE_MASK | KSEG2_FLAGS_MASK);
	//	entrylo |= (kseg2_addr & 0x000ff000);
#if defined(KSEG2_DEBUG) && !(defined(CONFIG_FB) && defined(CONFIG_PLUM2))
	printk("kseg2: addr=0x%08lx page=%d entry=0x%08lx\n",
	       kseg2_addr, page, entrylo);
#endif
	entryhi = (kseg2_addr & KSEG2_PAGE_MASK) | ctx;
	set_pagemask(PM_1M);
	set_entryhi(entryhi);
	set_entrylo0(entrylo);
	set_index((kseg2_index << 8) & 0x700);
	kseg2_index++;
	if (kseg2_index > 7)
		kseg2_index = 0;
	tlb_write_indexed();
	set_pagemask(old_pagemask);
	restore_flags(flags);

	return 0;
}

static int add_kseg2_map(kseg2_map_t *map, unsigned long phys_addr,
                         unsigned long size, unsigned long kseg2_addr)
{
	kseg2_map_t *new;

	new = kmalloc(sizeof(kseg2_map_t), GFP_KERNEL);
	if (!new)
		return -ENOMEM;

	new->phys_addr = phys_addr;
	new->size = size;
	new->kseg2_addr = kseg2_addr;
	new->next = map->next;
	map->next = new;

	return 0;
}

static int get_blank_entrys(unsigned long start, unsigned long end)
{
	kseg2_table_t *table = kseg2_table;
	int entry_start;
	int found;
	int i = 0;

	do {
		found = 1;
		while (i < KSEG2_TABLE_SIZE) {
			if (table->pfn == 0xffffffff)
				break;
			table++;
			i++;
		}

		entry_start = i;
		while (start < end) {
			if (i >= KSEG2_TABLE_SIZE) {
				entry_start = -1;
				break;
			}
			if (table->pfn != 0xffffffff) {
				found = 0;
				i++;
				break;
			}
			start += KSEG2_PAGE_SIZE;
			i++;
		}
	} while (!found);

	return entry_start;
}

int kseg2_page_alloc(unsigned long phys_addr, unsigned long size,
                     unsigned long flags, unsigned long *kseg2_addr)
{
	kseg2_map_t *map;
	kseg2_table_t *table;
	unsigned long f, end, entry;
	int entry_start = -1, entry_end, entry_size;
	int i;

	end = phys_addr + size;
#ifdef KSEG2_DEBUG
	printk("kseg2: phys_addr=0x%08lx size=0x%08lx\n", phys_addr, size);
#endif

	save_and_cli(f);
	for (map = &kseg2_map; map; map = map->next) {
		if (KSEG2_PFN_START(map) <= phys_addr &&
		    KSEG2_PFN_END(map) >= end) {
			entry_start = (KSEG2_VPN_START(map) +
				       KSEG2_OFFSET_SIZE(phys_addr, map->phys_addr))
					       >> KSEG2_PAGE_SHIFT;
			break;
		}
	}

	if (entry_start < 0) {
		entry_start = get_blank_entrys(phys_addr, end);
		if (entry_start < 0) {
			restore_flags(f);
			return -ENOMEM;
		}
	}

	entry_size = (size + KSEG2_PAGE_SIZE - 1) >> KSEG2_PAGE_SHIFT;
	if (!entry_size) {
		restore_flags(f);
		return -ENOMEM;
	}

	entry_end = entry_start + entry_size;

#ifdef KSEG2_DEBUG
	printk("kseg2: entry start=%d end=%d size=%d\n",
	       entry_start, entry_end, entry_size);
#endif

	entry = (phys_addr & KSEG2_PAGE_MASK) | (flags & KSEG2_FLAGS_MASK);
	table = kseg2_table + entry_start;
	for (i = entry_start; i < entry_end; i++) {
		table->pfn = entry;
		table->count++;
		table++;
		entry += KSEG2_PAGE_SIZE;
	}
	*kseg2_addr = KSEG2 | (entry_start << KSEG2_PAGE_SHIFT) |
	              (phys_addr & ~KSEG2_PAGE_MASK);
	add_kseg2_map(&kseg2_map, phys_addr, size, *kseg2_addr);
	restore_flags(f);

#ifdef KSEG2_DEBUG
	printk("kseg2: alloc page=0x%08lx size=0x%08lx\n", *kseg2_addr, size);
#endif
	return 0;
}

void kseg2_page_free(unsigned long phys_addr, unsigned long size,
                     unsigned long kseg2_addr)
{
	kseg2_map_t *map;
	kseg2_table_t *table;
	unsigned long flags;
	int entry_start, entry_end;
	int i;

	save_and_cli(flags);
	for (map = &kseg2_map; map; map = map->next) {
		if (map->phys_addr == phys_addr && map->size == size &&
		    map->kseg2_addr== kseg2_addr)
			break;
	}

	if (map) {
		entry_start = KSEG2_VPN_START(map) >> KSEG2_PAGE_SHIFT;
		entry_end = entry_start +
		            (KSEG2_VPN_SIZE(map) >> KSEG2_PAGE_SHIFT);
		table = kseg2_table + entry_start;
		for (i = entry_start; i < entry_end; i++) {
			table->count--;
			if (table->count <= 0)
				table->pfn = 0xffffffff;
		} 
	}
}

void __init kseg2_init(void)
{
	kseg2_table_t *table = kseg2_table;
	int i;

#ifdef KSEG2_DEBUG
	printk("kseg2: kseg2 init\n");
#endif
	for (i = 0; i < KSEG2_TABLE_SIZE; i++) {
		table->pfn = 0xffffffff;
		table->count = 0UL;
		table++;
	}
}
