/*
 * power - Routines for power management on SHARP Telios series
 *
 * Copyright (C) 2001 Hirokazu FUKUI <fuku@axe-inc.co.jp>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 */
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <asm/mipsregs.h>
#include <asm/system.h>
#include <asm/tx3920.h>

#ifndef CONFIG_PM

void r39xx_wait(void)
{
	/* not doing a HALT, but the more powersaving STOPCPU */
	PowerControl |= PWR_STOPCPU;
	/* now we wait... */
	PowerControl &= ~PWR_STOPCPU;
	asm volatile (
		"nop\n"
		"nop\n"
		"nop\n"
		"nop\n"
		"nop\n"
		"nop\n"
		"nop\n"
		);
}

#else /* CONFIG_PM */
void r39xx_wait(void)
{
	unsigned long flags;
	unsigned long oldrf;
	unsigned long int6;

	/* don't actualy handle the interrupts anymore */
	save_and_cli(flags);

	/* save and go to 9mhz (low power mode) */
	oldrf = read_32bit_cp0_register(CP0_CONF) & CONFF_RF;
	set_cp0_conf(CONFF_RF,3<<CONFB_RF);

	/* Enable the global interruptbit in case it was off.
	 * If it was off, we would never awaken from the
	 * STOPCPU. Which would be a bit sad for the user */
	int6 = IntEnable6;
	IntEnable6 |= INT6_GLOBALEN;
	PowerControl |=  PWR_STOPCPU;
	/* ...we wait here for an interrupt to happen... */
	/* Most probably a 10ms periodic interrupt */
	PowerControl &= ~PWR_STOPCPU;
	IntEnable6 = int6;

	/* restore previous speed */
	set_cp0_conf(CONFF_RF,oldrf);

	/* restore interrupts */
	restore_flags(flags);
}
#endif /* CONFIG_PM */

#ifdef CONFIG_SHARP_TELIOS
#include "sharp_telios.h"

#undef DEBUG
#define FAKE_APM

#define CLOCK_FREQ		3072
#define CLOCK_PERIOD_HALF	((RTC_PERSEC_COUNT / CLOCK_FREQ) / 2)

#ifdef CONFIG_PROC_FS
static int __init telios_battery_proc_init(void);
#endif

static inline unsigned period_rtc_sync(void)
{
    unsigned high, low, low2;
    tx3920_read_rtc(high, low);
    do{
	tx3920_read_rtc(high, low2);
    }while( low == low2 );
    return low2;
}

static inline int period_rtc_wait(unsigned wlow)
{
    unsigned high, low;
    do{
	tx3920_read_rtc(high, low);
	if( low == wlow )
	    return 0;
    }while( (int)(wlow - low) > 0 );
    return -1;
}


static void send_cmd(int cmd)
{
    unsigned cnt, tmp, bit, flags;
    int t, err=0;

    save_and_cli(flags);
    cnt = period_rtc_sync();
    for( t=0, bit=0x80; t<8; t++, bit>>=1 ){
	tmp = IODataInOut;
	tmp |= TELIOS_IO_MBU_CLOCK;
	if( cmd & bit )
	    tmp |= (TELIOS_IO_MBU_DATA_OUT);
	else
	    tmp &= ~(TELIOS_IO_MBU_DATA_OUT);
	IODataInOut = tmp;
	if( period_rtc_wait(cnt = (cnt + CLOCK_PERIOD_HALF)) != 0 )
	    err = -1;
	IODataInOut = tmp & ~TELIOS_IO_MBU_CLOCK;
	if( period_rtc_wait(cnt = (cnt + CLOCK_PERIOD_HALF)) != 0 )
	    err = -1;
    }
    restore_flags(flags);
    if( err != 0 )
	printk("battery.c: send error.\n");
}

static int recv_data(void)
{
    unsigned cnt, tmp, bit, data, flags;
    int t, err=0;

    data = 0;
    save_and_cli(flags);
    cnt = period_rtc_sync();
    for( t=0, bit=0x80; t<8; t++, bit>>=1 ){
	IODataInOut |= TELIOS_IO_MBU_CLOCK;
	if( period_rtc_wait(cnt = (cnt + CLOCK_PERIOD_HALF)) != 0 )
	    err = -1;
	tmp = IODataInOut;
	if( tmp & TELIOS_IO_MBU_DATA_IN )
	    data |= bit;
	IODataInOut = tmp & ~TELIOS_IO_MBU_CLOCK;
	if( period_rtc_wait(cnt = (cnt + CLOCK_PERIOD_HALF)) != 0 )
	    err = -1;
    }
    restore_flags(flags);
    if( err != 0 )
	printk("battery.c: receive error.\n");
    return data;
}

static int get_BMU_status(int cmd)
{
    int data;

    send_cmd(cmd);
    udelay(7000);
    data = recv_data();
    udelay(7000);
    return data;
}

static int __init telios_battery_init(void)
{
  unsigned flags;

  /*  init_waitqueue_head(&battery_wait);*/

  save_and_cli(flags);
  IODataInOut &= ~TELIOS_IO_MBU_CLOCK;
  restore_flags(flags);

#ifdef CONFIG_PROC_FS
    telios_battery_proc_init();
#endif
  return 0;
}
__initcall(telios_battery_init);

#ifdef CONFIG_PROC_FS

static int atoi(const char *p)
{
    int n = 0;
    while (*p && '0' <= *p && *p <= '9') {
	n *= 10;
	n += *p - '0';
	p++;
    }
    return (n);
}

static int next_tokn(char *str, int p)
{
    while(str[p] >= '0' && str[p] <= '9')
	p++;
    while(!(str[p] >= '0' && str[p] <= '9') && str[p]!='\0')
	p++;
    return p;
}

#ifdef DEBUG
static int debug_cmd_num=0;
static unsigned char debug_cmd[2];

static ssize_t telios_battery_read(struct file *file, char *buf, size_t len,
					loff_t *ppos)
{
    static char tmpbuf[64];
    size_t rlen;
    int data, i;

    if (*ppos != 0)
	return 0;

    data = get_BMU_status(0x00);
    sprintf(tmpbuf, "0x%02x %d\n", (~data)&0xff, (((~data) & 0xff) >> 1));
    rlen = strlen(tmpbuf);

    for(i=0; i < debug_cmd_num; i++){
	data = get_BMU_status(debug_cmd[i]);
	sprintf(tmpbuf+rlen, "; cmd %02x, data %02x\n", debug_cmd[i], data);
	rlen += strlen(tmpbuf+rlen);
    }
    debug_cmd_num = 0;

    if (len < rlen)
	return -EIO;
    if (copy_to_user(buf, tmpbuf, rlen))
	return -EFAULT;
    *ppos += rlen;
    return rlen;
}

static ssize_t
telios_battery_write(struct file *file, const char *buf, size_t len, loff_t *ppos)
{
    static char tmpbuf[24];
    int i;

    if (*ppos != 0)
	return -EIO;

    if (len > 24-1)
	return -EIO;

    if (copy_from_user(tmpbuf, buf, len))
	return -EFAULT;
    tmpbuf[len] = '\0';

    i = 0;
    debug_cmd_num = 0;
    if (tmpbuf[i] >= '0' && tmpbuf[i] <= '9')	
	debug_cmd[debug_cmd_num++] = atoi(tmpbuf);
    else
	return -EIO;

    i = next_tokn(tmpbuf, i);

    if (tmpbuf[i] >= '0' && tmpbuf[i] <= '9')	
	debug_cmd[debug_cmd_num++] = atoi(tmpbuf+i);

   *ppos += len;
    return len;
}

static struct file_operations battery_fops = 
{
	read:	telios_battery_read,
	write:	telios_battery_write,
};

static struct proc_dir_entry *proc_root_battery;

/*  power ctrl debug */
#ifdef CONFIG_PLUM2
#include <asm/plum2.h>
#endif

static ssize_t debug_powctrl_read(struct file *file, char *buf, size_t len,
					loff_t *ppos)
{
    static char tmpbuf[64];
    size_t rlen;

    if (*ppos != 0)
	return 0;

    rlen = 0;
#ifdef CONFIG_PLUM2
    sprintf(tmpbuf+rlen, "%d: %s %d(0x%x)\n",
	    0, "Plum2_PowerCtrl",
	    Plum2_PowerCtrl, Plum2_PowerCtrl);
    rlen += strlen(tmpbuf+rlen);
#endif

#ifdef CONFIG_SHARP_TELIOS
    sprintf(tmpbuf+rlen, "%d: %s %d(0x%x)\n",
	    1, "BackLight_VR",
	    BackLight_VR, BackLight_VR);
    rlen += strlen(tmpbuf+rlen);

    sprintf(tmpbuf+rlen, "%d: %s %d(0x%x)\n",
	    2, "BackLight_Reset",
	    BackLight_Reset, BackLight_Reset);
    rlen += strlen(tmpbuf+rlen);

#endif

    if (len < rlen)
	return -EIO;
    if (copy_to_user(buf, tmpbuf, rlen))
	return -EFAULT;
    *ppos += rlen;
    return rlen;
}

static ssize_t
debug_powctrl_write(struct file *file, const char *buf, size_t len, loff_t *ppos)
{
    static char tmpbuf[80];
    int i;
    int reg, val;

    if (*ppos != 0)
	return -EIO;

    if (len > 80-1)
	return -EIO;

    if (copy_from_user(tmpbuf, buf, len))
	return -EFAULT;
    tmpbuf[len] = '\0';

    i = 0;
    
    debug_cmd_num = 0;
    if (tmpbuf[i] >= '0' && tmpbuf[i] <= '9')	
	reg = atoi(tmpbuf);
    else
	return -EIO;

    i = next_tokn(tmpbuf, i);

    if (tmpbuf[i] >= '0' && tmpbuf[i] <= '9')	
	val = atoi(tmpbuf+i);
    else
	val = 0;

    switch(reg){
#ifdef CONFIG_PLUM2
    case 0:
	Plum2_PowerCtrl = val;
	printk("Plum2_PowerCtrl = 0x%x\n", val);
	break;
#endif
#ifdef CONFIG_SHARP_TELIOS
    case 1:
	BackLight_VR = val;
	printk("BackLight_VR = 0x%x\n", val);
	break;
    case 2:
	BackLight_Reset = val;
	printk("BackLight_Reset = 0x%x\n", val);
	break;
#endif
    }

   *ppos += len;
    return len;
}

static struct file_operations debug_powctrl_fops = 
{
	read:	debug_powctrl_read,
	write:	debug_powctrl_write,
};

static struct proc_dir_entry *proc_root_debug_powctrl;

#endif /* DEBUG */

#ifdef FAKE_APM
static char fake_apm_driver_version[] = "1.13";	/* no spaces */
static int fake_apm_bios_version = 0x12;

static int telios_apm_get_info(char *buf, char **start, off_t fpos, int length)
{
	char *		p;
	unsigned short  ac_line_status = 0xff;
	unsigned short  battery_status = 0xff;
	unsigned short  battery_flag   = 0xff;
	int		percentage     = -1;
	int             time_units     = -1;
	char            *units         = "?";

	p = buf;

	ac_line_status = (IODataInOut & TELIOS_IO_DC_JACK_N) ? 0x00 : 0x01;

	percentage = (~get_BMU_status(0x00) & 0xff) >> 1;

	/* Arguments, with symbols from linux/apm_bios.h.  Information is
	   from the Get Power Status (0x0a) call unless otherwise noted.

	   0) Linux driver version (this will change if format changes)
	   1) APM BIOS Version.  Usually 1.0, 1.1 or 1.2.
	   2) APM flags from APM Installation Check (0x00):
	      bit 0: APM_16_BIT_SUPPORT
	      bit 1: APM_32_BIT_SUPPORT
	      bit 2: APM_IDLE_SLOWS_CLOCK
	      bit 3: APM_BIOS_DISABLED
	      bit 4: APM_BIOS_DISENGAGED
	   3) AC line status
	      0x00: Off-line
	      0x01: On-line
	      0x02: On backup power (BIOS >= 1.1 only)
	      0xff: Unknown
	   4) Battery status
	      0x00: High
	      0x01: Low
	      0x02: Critical
	      0x03: Charging
	      0x04: Selected battery not present (BIOS >= 1.2 only)
	      0xff: Unknown
	   5) Battery flag
	      bit 0: High
	      bit 1: Low
	      bit 2: Critical
	      bit 3: Charging
	      bit 7: No system battery
	      0xff: Unknown
	   6) Remaining battery life (percentage of charge):
	      0-100: valid
	      -1: Unknown
	   7) Remaining battery life (time units):
	      Number of remaining minutes or seconds
	      -1: Unknown
	   8) min = minutes; sec = seconds */

	p += sprintf(p, "%s %d.%d 0x%02x 0x%02x 0x%02x 0x%02x %d%% %d %s\n",
		     fake_apm_driver_version,
		     (fake_apm_bios_version >> 8) & 0xff,
		     fake_apm_bios_version & 0xff,
		     0x02,
		     ac_line_status,
		     battery_status,
		     battery_flag,
		     percentage,
		     time_units,
		     units);

	return p - buf;
}
#endif /* FAKE_APM */

static int __init telios_battery_proc_init(void)
{
#ifdef DEBUG
	proc_root_battery = create_proc_entry("battery", S_IWUSR | S_IRUGO,
					      &proc_root);
	proc_root_battery->proc_fops = &battery_fops;
	proc_root_debug_powctrl = create_proc_entry("debug_powctrl", S_IWUSR | S_IRUGO,
					      &proc_root);
	proc_root_debug_powctrl->proc_fops = &debug_powctrl_fops;
#endif
#ifdef FAKE_APM
	create_proc_info_entry("apm", 0, NULL, telios_apm_get_info);
#endif

	return 0;
}
#endif /* CONFIG_PROC_FS */
#endif /* CONFIG_SHARP_TELIOS */
