/*
 * SPI - Routines for SPI Module on TMPR3922
 *       (The SPI is a serial interface consisting of clock, data out and
 *         data in)
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2001 Hirokazu FUKUI <fuku@axe-inc.co.jp>
 */

#include <linux/sched.h>
#include <asm/mipsregs.h>
#include <asm/system.h>
#include <asm/tx3920.h>

#include "spi.h"

#define NO_INTERRUPT

int spi_is_empty(void)
{
    return SPIControl & SPI_EMPTY;
}

int spi_put_word(int w)
{
#ifdef NO_INTERRUPT
    while(!(IntStatus5&INT5_SPIBUFAVAILINT))
	;
    IntClear5 = INT5_SPIBUFAVAILINT;
#endif
    SPITransmitHolding = (w & 0xffff);
    return 0;
}


int spi_get_word(void)
{
#ifdef NO_INTERRUPT
    while(!(IntStatus5&INT5_SPIRCVINT))
	;
    IntClear5 = INT5_SPIRCVINT;
#endif
    return SPIReceiveHolding & 0xffff;
}

void spi_enable(int enable)
{
    if( enable )
	SPIControl |= SPI_ENSPI;
    else
	SPIControl &= ~SPI_ENSPI;
}

void spi_enable_interrupt(int int_flags)
{
    unsigned flags;
    unsigned temp;

    save_and_cli(flags);

    temp = IntEnable5;
    temp &= ~(INT5_SPIBUFAVAILINT|INT5_SPIERRINT|INT5_SPIRCVINT|
		    INT5_SPIEMPTYINT);
    temp |= (INT5_SPIBUFAVAILINT|INT5_SPIERRINT|INT5_SPIRCVINT|
	     INT5_SPIEMPTYINT) & int_flags;

    restore_flags(flags);
}

/*
 *  Interrupt routine
 */
static void spi_bufvalint_handler_default(void)
{
    printk("spi_errint_handler_default: not set spi_bufvalint_handler\n");
}

static void spi_errint_handler_default(void)
{
    printk("spi_errint_handler_default: not set spi_errint_handler\n");
}

static void spi_rcvint_handler_default(void)
{
    printk("spi_errint_handler_default: not set spi_rcvint_handler\n");
}

static void spi_emptyint_handler_default(void)
{
    printk("spi_errint_handler_default: not set spi_emptyint_handler\n");
}

void (*spi_bufvalint_handler)(void) = spi_bufvalint_handler_default;
void (*spi_errint_handler)(void) = spi_errint_handler_default;
void (*spi_rcvint_handler)(void) = spi_rcvint_handler_default;
void (*spi_emptyint_handler)(void) = spi_emptyint_handler_default;

#ifndef NO_INTERRUPT
static void spi_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
    unsigned status;

    status = IntStatus5;
    IntEnable5 &= ~(INT5_SPIBUFAVAILINT|INT5_SPIERRINT|INT5_SPIRCVINT|
		    INT5_SPIEMPTYINT);
    IntClear5 = INT5_SPIBUFAVAILINT|INT5_SPIERRINT|INT5_SPIRCVINT|
	INT5_SPIEMPTYINT;


    if(status & INT5_SPIBUFAVAILINT ){
	(*spi_bufvalint_handler)();
    }

    if( status & INT5_SPIRCVINT ){
	(*spi_rcvint_handler)();
    }
    if( status & INT5_SPIERRINT ){
	(*spi_errint_handler)();
    }
    if( status & INT5_SPIEMPTYINT ){
	(*spi_emptyint_handler)();
    }
}
#endif /* !NO_INTERRUPT */

/*
 *  Initialization routine
 */

void __init spi_init(void)
{
    int flags;

    save_and_cli(flags);

    printk(KERN_ERR "spi: SPIControl %08x\n", SPIControl);
    SPIControl &= ~SPI_ENSPI;

#ifdef CONFIG_SHARP_TELIOS
    {
	unsigned temp;
	/* Specific settings for OptPoint of HC-AJ2 */
	temp = SPIControl;
	temp &= ~SPI_DELAYVAL_MASK;
	temp |= (0 << SPI_DELAYVAL_SHIFT);
	temp &= ~SPI_BAUDRATE_MASK;
	temp |= (4 << SPI_BAUDRATE_SHIFT);	/* SPICLK Rate = 737.3 kHz */
	temp &= ~SPI_WORD;			/* Use 8bits of data */
	temp &= ~SPI_PHAPOL;
	temp |= SPI_CLKPOL;
	temp &= ~SPI_LSB;			/* MSB first */
	SPIControl = temp;
    }
#endif

    IntClear5 = INT5_SPIBUFAVAILINT|INT5_SPIERRINT|INT5_SPIRCVINT|
	INT5_SPIEMPTYINT;
#ifndef NO_INTERRUPT
    i = request_irq(21, spi_interrupt, SA_INTERRUPT, "spi", NULL);
    if ( i != 0 ) {
	printk(KERN_ERR "spi: Cannot allocate irq. (%d)\n", i);
    }
#endif /* !NO_INTERRUPT */

    restore_flags(flags);
}
