/*
 * vrtp - Generate Linux Input driver events from vrtpanel raw events.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2001 Hirokazu FUKUI <fuku@axe-inc.co.jp>
 *
 * Merge a part of:
 *   drivers/macintosh/mac_hid.c
 *   Copyright (C) 2000 Franz Sirl.
 */

#include <linux/input.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>

#ifndef CONFIG_INPUT
#error Touch-panel event support is need Input core support.
#endif

#ifndef CONFIG_INPUT_KEYBDEV
#error Touch-panel event support is need Input Keyboard support.
#endif

static int mouse_move_shift_keycode = KEY_INTL2;	/* Hiragana key */
static int mouse_button2_shift_keycode = KEY_INTL5;	/* Muhenkan key */
static int mouse_button3_shift_keycode = KEY_INTL4;	/* Hekan key */

#define FL_TOUCH   1
#define FL_LEFT    2
#define FL_MIDDLE  4
#define FL_RIGHT   8

static int button_mode = FL_LEFT|FL_TOUCH;

struct vrtpcal_pack{
    int a, b, c, d, e, f, s;
};

struct noise_filter_param{
    int pressure_low_limit;
    int shift_bits;
    int output_threshold;
    int data_change_limit;
    int ignore_count;
};

struct noise_filter_stat{
    int ignore;
    int count;
    int accum_x;
    int accum_y;
    int accum_z;
    int have_last_data;
    int last_x;
    int last_y;
};

struct vrtp{
    struct input_dev dev;
    struct vrtpcal_pack tpcal;
    struct noise_filter_param nf_param;
    struct noise_filter_stat nf_stat;
    
    int max_absx;
    int max_absy;
};

struct vrtp vrtp;

#if defined(CONFIG_SHARP_TELIOS)
static struct vrtpcal_pack default_tpcal = {
    14783, 54, 110441492, -102, -10737, 63128930, 65536,
};

static struct noise_filter_param default_nf_param = {
    600,	/* pressure_low_limit */
    2,		/* shift_bits */
    1,		/* output_threshold */
    360,	/* data_change_limit */
    0,		/* ignore_count */
};
#elif defined(CONFIG_NEC_MOBILEGEAR2_R530)
static struct vrtpcal_pack default_tpcal = {
    14783, 54, 110441492, -102, -10737, 63128930, 65536,
};

static struct noise_filter_param default_nf_param = {
    600,	/* pressure_low_limit */
    2,		/* shift_bits */
    1,		/* output_threshold */
    360,	/* data_change_limit */
    3,		/* ignore_count */
};
#else
static struct vrtpcal_pack default_tpcal = {
    14783, 54, 110441492, -102, -10737, 63128930, 65536,
};

static struct noise_filter_param default_nf_param_vj1 = {
    600,	/* pressure_low_limit */
    2,		/* shift_bits */
    1,		/* output_threshold */
    360,	/* data_change_limit */
    3,		/* ignore_count */
};
#endif

static inline int vrtpanel_calx(int rx, int ry)
{
    return (vrtp.tpcal.a * rx +
	    vrtp.tpcal.b * ry +
	    vrtp.tpcal.c) / vrtp.tpcal.s;
}

static inline int vrtpanel_caly(int rx, int ry)
{
    return (vrtp.tpcal.d * rx +
	    vrtp.tpcal.e * ry +
	    vrtp.tpcal.f) / vrtp.tpcal.s;
}

static int report_button(int down)
{
    static int button_status = 0;
    int last_status, chg, rep=0;

    last_status = button_status;

    if( down == -1 ){
	down = (last_status & FL_TOUCH) ? 1 : 0;
	button_status &= ~(FL_TOUCH|FL_LEFT|FL_MIDDLE|FL_RIGHT);
    }

    if( down ){
	button_status |= (button_mode &
			  (FL_TOUCH|FL_LEFT|FL_MIDDLE|FL_RIGHT));
    }else{
	button_status &= ~(FL_TOUCH|FL_LEFT|FL_MIDDLE|FL_RIGHT);
    }
    chg = last_status ^ button_status;
#ifdef NEVER
    if( chg & FL_TOUCH ){
	input_report_key(&vrtp.dev, BTN_TOUCH, (button_status&FL_TOUCH)?1:0);
	rep = 1;
    }
#endif
    if( chg & FL_LEFT ){
	input_report_key(&vrtp.dev, BTN_LEFT, (button_status&FL_LEFT)?1:0);
	rep = 1;
    }
    if( chg & FL_MIDDLE ){
	input_report_key(&vrtp.dev, BTN_MIDDLE, (button_status&FL_MIDDLE)?1:0);
	rep = 1;
    }
    if( chg & FL_RIGHT ){
	input_report_key(&vrtp.dev, BTN_RIGHT, (button_status&FL_RIGHT)?1:0);
	rep = 1;
    }
    /*
    printk(KERN_INFO "report_button: stat(%02x>%02x) chg=%02x mode=%02x\n",
	   last_status, button_status, chg, button_mode);
    */
    return rep;
}

static void noise_filter_reset(void)
{
    vrtp.nf_stat.ignore = 1;
    vrtp.nf_stat.count = 0;
    vrtp.nf_stat.accum_x = 0;
    vrtp.nf_stat.accum_y = 0;
    vrtp.nf_stat.accum_z = 0;
    vrtp.nf_stat.have_last_data = 0;
}

static int noise_filter(int *px, int *py, int *pz)
{
    if( vrtp.nf_stat.ignore ){
	if( vrtp.nf_stat.count++ < vrtp.nf_param.ignore_count ){
	    return 0;
	}else{
	    vrtp.nf_stat.ignore = 0;
	    vrtp.nf_stat.count = 0;
	}
    }
    if( *pz < vrtp.nf_param.pressure_low_limit )
	return 0;
    if( vrtp.nf_stat.have_last_data ){
	if( (abs(*px - vrtp.nf_stat.last_x) >
	     vrtp.nf_param.data_change_limit)
	    || (abs(*py - vrtp.nf_stat.last_y) >
		vrtp.nf_param.data_change_limit) )
	    return 0;
    }

    vrtp.nf_stat.have_last_data = 1;
    vrtp.nf_stat.last_x = *px;
    vrtp.nf_stat.last_y = *py;

    if( vrtp.nf_param.shift_bits <= 0 ){
	*px *= 8;
	*py *= 8;
	*pz *= 8;
	return 1;
    }

    if( vrtp.nf_stat.count == (1<<vrtp.nf_param.shift_bits) ){
	vrtp.nf_stat.accum_x -=
	    vrtp.nf_stat.accum_x >> vrtp.nf_param.shift_bits;
	vrtp.nf_stat.accum_y -=
	    vrtp.nf_stat.accum_y >> vrtp.nf_param.shift_bits;
	vrtp.nf_stat.accum_z -=
	    vrtp.nf_stat.accum_z >> vrtp.nf_param.shift_bits;
	vrtp.nf_stat.count--;
    }

    vrtp.nf_stat.accum_x += *px;
    vrtp.nf_stat.accum_y += *py;
    vrtp.nf_stat.accum_z += *pz;
    vrtp.nf_stat.count++;

    if( vrtp.nf_stat.count <= vrtp.nf_param.output_threshold )
	return 0;

    *px = (vrtp.nf_stat.accum_x * 8) / vrtp.nf_stat.count;
    *py = (vrtp.nf_stat.accum_y * 8) / vrtp.nf_stat.count;
    *pz = (vrtp.nf_stat.accum_z * 8) / vrtp.nf_stat.count;
    return 1;
}

void vrtp_report_event(int pen_state, int lost,
		       int xp, int yp, int xm, int ym, int pressure)
{
    int rx, ry, rz, tx, ty, tz;

    switch(pen_state){
    case 0:	/* pen up */
	noise_filter_reset();
	if( report_button(0) )
	    input_event(&vrtp.dev, EV_MSC, MSC_SERIAL, 0);
	break;
    case 1:	/* pen down */
	break;
    case 2:
	rx = (xm - xp);
	ry = (ym - yp);
	rz = pressure;
	if( noise_filter(&rx, &ry, &rz) == 0 )
	    break;
	tx = vrtpanel_calx(rx, ry);
	ty = vrtpanel_caly(rx, ry);
	tz = rz;
	if( tx < 0 )
	    tx = 0;
	else if( tx > vrtp.dev.absmax[ABS_X] )
	    tx = vrtp.dev.absmax[ABS_X];
	if( ty < 0 )
	    ty = 0;
	else if( ty > vrtp.dev.absmax[ABS_Y] )
	    ty = vrtp.dev.absmax[ABS_Y];
	input_report_abs(&vrtp.dev, ABS_X, tx);
	input_report_abs(&vrtp.dev, ABS_Y, ty);
	input_report_abs(&vrtp.dev, ABS_PRESSURE, tz);

	report_button(1);
	input_event(&vrtp.dev, EV_MSC, MSC_SERIAL, 0);
	break;
    }
}

int tpev_report_key(unsigned int keycode, int down)
{
    /*
    printk(KERN_INFO "tpev_report_key: keycode=%d down=%d\n", keycode, down);
    */
    if( down == 2 )	/* key repeat event */
	return 0;

    if( keycode == mouse_move_shift_keycode ){
	if( down ){
	    button_mode &= ~FL_LEFT;
	}else{
	    if( !(button_mode & (FL_MIDDLE | FL_RIGHT)) )
		button_mode |= FL_LEFT;
	}
	if( report_button(-1) )
	    input_event(&vrtp.dev, EV_MSC, MSC_SERIAL, 0);
    }else if( keycode == mouse_button2_shift_keycode ){
	if( down ){
	    button_mode &= ~FL_LEFT;
	    button_mode |= FL_MIDDLE;
	}else{
	    button_mode &= ~FL_MIDDLE;
	    if( !(button_mode & (FL_MIDDLE | FL_RIGHT)) )
		button_mode |= FL_LEFT;
	}
	if( report_button(-1) )
	    input_event(&vrtp.dev, EV_MSC, MSC_SERIAL, 0);
    }else if( keycode == mouse_button3_shift_keycode ){
	if( down ){
	    button_mode &= ~FL_LEFT;
	    button_mode |= FL_RIGHT;
	}else{
	    button_mode &= ~FL_RIGHT;
	    if( !(button_mode & (FL_MIDDLE | FL_RIGHT)) )
		button_mode |= FL_LEFT;
	}
	if( report_button(-1) )
	    input_event(&vrtp.dev, EV_MSC, MSC_SERIAL, 0);
    }
    return 0;
}

#if defined(CONFIG_PROC_FS)
static int atoi(const char *p)
{
	int n = 0, sign;

	if( *p == '-' ){
	    sign = -1;
	    p++;
	}else
	    sign = 1;

	while (*p && '0' <= *p && *p <= '9') {
		n *= 10;
		n += *p - '0';
		p++;
	}
	return n * sign;
}

static int next_tokn(char *str, int p)
{
    if(str[p] == '-' )
	p++;
    while(str[p] >= '0' && str[p] <= '9')
	p++;
    while(!((str[p] >= '0' && str[p] <= '9') || str[p] == '-') )
	p++;
    return p;
}

static ssize_t
vrtp_read_int(struct file *file, char *buf, size_t len, loff_t *ppos, int val)
{
    size_t rlen;
    static char tmpbuf[16];

    if (*ppos != 0)
	return 0;

    rlen = sprintf(tmpbuf, "%d\n", val);

    if (len < rlen)
	return -EIO;
    if (copy_to_user(buf, tmpbuf, rlen))
	return -EFAULT;
    *ppos += rlen;
    return rlen;
}

static ssize_t
vrtp_write_int(struct file *file, const char *buf, size_t len, loff_t *ppos,
	   int *pval)
{
    static char tmpbuf[64];
    int i;

    if (*ppos != 0)
	return -EIO;

    if (len > 63)
	return -EIO;

    if (copy_from_user(tmpbuf, buf, len))
	return -EFAULT;
    tmpbuf[len] = '\0';

    i = 0;
    if (tmpbuf[i] >= '0' && tmpbuf[i] <= '9')	
	*pval = atoi(tmpbuf);
    else
	return -EIO;

    *ppos += len;
    return len;
}

static ssize_t
vrtp_pointercal_read(struct file *file, char *buf, size_t len, loff_t *ppos)
{
    size_t rlen;
    static char tmpbuf[80];

    if (*ppos != 0)
	return 0;

    rlen = sprintf(tmpbuf, "%d %d %d %d %d %d %d\n",
		   vrtp.tpcal.a, vrtp.tpcal.b, vrtp.tpcal.c, vrtp.tpcal.d, 
		   vrtp.tpcal.e, vrtp.tpcal.f, vrtp.tpcal.s);

    if (len < rlen)
	return -EIO;
    if (copy_to_user(buf, tmpbuf, rlen))
	return -EFAULT;
    *ppos += rlen;
    return rlen;
}

static ssize_t
vrtp_pointercal_write(struct file *file, const char *buf, size_t len,
		      loff_t *ppos)
{
    static char tmpbuf[128];
    struct vrtpcal_pack p;
    int i;

    if (*ppos != 0)
	return -EIO;

    if (len > 128)
	return -EIO;

    if (copy_from_user(tmpbuf, buf, len))
	return -EFAULT;
    tmpbuf[len] = '\0';

    i = 0;
    if( (tmpbuf[i] >= '0' && tmpbuf[i] <= '9') || tmpbuf[i] == '-' )
	p.a = atoi(tmpbuf);
    else
	return -EIO;

    i = next_tokn(tmpbuf, i);
    if( (tmpbuf[i] >= '0' && tmpbuf[i] <= '9') || tmpbuf[i] == '-' )
	p.b = atoi(tmpbuf+i);
    else
	return -EIO;

    i = next_tokn(tmpbuf, i);
    if( (tmpbuf[i] >= '0' && tmpbuf[i] <= '9') || tmpbuf[i] == '-' )
	p.c = atoi(tmpbuf+i);
    else
	return -EIO;

    i = next_tokn(tmpbuf, i);
    if( (tmpbuf[i] >= '0' && tmpbuf[i] <= '9') || tmpbuf[i] == '-' )
	p.d = atoi(tmpbuf+i);
    else
	return -EIO;

    i = next_tokn(tmpbuf, i);
    if( (tmpbuf[i] >= '0' && tmpbuf[i] <= '9') || tmpbuf[i] == '-' )
	p.e = atoi(tmpbuf+i);
    else
	return -EIO;

    i = next_tokn(tmpbuf, i);
    if( (tmpbuf[i] >= '0' && tmpbuf[i] <= '9') || tmpbuf[i] == '-' )
	p.f = atoi(tmpbuf+i);
    else
	return -EIO;

    i = next_tokn(tmpbuf, i);
    if( (tmpbuf[i] >= '0' && tmpbuf[i] <= '9') || tmpbuf[i] == '-' )
	p.s = atoi(tmpbuf+i);
    else
	return -EIO;

    vrtp.tpcal = p;
    *ppos += len;
    return len;
}

static ssize_t
vrtp_mouse_move_shift_keycode_read(struct file * file, char * buf,
					size_t count, loff_t *ppos)
{
    return vrtp_read_int(file, buf, count, ppos, mouse_move_shift_keycode);
}

static ssize_t
vrtp_mouse_move_shift_keycode_write(struct file * file, const char * buf,
					 size_t count, loff_t *ppos)
{
    return vrtp_write_int(file, buf, count, ppos,
			  &mouse_move_shift_keycode);
}

static ssize_t
vrtp_mouse_button2_shift_keycode_read(struct file * file, char * buf,
					size_t count, loff_t *ppos)
{
    return vrtp_read_int(file, buf, count, ppos, mouse_button2_shift_keycode);
}

static ssize_t
vrtp_mouse_button2_shift_keycode_write(struct file * file, const char * buf,
					 size_t count, loff_t *ppos)
{
    return vrtp_write_int(file, buf, count, ppos,
			  &mouse_button2_shift_keycode);
}

static ssize_t
vrtp_mouse_button3_shift_keycode_read(struct file * file, char * buf,
					size_t count, loff_t *ppos)
{
    return vrtp_read_int(file, buf, count, ppos, mouse_button3_shift_keycode);
}

static ssize_t
vrtp_mouse_button3_shift_keycode_write(struct file * file, const char * buf,
					 size_t count, loff_t *ppos)
{
    return vrtp_write_int(file, buf, count, ppos,
			  &mouse_button3_shift_keycode);
}

static ssize_t
vrtp_nf_shift_bits_read(struct file * file, char * buf,
					size_t count, loff_t *ppos)
{
    return vrtp_read_int(file, buf, count, ppos, vrtp.nf_param.shift_bits);
}

static ssize_t
vrtp_nf_shift_bits_write(struct file * file, const char * buf,
			 size_t count, loff_t *ppos)
{
    return vrtp_write_int(file, buf, count, ppos, &vrtp.nf_param.shift_bits);
}

static ssize_t
vrtp_nf_ignore_count_read(struct file * file, char * buf,
			  size_t count, loff_t *ppos)
{
    return vrtp_read_int(file, buf, count, ppos, vrtp.nf_param.ignore_count);
}

static ssize_t
vrtp_nf_ignore_count_write(struct file * file, const char * buf,
			   size_t count, loff_t *ppos)
{
    return vrtp_write_int(file, buf, count, ppos, &vrtp.nf_param.ignore_count);
}

struct file_operations vrtp_pointercal_fops = {
    read:		vrtp_pointercal_read,
    write:		vrtp_pointercal_write,
};

struct file_operations vrtp_mouse_move_shift_keycode_fops = {
    read:		vrtp_mouse_move_shift_keycode_read,
    write:		vrtp_mouse_move_shift_keycode_write,
};

struct file_operations vrtp_mouse_button2_shift_keycode_fops = {
    read:		vrtp_mouse_button2_shift_keycode_read,
    write:		vrtp_mouse_button2_shift_keycode_write,
};

struct file_operations vrtp_mouse_button3_shift_keycode_fops = {
    read:		vrtp_mouse_button3_shift_keycode_read,
    write:		vrtp_mouse_button3_shift_keycode_write,
};

struct file_operations vrtp_nf_shift_bits_fops = {
    read:		vrtp_nf_shift_bits_read,
    write:		vrtp_nf_shift_bits_write,
};

struct file_operations vrtp_nf_ignore_count_fops = {
    read:		vrtp_nf_ignore_count_read,
    write:		vrtp_nf_ignore_count_write,
};

static struct proc_dir_entry *proc_vrtp_root = NULL;

static int __init vrtp_proc_init(void)
{
    struct proc_dir_entry *proc_pointercal, *proc_move,
	*proc_button2, *proc_button3,
	*proc_nf_shift_bits, *proc_nf_ignore_count;

    proc_vrtp_root = proc_mkdir("vrtp", 0);
    if( proc_vrtp_root ){
	proc_pointercal = create_proc_entry("pointercal",
					 S_IWUSR | S_IRUGO, proc_vrtp_root);
	if( proc_pointercal )
	    proc_pointercal->proc_fops = &vrtp_pointercal_fops;

	proc_move = create_proc_entry("mouse_move_shift_keycode",
					 S_IWUSR | S_IRUGO, proc_vrtp_root);
	if( proc_move )
	    proc_move->proc_fops = &vrtp_mouse_move_shift_keycode_fops;

	proc_button2 = create_proc_entry("mouse_button2_shift_keycode",
					 S_IWUSR | S_IRUGO, proc_vrtp_root);
	if( proc_button2 )
	    proc_button2->proc_fops = &vrtp_mouse_button2_shift_keycode_fops;

	proc_button3 = create_proc_entry("mouse_button3_shift_keycode",
					 S_IWUSR | S_IRUGO, proc_vrtp_root);
	if( proc_button3 )
	    proc_button3->proc_fops = &vrtp_mouse_button3_shift_keycode_fops;

	proc_nf_shift_bits = create_proc_entry("nf_shift_bits",
					 S_IWUSR | S_IRUGO, proc_vrtp_root);
	if( proc_nf_shift_bits )
	    proc_nf_shift_bits->proc_fops = &vrtp_nf_shift_bits_fops;

	proc_nf_ignore_count = create_proc_entry("nf_ignore_count",
					 S_IWUSR | S_IRUGO, proc_vrtp_root);
	if( proc_nf_ignore_count )
	    proc_nf_ignore_count->proc_fops = &vrtp_nf_ignore_count_fops;
    }
    return 0;
}
__initcall(vrtp_proc_init);
#endif /* CONFIG_PROC_FS */

static int __init vrtp_init(void)
{
    memset(&vrtp, 0, sizeof(struct vrtp));

    vrtp.tpcal = default_tpcal;
    vrtp.nf_param = default_nf_param;

    set_bit(EV_ABS, vrtp.dev.evbit);
    set_bit(ABS_X, vrtp.dev.absbit);
    set_bit(ABS_Y, vrtp.dev.absbit);
    set_bit(ABS_PRESSURE, vrtp.dev.absbit);
    vrtp.dev.absmax[ABS_X] = 799 * 4;
    vrtp.dev.absmax[ABS_Y] = 479 * 4;
    vrtp.dev.absmax[ABS_PRESSURE] = 1023;
    vrtp.dev.absfuzz[ABS_X] = 0;
    vrtp.dev.absfuzz[ABS_Y] = 0;

    set_bit(EV_KEY, vrtp.dev.evbit);
    set_bit(BTN_TOUCH, vrtp.dev.keybit);
    set_bit(BTN_LEFT, vrtp.dev.keybit);
    set_bit(BTN_MIDDLE, vrtp.dev.keybit);
    set_bit(BTN_RIGHT, vrtp.dev.keybit);

    set_bit(EV_MSC, vrtp.dev.evbit);
    set_bit(MSC_SERIAL, vrtp.dev.mscbit);

    vrtp.dev.name = "vrtp";

    input_register_device(&vrtp.dev);
    return 0;
}
module_init(vrtp_init);

static void __exit vrtp_exit(void)
{
    input_unregister_device(&vrtp.dev);
}
module_exit(vrtp_exit);
