/*
 *
 *  linux/drivers/char/tx39uart.c
 *
 *  Toshiba TMPR39xx UART module support.
 *
 *  Copyright (C) 2000  Yoichi Yuasa <yuasa@users.sourceforge.net>
 *  Copyright (C) 2001  Hirokazu FUKUI <fuku@axe-inc.co.jp>
 *
 * TTY code is based on sx.c (Specialix SX driver) by:
 *
 *   (C) 1998 - 2000  R.E.Wolff@BitWizard.nl
 *
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/errno.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/mm.h>
#include <linux/serial.h>
#include <linux/fcntl.h>
#include <linux/major.h>
#include <linux/init.h>
#include <linux/console.h>
#include <linux/generic_serial.h>
#include <linux/delay.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/tx3920.h>

#include "tx39uart.h"
#ifdef CONFIG_SHARP_TELIOS
#include "../../arch/mips/tx3920/sharp_telios.h"
#endif	

#undef SERIAL_DEBUG_OPEN
#undef SERIAL_DEBUG_CRTSCTS

/* Function prototypes */
static void tx39uart_disable_tx_interrupts (void * ptr);
static void tx39uart_enable_tx_interrupts (void * ptr);
static void tx39uart_disable_rx_interrupts (void * ptr);
static void tx39uart_enable_rx_interrupts (void * ptr);
static int  tx39uart_get_CD (void * ptr);
static void tx39uart_shutdown_port (void * ptr);
static int  tx39uart_set_real_termios (void  *ptr);
static int tx39uart_chars_in_buffer (void * ptr);
static void tx39uart_close (void  *ptr);
static void tx39uart_hungup (void  *ptr);

static struct tty_driver tx39uart_driver, tx39uart_callout_driver;

static struct tty_struct *tx39uart_table[TX39UART_NPORTS] = { NULL, };
static struct termios *tx39uart_termios[TX39UART_NPORTS];
static struct termios *tx39uart_termios_locked[TX39UART_NPORTS];

struct tx39uart_port tx39uart_ports[TX39UART_NPORTS];
int tx39uart_refcount;
int tx39uart_initialized = 0;
int tx39uart_debug = 0;
#ifdef CONFIG_PCMCIA_SERIAL
int tx39uart_enable = 0;
#else
int tx39uart_enable = 1;
#endif

static int tx_hold_reg_empty = 1;

static struct real_driver tx39uart_real_driver = {
	tx39uart_disable_tx_interrupts,
	tx39uart_enable_tx_interrupts,
	tx39uart_disable_rx_interrupts,
        tx39uart_enable_rx_interrupts,
        tx39uart_get_CD,
        tx39uart_shutdown_port,
        tx39uart_set_real_termios,
        tx39uart_chars_in_buffer,
        tx39uart_close,
        tx39uart_hungup,
        NULL
};

static inline u32 tx39uart_read(struct tx39uart_port *port, int offset)
{
	u32 val;

	val = *((volatile u32 *)(port->base + offset));
	barrier();
	return val;
}

static inline void tx39uart_write(struct tx39uart_port *port, int offset,
                                  u32 value)
{
	*((volatile u32 *)(port->base + offset)) = value;
	barrier();
}

static void tx39uart_setsignals(struct tx39uart_port *port, int dtr, int rts)
{
#ifdef CONFIG_SHARP_TELIOS
    unsigned long flags;

    if( port->line == 0 ){
	save_and_cli(flags);
	if( dtr >= 0 )
	    IODataInOut = !dtr ? (IODataInOut | TELIOS_IO_UARTA_DTR) 
			  : (IODataInOut & ~TELIOS_IO_UARTA_DTR);
	if( rts >= 0 )
	    MFIOOutput = !rts ? (MFIOOutput | TELIOS_MFIO_UARTA_RTS)
		: (MFIOOutput & ~TELIOS_MFIO_UARTA_RTS);
	restore_flags(flags);
    }
#endif
}

static int tx39uart_getsignals(struct tx39uart_port *port)
{
#ifdef CONFIG_SHARP_TELIOS
    int d;

    if( port->line == 0 ){
	d = TIOCM_DSR;
	d |= !(IODataInOut & TELIOS_IO_UARTA_DTR) ? TIOCM_DTR : 0;
	d |= !(MFIOOutput & TELIOS_MFIO_UARTA_RTS) ? TIOCM_RTS : 0;
	d |= !(MFIOInput & TELIOS_MFIO_UARTA_CTS) ? TIOCM_CTS : 0;
	return d;
    }else
	return TIOCM_DTR|TIOCM_RTS|TIOCM_DSR;
#else
    return TIOCM_DTR|TIOCM_RTS|TIOCM_DSR;
#endif
}

static inline int tx39uart_detect_change_CD(void *ptr)
{
#ifdef CONFIG_SHARP_TELIOS
    struct tx39uart_port *port = ptr;
    static int prev = -1;
    int cd, r;

    if( port->line == 0 ){
	cd = !(IODataInOut & TELIOS_IO_UARTA_CD) ? 1 : 0;
	r = (cd != prev) ? 1 : 0;
	prev = cd;
	return  r;
    }else
	return 0;
#else
    return 0;
#endif
}

static inline int tx39uart_get_CTS(void *ptr)
{
#ifdef CONFIG_SHARP_TELIOS
    struct tx39uart_port *port = ptr;

    if( port->line == 0 )
	return !(MFIOInput & TELIOS_MFIO_UARTA_CTS) ? TIOCM_CTS : 0;
    else
	return 0;
#else
	return 0;
#endif
}

static inline void tx39uart_set_RTS(void *ptr, int rts)
{
#ifdef CONFIG_SHARP_TELIOS
    struct tx39uart_port *port = ptr;
    unsigned long flags;

    if( port->line == 0 ){
	save_and_cli(flags);
	MFIOOutput = !rts ? (MFIOOutput|TELIOS_MFIO_UARTA_RTS) :
	    (MFIOOutput&~TELIOS_MFIO_UARTA_RTS);
	restore_flags(flags);
    }
#endif
}

static void tx39uart_clock(int line, int enable)
{
    u32 ctrl1, clock;

    clock = ClockControl;
    if( enable )
	clock |= (line == 0) ? CLK_EN_UART_A : CLK_EN_UART_B;
    else
	clock &= ~((line == 0) ? CLK_EN_UART_A : CLK_EN_UART_B);
    if( clock & (CLK_EN_UART_A | CLK_EN_UART_B) ){
	clock &= ~CLK_CSERDIV_MASK;
	/* UART clock is 9.216MHz (= (129.024MHz / 2) / (5+2)) under Telios */
	clock |= (5 << CLK_CSERDIV_SHIFT) | CLK_ENCSERCLK | CLK_CSERSEL;
    }else{
	unsigned long timeout=1000;

	do {
		udelay(10);
		if( line == 0 )
		    ctrl1 = UartA_Ctrl1;
		else
		    ctrl1 = UartB_Ctrl1;
	} while (--timeout && (ctrl1 & TX39UART_UARTON));
	if ((ctrl1 & TX39UART_UARTON))
		printk("tx39uart: Error UART power down control\n");
	clock &= ~CLK_ENCSERCLK;
    }
    ClockControl = clock;
#ifdef CONFIG_SHARP_TELIOS
    if( clock & (CLK_EN_UART_A | CLK_EN_UART_B) ){
	IODataInOut |= TELIOS_IO_RS232C_EN;
    }
#endif	
}

static void tx39uart_enable_port(struct tx39uart_port *port)
{
	unsigned long flags, timeout=1000;
	u32 ctrl1;

	save_and_cli(flags);
	tx39uart_clock(port->line, 1);
	restore_flags(flags);
	udelay(100);
	save_and_cli(flags);
	ctrl1 = tx39uart_read(port, TX39UART_CTRL1);
	ctrl1 |= TX39UART_ENUART;
	tx39uart_write(port, TX39UART_CTRL1, ctrl1);

	do {
		udelay(10);
		ctrl1 = tx39uart_read(port, TX39UART_CTRL1);
	} while (--timeout && !(ctrl1 & TX39UART_UARTON));
	restore_flags(flags);

	if (!(ctrl1 & TX39UART_UARTON))
		printk("tx39uart: Error UART power on control\n");
}

static void tx39uart_disable_port(struct tx39uart_port *port)
{
	unsigned long flags;
	u32 ctrl1;

	save_and_cli(flags);
	ctrl1 = tx39uart_read(port, TX39UART_CTRL1);
	ctrl1 &= ~TX39UART_ENUART;
	tx39uart_write(port, TX39UART_CTRL1, ctrl1);
	tx39uart_clock(port->line, 0);
	restore_flags(flags);
}

static void tx39uart_set_baud (struct tx39uart_port *port)
{
        u32 quot;

	switch (port->gs.baud) {
	case 300:
		quot = TX39UART_BAUD_300;
		break;
	case 600:
		quot = TX39UART_BAUD_600;
		break;
	case 1200:
		quot = TX39UART_BAUD_1200;
		break;
	case 2400:
		quot = TX39UART_BAUD_2400;
		break;
	case 4800:
		quot = TX39UART_BAUD_4800;
		break;
	case 9600:
		quot = TX39UART_BAUD_9600;
		break;
	case 19200:
		quot = TX39UART_BAUD_19200;
		break;
	case 38400:
		quot = TX39UART_BAUD_38400;
		break;
	case 57600:
		quot = TX39UART_BAUD_57600;
		break;
	case 115200:
		quot = TX39UART_BAUD_115200;
		break;
	default:
		quot = TX39UART_BAUD_9600;
#if 0
		printk (KERN_INFO "tx39uart: unsupported baud rate: %d.\n",
		        port->gs.baud);
#endif
		break;
	}

	tx39uart_write(port, TX39UART_CTRL2, quot);
}

static int tx39uart_set_real_termios(void *ptr)
{
	struct tx39uart_port *port = ptr;
	unsigned long flags, c_cflag;
	u32 ctrl1;

	c_cflag = port->gs.tty->termios->c_cflag;

	tx39uart_enable_port(port);
	ctrl1 = TX39UART_ENUART;
	switch (c_cflag & CSIZE) {
	case CS7:
	        ctrl1 |= TX39UART_BIT_7;
		break;
	case CS8:
	default:
		break;
	}
	if (c_cflag & CSTOPB)
		ctrl1 |= TX39UART_TWOSTOP;
	if (c_cflag & PARENB){
		ctrl1 |= TX39UART_ENPARITY;
		if (!(c_cflag & PARODD))
			ctrl1 |= TX39UART_EVENPARITY;
	}

	save_and_cli(flags);
	tx39uart_set_baud(port);
	tx39uart_write(port, TX39UART_CTRL1, ctrl1);
	restore_flags(flags);

	if (!(c_cflag & CRTSCTS))
		port->gs.tty->hw_stopped = 0;

	tx39uart_enable_rx_interrupts(port);

	/* Tell line discipline whether we will do input cooking */
	if(I_OTHER(port->gs.tty)) {
		clear_bit(TTY_HW_COOK_IN, &port->gs.tty->flags);
	} else {
		set_bit(TTY_HW_COOK_IN, &port->gs.tty->flags);
	}

/* Tell line discipline whether we will do output cooking.
 * If OPOST is set and no other output flags are set then we can do output
 * processing.  Even if only *one* other flag in the O_OTHER group is set
 * we do cooking in software.
 */
	if(O_OPOST(port->gs.tty) && !O_OTHER(port->gs.tty)) {
		set_bit(TTY_HW_COOK_OUT, &port->gs.tty->flags);
	} else {
		clear_bit(TTY_HW_COOK_OUT, &port->gs.tty->flags);
	}

        return 0;
}

/* ********************************************************************** *
 *                   the interrupt related routines                       *
 * ********************************************************************** */

void tx39uart_transmit_chars(struct tx39uart_port *port)
{
	u32 c;

	if (port->gs.xmit_cnt <= 0 && !port->x_char) {
		port->gs.flags &= ~GS_TX_INTEN;
		tx39uart_disable_tx_interrupts(port);
		printk("tx39uart_transmit_chars: ERROR: transmit 0bytes\n");
		return;
	}

	if (port->x_char) {
		c = port->x_char;
		port->x_char = 0;
	}
	else
		c = port->gs.xmit_buf[port->gs.xmit_tail++] & TX39UART_TX_DATA;

	IntClear2 = (port->line == 0) ? INT2_UARTATXINT : INT2_UARTBTXINT;

	tx_hold_reg_empty = 0;
	tx39uart_write(port, TX39UART_TX, c);
	port->gs.xmit_tail &= (SERIAL_XMIT_SIZE - 1);
	port->gs.xmit_cnt--;

	if (port->gs.xmit_cnt <= 0) {
		port->gs.flags &= ~GS_TX_INTEN;
		tx39uart_disable_tx_interrupts(port);
	}

	if (port->gs.xmit_cnt <= port->gs.wakeup_chars) {
		if ((port->gs.tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
		    port->gs.tty->ldisc.write_wakeup)
			(port->gs.tty->ldisc.write_wakeup)(port->gs.tty);
		wake_up_interruptible(&port->gs.tty->write_wait);
        }
}

static inline void tx39uart_receive_chars(struct tx39uart_port *port)
{
	struct tty_struct *tty;

        tty = port->gs.tty;

	/* Don't copy more bytes than there is room for in the buffer */
	if (tty->flip.count + 1 > TTY_FLIPBUF_SIZE)
		return;

	tty->flip.char_buf_ptr[0] = tx39uart_read(port, TX39UART_RX);

	memset(tty->flip.flag_buf_ptr, TTY_NORMAL, 1);

	/* Update the kernel buffer end */
	tty->flip.count++;
	tty->flip.char_buf_ptr++;
	tty->flip.flag_buf_ptr++;

	/* Tell the rest of the system the news. New characters! */
	tty_flip_buffer_push(tty);
}

static inline void tx39uart_check_modem_status(struct tx39uart_port *port)
{
    unsigned long c_cflag, cts_flag;

    c_cflag = port->gs.tty->termios->c_cflag;

    if( !(c_cflag & CLOCAL) && tx39uart_detect_change_CD(port)){
	if( tx39uart_get_CD(port) )
	    wake_up_interruptible(&port->gs.open_wait);
	else if (!((port->gs.flags & ASYNC_CALLOUT_ACTIVE) &&
		   (port->gs.flags & ASYNC_CALLOUT_NOHUP))) {
	    if (port->gs.tty)
		tty_hangup(port->gs.tty);
	}
    }

    if (c_cflag & CRTSCTS) {
	cts_flag = tx39uart_get_CTS(port);
	if (port->gs.tty->hw_stopped) {
	    if (cts_flag) {
#ifdef SERIAL_DEBUG_CRTSCTS
		printk("tx39uart: hw_stopped=0\n");
#endif
		port->gs.tty->hw_stopped = 0;
		if ((port->gs.tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
		    port->gs.tty->ldisc.write_wakeup)
		    (port->gs.tty->ldisc.write_wakeup)(port->gs.tty);
		wake_up_interruptible(&port->gs.tty->write_wait);
		return;
	    }
	} else {
	    if (!cts_flag){
#ifdef SERIAL_DEBUG_CRTSCTS
		printk("tx39uart: hw_stopped=1\n");
#endif
		port->gs.tty->hw_stopped = 1;
	    }
	}
    }
}

extern volatile u32 tx39uart_status;

static void tx39uart_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct tx39uart_port *port = (struct tx39uart_port *)dev_id;
	u32 status;

	if (!(port->gs.flags & GS_ACTIVE))
		return;

	status = tx39uart_status & INT2_UART_INT(port->line);
	IntClear2 = status;
	tx39uart_status &= ~status;

	if (status & INT2_UARTRXINT)
		tx39uart_receive_chars(port);
	tx39uart_check_modem_status(port);
	if (status & INT2_UARTTXINT){
	    tx_hold_reg_empty = 1;
	    if (port->gs.xmit_cnt > 0 || port->x_char) {
		tx39uart_transmit_chars(port);
	    }else{
		port->gs.flags &= ~GS_TX_INTEN;
		tx39uart_disable_tx_interrupts(port);
	    }
	}
	if (status & (INT2_UARTRXOVERRUNINT | INT2_UARTFRAMEERRINT |
	              INT2_UARTPARITYERRINT | INT2_UARTBREAKINT))
		printk("tx39uart: Rx error\n");

	if (status & INT2_UARTTXOVERRUNINT)
		printk("tx39uart: Tx overrun\n");
}

/* ********************************************************************** *
 *                Here are the routines that actually                     *
 *              interface with the generic_serial driver                  *
 * ********************************************************************** */

static void tx39uart_disable_tx_interrupts(void *ptr)
{
	struct tx39uart_port *port = ptr;
	unsigned long flags;
	u32 ien;

	save_and_cli(flags);
	ien = IntEnable2;
	ien &= ~INT2_UART_TXINT(port->line);
	IntEnable2 = ien;
	/* tx_hold_reg_empty = ???; */
	restore_flags(flags);
}

static void tx39uart_enable_tx_interrupts(void *ptr)
{
	struct tx39uart_port *port = ptr;
	unsigned long flags;
	u32 ien;

	save_and_cli(flags);
	if (tx_hold_reg_empty ||
	    (tx39uart_read(port, TX39UART_CTRL1) & TX39UART_EMPTY)){
	    tx39uart_transmit_chars(port);
	}
	ien = IntEnable2;
	ien |= INT2_UART_TXINT(port->line);
	IntEnable2 = ien;
	restore_flags(flags);
}

static void tx39uart_disable_rx_interrupts(void *ptr)
{
	struct tx39uart_port *port = ptr;
	unsigned long flags;
	u32 ien;

	save_and_cli(flags);
	ien = IntEnable2;
	ien &= ~INT2_UART_RXINT(port->line);
	IntEnable2 = ien;
	restore_flags(flags);
}

static void tx39uart_enable_rx_interrupts(void *ptr)
{
	struct tx39uart_port *port = ptr;
	unsigned long flags;
	u32 ien;

	save_and_cli(flags);
	if (tx39uart_read(port, TX39UART_CTRL1) & TX39UART_RXHOLDFULL)
		tx39uart_receive_chars(port);
	ien = IntEnable2;
	ien |= INT2_UART_RXINT(port->line);
	IntEnable2 = ien;
	restore_flags(flags);
}

static int tx39uart_get_CD(void *ptr)
{
#ifdef CONFIG_SHARP_TELIOS
    struct tx39uart_port *port = ptr;

    if( port->line == 0 )
	return !(IODataInOut & TELIOS_IO_UARTA_CD) ? 1 : 0;
    else
	return 1;
#else
    /* If you have signal for CD (Carrier Detect), please change here. */
    return 1;
#endif
}

static int tx39uart_chars_in_buffer(void *ptr)
{
	struct tx39uart_port *port = ptr;

	return (tx39uart_read(port, TX39UART_CTRL1) & TX39UART_EMPTY) ? 0 : 1;
}

static void tx39uart_shutdown_port(void *ptr)
{
	struct tx39uart_port *port = ptr;
/*
	unsigned long flags;
	u32 status, val;
*/

	port->gs.flags &= ~GS_ACTIVE;
	if (port->gs.tty && port->gs.tty->termios->c_cflag & HUPCL) {
		tx39uart_setsignals(port, 0, 0);
/*
		save_and_cli(flags);
		val = tx39uart_read(port, TX39UART_CTRL1);
		val &= ~TX39UART_ENUART;
		tx39uart_write(port, TX39UART_CTRL1, val);
		do {
			status = tx39uart_read(port, TX39UART_CTRL1);
		} while (status & TX39UART_UARTON);
		restore_flags(flags);
*/
	}
	tx39uart_disable_port(port);
}

/* ********************************************************************** *
 *                Here are the routines that actually                     *
 *               interface with the rest of the system                    *
 * ********************************************************************** */

static int tx39uart_open(struct tty_struct *tty, struct file *filp)
{
	struct tx39uart_port *port;
	int retval, line;

	if (!tx39uart_initialized) {
		return -EIO;
	}

	line = MINOR(tty->device) - tty->driver.minor_start;
	if ((line < 0) || (line >= TX39UART_NPORTS))
		return -ENODEV;

#ifdef SERIAL_DEBUG_OPEN
	printk("tx39uart: open line=%d\n", line);
#endif

	port = &tx39uart_ports[line];

	tty->driver_data = port;
	port->gs.tty = tty;
	port->gs.count++;

	/*
	 * Start up UART port
	 */
	retval = gs_init_port(&port->gs);
	if (retval) {
		port->gs.count--;
		return retval;
	}

	port->gs.flags |= GS_ACTIVE;
	tx39uart_setsignals(port, 1, 1);

	if (port->gs.count == 1) {
		MOD_INC_USE_COUNT;
	}

	retval = block_til_ready(port, filp);
	if (retval) {
		port->gs.count--;
		if (port->gs.count <= 0);
			MOD_DEC_USE_COUNT;
		return retval;
	}

	if ((port->gs.count == 1) && (port->gs.flags & ASYNC_SPLIT_TERMIOS)) {
		if (tty->driver.subtype == SERIAL_TYPE_NORMAL)
			*tty->termios = port->gs.normal_termios;
		else
			*tty->termios = port->gs.callout_termios;
		tx39uart_set_real_termios (port);
	}

	tx39uart_enable_rx_interrupts(port);

	port->gs.session = current->session;
	port->gs.pgrp = current->pgrp;

        return 0;
}

static void tx39uart_hungup(void *ptr)
{
	MOD_DEC_USE_COUNT;
}

static void tx39uart_close(void *ptr)
{
#ifdef SERIAL_DEBUG_OPEN
	struct tx39uart_port *port = ptr;
	printk("tx39uart: close line=%d\n", port->line);
#endif
	MOD_DEC_USE_COUNT;
}

static int tx39uart_ioctl(struct tty_struct * tty, struct file * filp,
                          unsigned int cmd, unsigned long arg)
{
	int rc;
	struct tx39uart_port *port = tty->driver_data;
	int ival;

	rc = 0;
	switch (cmd) {
	case TIOCGSOFTCAR:
		rc = put_user(((tty->termios->c_cflag & CLOCAL) ? 1 : 0),
		              (unsigned int *)arg);
		break;
	case TIOCSSOFTCAR:
		if ((rc = verify_area(VERIFY_READ, (void *)arg,
		                      sizeof(int))) == 0) {
			get_user(ival, (unsigned int *)arg);
			tty->termios->c_cflag =
			        (tty->termios->c_cflag & ~CLOCAL) |
			        (ival ? CLOCAL : 0);
		}
		break;
	case TIOCGSERIAL:
		if ((rc = verify_area(VERIFY_WRITE, (void *)arg,
		                      sizeof(struct serial_struct))) == 0)
		        gs_getserial(&port->gs, (struct serial_struct *)arg);
		break;
	case TIOCSSERIAL:
		if ((rc = verify_area(VERIFY_READ, (void *)arg,
		                      sizeof(struct serial_struct))) == 0)
			rc = gs_setserial(&port->gs, (struct serial_struct *)arg);
		break;
	case TIOCMGET:
		if ((rc = verify_area(VERIFY_WRITE, (void *)arg,
		                      sizeof(unsigned int))) == 0) {
			ival = tx39uart_getsignals(port);
			put_user(ival, (unsigned int *)arg);
		}
		break;
	case TIOCMBIS:
		if ((rc = verify_area(VERIFY_READ, (void *)arg,
		                      sizeof(unsigned int))) == 0) {
			get_user(ival, (unsigned int *)arg);
                        tx39uart_setsignals(port, ((ival & TIOCM_DTR) ? 1 : -1),
			                          ((ival & TIOCM_RTS) ? 1 : -1));
		}
		break;
	case TIOCMBIC:
		if ((rc = verify_area(VERIFY_READ, (void *)arg,
		                      sizeof(unsigned int))) == 0) {
			get_user(ival, (unsigned int *)arg);
			tx39uart_setsignals(port, ((ival & TIOCM_DTR) ? 0 : -1),
			                          ((ival & TIOCM_RTS) ? 0 : -1));
		}
		break;
	case TIOCMSET:
		if ((rc = verify_area(VERIFY_READ, (void *)arg,
		                      sizeof(unsigned int))) == 0) {
			get_user(ival, (unsigned int *)arg);
                        tx39uart_setsignals(port, ((ival & TIOCM_DTR) ? 1 : 0),
			                          ((ival & TIOCM_RTS) ? 1 : 0));
		}
		break;
	default:
		rc = -ENOIOCTLCMD;
		break;
	}

	return rc;
}

static void tx39uart_send_xchar(struct tty_struct *tty, char ch)
{
	struct tx39uart_port *port = (struct tx39uart_port *)tty->driver_data;

	if (ch) {
		port->x_char = ch;
		tx39uart_enable_tx_interrupts(tty);
	}
}

static void tx39uart_throttle(struct tty_struct *tty)
{
	struct tx39uart_port *port = (struct tx39uart_port *)tty->driver_data;

	/* If the port is using any type of input flow
	 * control then throttle the port.
	 */
	if (I_IXOFF(tty))
		tx39uart_send_xchar(tty, STOP_CHAR(tty));
	if (tty->termios->c_cflag & CRTSCTS)
		port->gs.flags |= TX39UART_RX_THROTTLE;

	tx39uart_set_RTS(port, (port->gs.flags & TX39UART_RX_THROTTLE)?0:1);
#ifdef SERIAL_DEBUG_CRTSCTS
	printk("tx39uart: RTS=%d\n", (port->gs.flags & TX39UART_RX_THROTTLE)?0:1);
#endif
}

static void tx39uart_unthrottle (struct tty_struct *tty)
{
        struct tx39uart_port *port = (struct tx39uart_port *)tty->driver_data;

	/* Always unthrottle even if flow control is not enabled on
	 * this port in case we disabled flow control while the port
	 * was throttled
	 */
	if(I_IXOFF(tty)) {
		if (port->x_char)
			port->x_char = 0;
		else
			tx39uart_send_xchar(tty, START_CHAR(tty));
	}
	port->gs.flags &= ~TX39UART_RX_THROTTLE;
	tx39uart_set_RTS(port, 1);
#ifdef SERIAL_DEBUG_CRTSCTS
	printk("tx39uart: RTS=%d\n", 1);
#endif
	return;
}

/* ********************************************************************** *
 *                    Here are the initialization routines.               *
 * ********************************************************************** */

static void __init tx39uart_init_portstructs(void)
{
	struct tx39uart_port *port = &tx39uart_ports[0];
	int i;

	for (i = 0; i < TX39UART_NPORTS; i++, port++) {
		port->gs.callout_termios = tty_std_termios;
		port->gs.normal_termios = tty_std_termios;
		port->gs.magic = TX39UART_MAGIC;
		port->gs.close_delay = HZ/2;
		port->gs.closing_wait = 30 * HZ;
		port->gs.rd = &tx39uart_real_driver;
		init_waitqueue_head(&port->gs.open_wait);
		init_waitqueue_head(&port->gs.close_wait);

		port->line = i;
		port->base = (i == 0) ? TX39UART_UARTA_BASE : TX39UART_UARTB_BASE;
		port->irq = (i == 0) ? TX39UART_UARTA_IRQ : TX39UART_UARTB_IRQ;
	}
}

static int __init tx39uart_init_drivers(void)
{
	int error;

	memset(&tx39uart_driver, 0, sizeof(tx39uart_driver));
	tx39uart_driver.magic = TTY_DRIVER_MAGIC;
	tx39uart_driver.driver_name = "serial";
	tx39uart_driver.name = "ttyS";
	tx39uart_driver.major = TTY_MAJOR;
	tx39uart_driver.minor_start = TX39UART_MINOR_START;
	tx39uart_driver.num = TX39UART_NPORTS;
	tx39uart_driver.type = TTY_DRIVER_TYPE_SERIAL;
	tx39uart_driver.subtype = SERIAL_TYPE_NORMAL;
	tx39uart_driver.init_termios = tty_std_termios;
	tx39uart_driver.init_termios.c_cflag =
		B9600 | CS8 | CREAD | HUPCL | CLOCAL;
	tx39uart_driver.flags = TTY_DRIVER_REAL_RAW;
	tx39uart_driver.refcount = &tx39uart_refcount;
	tx39uart_driver.table = tx39uart_table;
	tx39uart_driver.termios = &tx39uart_termios[0];
	tx39uart_driver.termios_locked = &tx39uart_termios_locked[0];

	tx39uart_driver.open = tx39uart_open;
	tx39uart_driver.close = gs_close;
	tx39uart_driver.write = gs_write;
	tx39uart_driver.put_char = gs_put_char;
	tx39uart_driver.flush_chars = gs_flush_chars;
	tx39uart_driver.write_room = gs_write_room;
	tx39uart_driver.chars_in_buffer = gs_chars_in_buffer;
	tx39uart_driver.flush_buffer = gs_flush_buffer;
	tx39uart_driver.ioctl = tx39uart_ioctl;
	tx39uart_driver.throttle = tx39uart_throttle;
	tx39uart_driver.unthrottle = tx39uart_unthrottle;
	tx39uart_driver.set_termios = gs_set_termios;
	tx39uart_driver.stop = gs_stop;
	tx39uart_driver.start = gs_start;
	tx39uart_driver.hangup = gs_hangup;

	tx39uart_callout_driver = tx39uart_driver;
	tx39uart_callout_driver.name = "cua";
	tx39uart_callout_driver.major = TTYAUX_MAJOR;
	tx39uart_callout_driver.subtype = SERIAL_TYPE_CALLOUT;

	if ((error = tty_register_driver(&tx39uart_driver))) {
		printk(KERN_ERR "tx39uart: Couldn't register TX39xx UART driver, error = %d\n", error);
		return -ENODEV;
	}
	if ((error = tty_register_driver(&tx39uart_callout_driver))) {
		tty_unregister_driver(&tx39uart_driver);
		printk(KERN_ERR "tx39uart: Couldn't register TX39xx UART callout driver, error = %d\n", error);
		return -ENODEV;
	}

	return 0;
}

int __init tx39uart_init(void)
{
	unsigned long flags;
	struct tx39uart_port *port;
	int retval, i;

	if( !tx39uart_enable ){
	    printk(KERN_INFO "tx39uart: disabled by serial driver.\n");
	    return 0;
	}

	tx39uart_init_portstructs();
	retval = tx39uart_init_drivers();
	if (retval) {
		return retval;
	}

#if 0
#ifdef CONFIG_SHARP_TELIOS
	save_and_cli(flags);
	IODataInOut = IODataInOut & ~TELIOS_IO_UARTA_DTR; 
	restore_flags(flags);
#endif
#endif

	port = &tx39uart_ports[0];
	for (i = 0; i < TX39UART_NPORTS; i++, port++) {
		save_and_cli(flags);
		IntEnable2 &= ~INT2_UART_INT(port->line);
		restore_flags(flags);

		if (request_irq(port->irq, tx39uart_interrupt, SA_INTERRUPT,
		                "serial", port)) {
			printk(KERN_ERR "tx39uart: Cannot allocate irq=%d\n",
			       port->irq);
			return -ENODEV;
		}
		printk(KERN_INFO "ttyS%02d at 0x%08x (irq = %d) is a %s\n",
		       port->line, port->base, port->irq,
		       ((port->line == 0) ? "UARTA" : "UARTB"));
	}

	tx39uart_initialized++;

	return 0;
}

void tx39uart_fini(void)
{
	struct tx39uart_port *port = &tx39uart_ports[0];
        int i;

	if( !tx39uart_enable )
	    return;

	for (i = 0; i < TX39UART_NPORTS; i++, port++)
		free_irq(port->irq, port);

	tty_unregister_driver(&tx39uart_driver);
	tty_unregister_driver(&tx39uart_callout_driver);
}

module_init(tx39uart_init);
module_exit(tx39uart_fini);

void serial_outc(char c)
{
	unsigned long flags, timeout = 1000000;
	u32 status;

	save_and_cli(flags);
	do {
		status = UartA_Ctrl1;
	} while (--timeout && !(status & TX39UART_EMPTY));

	UartA_Data = c;

	do {
		status = UartA_Ctrl1;
	} while (--timeout && (status & TX39UART_EMPTY));
	restore_flags(flags);
}

static int __init setup_tx39uart(char *str)
{
    tx39uart_enable = 1;
    return 1;
}

__setup("tx39uart=", setup_tx39uart);

#ifdef CONFIG_SERIAL_CONSOLE
/*
 * ------------------------------------------------------------
 * Serial console driver for Toshiba TMPR39xx UART module
 * ------------------------------------------------------------
 */

static void serial_console_write(struct console *co, const char *s,
                                 unsigned count)
{
	unsigned i;

	/*
	 *      Now, do each character
	 */
	for (i = 0; i < count; i++, s++) {
		/*
		 *      Send the character out.
		 *      If a LF, also do CR...
		 */
		if (*s == '\n')
			serial_outc('\r');
		serial_outc(*s);
	}
}

static int serial_console_wait_key(struct console *co)
{
	unsigned long flags;
	u32 status;
	int c;

	save_and_cli(flags);
	do {
		status = UartA_Ctrl1;
	} while (!(status & TX39UART_RXHOLDFULL));

	c = (int)(UartA_Data & TX39UART_RX_DATA);
	restore_flags(flags);

	return c;
}

static kdev_t serial_console_device(struct console *c)
{
	return MKDEV(TTY_MAJOR, 64 + c->index);
}

static int __init serial_console_setup(struct console *co, char *options)
{
	unsigned long flags;
	u32 ctrl1 = TX39UART_ENUART;
	u32 ctrl2 = 0;
	int baud = 9600;
	int bits = 8;
	int parity = 'n';
	int cflag = CREAD | HUPCL | CLOCAL;
	char    *s;

	if (options) {
		baud = simple_strtoul(options, NULL, 10);
		s = options;
		while(*s >= '0' && *s <= '9')
			s++;
		if (*s) parity = *s++;
		if (*s) bits   = *s - '0';
	}

	/*
	 * Now construct a cflag setting.
	 */
	switch(baud) {
	case 300:
		cflag |= B300;
		ctrl2 = TX39UART_BAUD_300;
		break;
	case 600:
		cflag |= B600;
		ctrl2 = TX39UART_BAUD_600;
		break;
	case 1200:
		cflag |= B1200;
		ctrl2 = TX39UART_BAUD_1200;
		break;
	case 2400:
		cflag |= B2400;
		ctrl2 = TX39UART_BAUD_2400;
		break;
	case 4800:
		cflag |= B4800;
		ctrl2 = TX39UART_BAUD_4800;
		break;
	case 9600:
	default:
		cflag |= B9600;
		ctrl2 = TX39UART_BAUD_9600;
		break;
	case 19200:
		cflag |= B19200;
		ctrl2 = TX39UART_BAUD_19200;
		break;
	case 38400:
		cflag |= B38400;
		ctrl2 = TX39UART_BAUD_38400;
		break;
	case 57600:
		cflag |= B57600;
		ctrl2 = TX39UART_BAUD_57600;
		break;
	case 115200:
		cflag |= B115200;
		ctrl2 = TX39UART_BAUD_115200;
		break;
	}
	switch(bits) {
	case 7:
		cflag |= CS7;
		ctrl1 |= TX39UART_BIT_7;
		break;
	case 8:
	default:
		cflag |= CS8;
		break;
	}
	switch(parity) {
	case 'o': case 'O':
		cflag |= PARODD;
		ctrl1 |= TX39UART_ENPARITY;
		break;
	case 'e': case 'E':
		cflag |= PARENB;
		ctrl1 |= TX39UART_ENPARITY | TX39UART_EVENPARITY;
		break;
	}
	co->cflag = cflag;

	/*
	 * setup UART module A
	 */
	save_and_cli(flags);
	tx39uart_clock(0, 1);
        UartA_Ctrl2 = ctrl2;
        UartA_Ctrl1 = ctrl1;
	restore_flags(flags);

	return 0;
}

static struct console sercons = {
	"ttyS",
	serial_console_write,
	NULL,
	serial_console_device,
	serial_console_wait_key,
	NULL,
	serial_console_setup,
	CON_PRINTBUFFER,
	-1,
	0,
	NULL
};

void __init tx39uart_serial_console_init(void)
{
	register_console(&sercons);
}
#endif
