/*
 * Toshiba PLUM2 PCMCIA Controller support.
 *
 * Copyright (C) 2000 Yoichi Yuasa <yuasa@users.sorceforge.net>
 * All rights reserved.
 *
 * This code is based on i82365.c (Intel 82365 driver) by:
 *
 *   Copyright (C) 1999 David A. Hinds
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/bus_ops.h>
#include <pcmcia/ss.h>
#include <pcmcia/cs.h>

#include <asm/io.h>
#include <asm/kseg2.h>
#include <asm/plum2.h>
#include "plum2pcc.h"

#ifdef PCMCIA_DEBUG
static int pcmcia_debug = PCMCIA_DEBUG;
#define DEBUG(n, args...)	if (pcmcia_debug > (n)) printk(KERN_INFO args)
#else
#define DEBUG(n, args...)
#endif

static u_short cycle_time = 120;

typedef struct socket_info_t {
	u8	psock;
	u32	reg_base;
	void	(*handler)(void *info, u_int events);
	void	*info;
} socket_info_t;

static int sockets = 0;
static socket_info_t socket_table[PLUM2_PCMCIA_SOCKETS] = {
	{0, },
};

static socket_cap_t plum2pcc_cap[PLUM2_PCMCIA_SOCKETS] = {
	{SS_CAP_PAGE_REGS | SS_CAP_MEM_ALIGN | SS_CAP_PCCARD,
	 0xffff, 0x1000, 0, 0},
	{SS_CAP_PAGE_REGS | SS_CAP_MEM_ALIGN | SS_CAP_PCCARD,
	 0xffff, 0x1000, 0, 0}
};

typedef struct plum2pcc_map_t {
	u_long base;
	u_long size;
	u_long offset;
} plum2pcc_map_t;

static plum2pcc_map_t plum2pcc_io_map[] = {
	{PLUM2_PCC_IO1_BASE, PLUM2_PCC_IO1_SIZE, 0},
	{PLUM2_PCC_IO2_BASE, PLUM2_PCC_IO2_SIZE, 0}
};

static plum2pcc_map_t plum2pcc_mem_map[] = {
	{PLUM2_PCC_MEM1_BASE, PLUM2_PCC_MEM1_SIZE, 0},
	{PLUM2_PCC_MEM2_BASE, PLUM2_PCC_MEM2_SIZE, 0},
	{PLUM2_PCC_MEM3_BASE, PLUM2_PCC_MEM3_SIZE, 0},
	{PLUM2_PCC_MEM4_BASE, PLUM2_PCC_MEM4_SIZE, 0}
};

static u32 plum2pcc_io_base;
static u32 plum2pcc_mem_base;

static spinlock_t isa_lock = SPIN_LOCK_UNLOCKED;
#define ISA_LOCK(n, f) spin_lock_irqsave(&isa_lock, f)
#define ISA_UNLOCK(n, f) spin_unlock_irqrestore(&isa_lock, f)

/*====================================================================*/

static inline u32 plum2pcc_reg_read(u8 socket, u32 offset)
{
	u32 val;
	val = *((volatile u32*)(socket_table[socket].reg_base + offset));
	barrier();
	return val;
}

static inline void plum2pcc_reg_write(u8 socket, u32 offset, u32 val)
{
	*((volatile u32*)(socket_table[socket].reg_base + offset)) = val;
	barrier();
}

static inline void plum2pcc_reg_set(u8 socket, u32 offset, u32 mask)
{
	u32 val;

	val = plum2pcc_reg_read(socket, offset);
	val |= mask;
	plum2pcc_reg_write(socket, offset, val);
	DEBUG(10, "plum2pcc: socket=%d offset=0x%08x value=0x%08x\n",
	       socket, offset, val);
}

static inline void plum2pcc_reg_clear(u32 socket, u32 offset, u32 mask)
{
	u32 val;

	val = plum2pcc_reg_read(socket, offset);
	val &= ~mask;
	plum2pcc_reg_write(socket, offset, val);
	DEBUG(10, "plum2pcc: socket=%d offset=0x%08x value=0x%08x\n",
	       socket, offset, val);
}

/*====================================================================*/

void plum2pcc_pccard_startup(void)
{
	*((volatile u32*)(PLUM2_PCC1REG_BASE + PLUM2_PCC_CARDPOWER)) =
	                                               PLUM2_PCC_CARD_STARTUP;
}

void plum2pcc_pccard_shutdown(void)
{
	*((volatile u32*)(PLUM2_PCC1REG_BASE + PLUM2_PCC_WINENABLE)) = 0;
	*((volatile u32*)(PLUM2_PCC2REG_BASE + PLUM2_PCC_WINENABLE)) = 0;
	*((volatile u32*)(PLUM2_PCC1REG_BASE + PLUM2_PCC_CARDPOWER)) =
	                                               PLUM2_PCC_CARD_SHUTDOWN;
}

/*====================================================================*/

#ifdef CONFIG_VIRTUAL_BUS

static u32 plum2pcc_bus_in(void *bus, u32 port, s32 sz)
{
	switch (sz) {
	case 0:
		return inb(plum2pcc_io_base + port);
	case 1:
		return inw(plum2pcc_io_base + port);
	case 2:
		return inl(plum2pcc_io_base + port);
	default:
		return 0;
	}
}

static void plum2pcc_bus_ins(void *bus, u32 port, void *buf, u32 count, s32 sz)
{
	switch (sz) {
	case 0:
		insb((plum2pcc_io_base + port), buf, count);
		break;
	case 1:
		insw((plum2pcc_io_base + port), buf, count);
		break;
	case 2:
		insl((plum2pcc_io_base + port), buf, count);
		break;
	}
}

static void plum2pcc_bus_out(void *bus, u32 val, u32 port, s32 sz)
{
	switch (sz) {
	case 0:
		outb(val, (plum2pcc_io_base  + port));
		break;
	case 1:
		outw(val, (plum2pcc_io_base + port));
		break;
	case 2:
		outl(val, (plum2pcc_io_base + port));
		break;
	}
}

static void plum2pcc_bus_outs(void *bus, u32 port, void *buf, u32 count, s32 sz)
{
	switch (sz) {
	case 0:
		outsb((plum2pcc_io_base + port), buf, count);
		break;
	case 1:
		outsw((plum2pcc_io_base + port), buf, count);
		break;
	case 2:
		outsl((plum2pcc_io_base + port), buf, count);
		break;
	}
}

static void *plum2pcc_bus_ioremap(void *bus, u_long ofs, u_long sz)
{
	return (void *)(plum2pcc_mem_base + (ofs & 0x03ffffff));
}

static void plum2pcc_bus_iounmap(void *bus, void *addr)
{
}

static u32 plum2pcc_bus_read(void *bus, void *addr, s32 sz)
{
	switch (sz) {
	case 0:
		return (*(volatile u8 *)(addr));
	case 1:
		return (*(volatile u16 *)(addr));
	case 2:
		return (*(volatile u32 *)(addr));
	default:
		return 0;
	}
}

static void plum2pcc_bus_write(void *bus, u32 val, void *addr, s32 sz)
{
	switch (sz) {
	case 0:
		(*(volatile u8 *)(addr)) = val;
		break;
	case 1:
		(*(volatile u16 *)(addr)) = val;
		break;
	case 2:
		(*(volatile u32 *)(addr)) = val;
		break;
	}
}

static void plum2pcc_bus_copy_from(void *bus, void *d, void *s, u32 count)
{
	memcpy(d, (void *)s, count);
}

static void plum2pcc_bus_copy_to(void *bus, void *d, void *s, u32 count)
{
	memcpy((void *)d, s, count);
}

static int plum2pcc_bus_request_irq(void *bus, u_int irq,
                                    void (*handler)(int, void *,
                                                    struct pt_regs *),
                                    u_long flags, const char *device,
                                    void *dev_id)
{
	return request_irq(irq, handler, flags, device, dev_id);
}

static void plum2pcc_bus_free_irq(void *bus, u_int irq, void *dev_id)
{
	free_irq(irq, dev_id);
}

static bus_operations plum2pcc_bus_operations = {
	NULL,
	plum2pcc_bus_in,
	plum2pcc_bus_ins,
	plum2pcc_bus_out,
	plum2pcc_bus_outs,
	plum2pcc_bus_ioremap,
	plum2pcc_bus_iounmap,
	plum2pcc_bus_read,
	plum2pcc_bus_write,
	plum2pcc_bus_copy_from,
	plum2pcc_bus_copy_to,
	plum2pcc_bus_request_irq,
	plum2pcc_bus_free_irq
};

#endif /* CONFIG_VIRTUAL_BUS */

/*====================================================================*/

static void plum2pcc_interrupt(int irq, void *dev, struct pt_regs *regs)
{
    unsigned long flags;
    u_int events;
    u32 csc;
    u8 i;

    if (irq != PLUM2_CSC_IRQ)
	return;

    for (i = 0; i < sockets; i++) {
	events = 0;
	ISA_LOCK(i, flags);
	csc = plum2pcc_reg_read(i, PLUM2_PCC_CSC);
	if ((csc == 0) || (!socket_table[i].handler) ||
	    (plum2pcc_reg_read(i, PLUM2_PCC_ID) & 0x70)) {
	    ISA_UNLOCK(i, flags);
	    continue;
	}
	if (csc & PLUM2_PCC_DETECT) events |= SS_DETECT;
	if (plum2pcc_reg_read(i, PLUM2_PCC_INTCTRL) & PLUM2_PCC_IOCARD)
	    if (csc & PLUM2_PCC_STSCHG) {
		events |= SS_STSCHG;
	    }
	else {
	    if (csc & PLUM2_PCC_BATTDEAD) events |= SS_BATDEAD;
	    if (csc & PLUM2_PCC_BATTWARN) events |= SS_BATWARN;
	    if (csc & PLUM2_PCC_CSCREADY) events |= SS_READY;
	}
	ISA_UNLOCK(i, flags);
	if (events)
	    socket_table[i].handler(socket_table[i].info, events);
    }
}

/*====================================================================*/

static int plum2pcc_get_status(unsigned int socket, u_int *value)
{
    unsigned long flags;
    u32 status;

    if (socket >= PLUM2_PCMCIA_SOCKETS) {
	*value = 0;
	return -EINVAL;
    }

    ISA_LOCK(socket, flags);

    status = plum2pcc_reg_read(socket, PLUM2_PCC_STATUS);
    *value = 0;
    if (status & PLUM2_PCC_CD) *value |= SS_DETECT;
    if (plum2pcc_reg_read(socket, PLUM2_PCC_INTCTRL) == PLUM2_PCC_IOCARD)
	if (!(status & PLUM2_PCC_STSCHG)) *value |= SS_STSCHG;
    else {
	if (!(status & PLUM2_PCC_BATTDEAD)) *value |= SS_BATDEAD;
	if (!(status & PLUM2_PCC_BATTWARN)) *value |= SS_BATWARN;
    }
    if (status & PLUM2_PCC_WRITEPROTECT) *value |= SS_WRPROT;
    if (status & PLUM2_PCC_READY) *value |= SS_READY;
    if (status & PLUM2_PCC_POWERACTIVE) *value |= SS_POWERON;

    status = plum2pcc_reg_read(socket, PLUM2_PCC_ADDGENCTRL);
    *value |= (status & PLUM2_PCC_VSENCE1) ? 0 : SS_3VCARD;
    *value |= (status & PLUM2_PCC_VSENCE2) ? 0 : SS_XVCARD;

    DEBUG(2, "plum2pcc: get_status socket %d 0x%08x\n", socket, *value);

    ISA_UNLOCK(socket, flags);

    return 0;
}

static int plum2pcc_get_socket(unsigned int socket, socket_state_t *state)
{
    unsigned long flags;
    u32 val, vcc, vpp;

    if (socket >= PLUM2_PCMCIA_SOCKETS)
	return -EINVAL;

    ISA_LOCK(socket, flags);

    /* Power Control */
    val = plum2pcc_reg_read(socket, PLUM2_PCC_POWERCTRL);
    state->flags = 0;
    if (val & PLUM2_PCC_AUTOPOWER) state->flags |= SS_PWR_AUTO;
    if (val & PLUM2_PCC_OUTPUTENABLE) state->flags |= SS_OUTPUT_ENA;

    if (val & PLUM2_PCC_CARDPOWENABLE) {
	vcc = val & PLUM2_PCC_VCC_MASK;
	vpp = val & PLUM2_PCC_VPP1_MASK;
	state->Vcc = state->Vpp = 0;
	if (vcc == PLUM2_PCC_VCC_5V) state->Vcc = 50;
	if (vcc == PLUM2_PCC_VCC_3V) state->Vcc = 33;
	if (vpp == PLUM2_PCC_VPP1_VCC) state->Vpp = state->Vcc;
    }

    /* Interrupt and General Control */
    val = plum2pcc_reg_read(socket, PLUM2_PCC_INTCTRL);
    if (!(val & PLUM2_PCC_CARDRESET)) state->flags |= SS_RESET;
    if (val & PLUM2_PCC_IOCARD) state->flags |= SS_IOCARD;
    if (socket == 0) state->io_irq = PLUM2_PCC1_IRQ;
    else state->io_irq = PLUM2_PCC2_IRQ;

    /* Card Status Change Interrupt Control */
    val = plum2pcc_reg_read(socket, PLUM2_PCC_CSCINTCTRL);
    state->csc_mask = 0;
    if (val & PLUM2_PCC_DETECT) state->csc_mask |= SS_DETECT;
    if (!(state->flags & SS_IOCARD)) {
	if (val & PLUM2_PCC_BATTDEAD) state->csc_mask |= SS_BATDEAD;
	if (val & PLUM2_PCC_BATTWARN) state->csc_mask |= SS_BATWARN;
	if (val & PLUM2_PCC_CSCREADY) state->csc_mask |= SS_READY;
    }

    DEBUG(2, "plum2pcc: get_socket socket %d (flags %#3.3x, Vcc %d, Vpp %d, "
          "irq %d, csc_mask %#2.2x)\n", socket, state->flags,
          state->Vcc, state->Vpp, state->io_irq, state->csc_mask);

    ISA_UNLOCK(socket, flags);

    return 0;
}

static int plum2pcc_set_socket(unsigned int socket, socket_state_t *state)
{
    unsigned long flags;
    u32 val = 0, old;

    if (socket >= PLUM2_PCMCIA_SOCKETS)
	return -EINVAL;

    ISA_LOCK(socket, flags);

    DEBUG(2, "plum2pcc: set_socket socket %d (flags %#3.3x, Vcc %d, Vpp %d, "
          "irq %d, csc_mask %#2.2x)\n", socket, state->flags,
          state->Vcc, state->Vpp, state->io_irq, state->csc_mask);

    /* Interrupt and General Control */
    if (!(state->flags & SS_RESET)) val |= PLUM2_PCC_CARDRESET;
    if (state->flags & SS_IOCARD) val |= PLUM2_PCC_IOCARD;
    plum2pcc_reg_write(socket, PLUM2_PCC_INTCTRL, val);

    /* Power Control */
#if 0
    val = PLUM2_PCC_DISABLERESET;
#else
    val = 0;
#endif
    if (state->flags & SS_PWR_AUTO) val |= PLUM2_PCC_AUTOPOWER;
    if (state->flags & SS_OUTPUT_ENA) val |= PLUM2_PCC_OUTPUTENABLE;

    /* Vpp1 */
    if (state->Vpp != 0) {
	if (state->Vpp == state->Vcc)
	    val |= PLUM2_PCC_VPP1_VCC;
	else
	    return -EINVAL;
    }

    /* Vcc */
    if (state->Vcc != 0) {
	if (state->Vcc == 33)
	    val |= PLUM2_PCC_VCC_3V;
	else if (state->Vcc == 50)
	    val |= PLUM2_PCC_VCC_5V;
	else
	    return -EINVAL;
    }
    if (state->Vcc != 0 || state->Vpp != 0) val |= PLUM2_PCC_CARDPOWENABLE;
    old = plum2pcc_reg_read(socket, PLUM2_PCC_POWERCTRL);
    if ( old != val) {
	plum2pcc_reg_write(socket, PLUM2_PCC_POWERCTRL, val);
	udelay(10000);
    }

    /* Card Status Change */
    if (state->csc_mask & SS_DETECT) val |= PLUM2_PCC_DETECT;
    if (state->csc_mask & SS_BATDEAD) val |= PLUM2_PCC_BATTDEAD;
    if (state->csc_mask & SS_BATWARN) val |= PLUM2_PCC_BATTWARN;
    if (state->csc_mask & SS_READY) val |= PLUM2_PCC_CSCREADY;
    plum2pcc_reg_write(socket, PLUM2_PCC_CSCINTCTRL, val);

    ISA_UNLOCK(socket, flags);

    return 0;
}

static int plum2pcc_get_io_map(unsigned int socket, struct pccard_io_map *io)
{
    unsigned long flags;
    u_char map;
    u32 ctrl, enable;

    if (socket >= PLUM2_PCMCIA_SOCKETS)
	return -EINVAL;

    ISA_LOCK(socket, flags);

    map = io->map;
    if (map > PLUM2_PCC_IO_MAX)
	return -EINVAL;

    io->start = plum2pcc_reg_read(socket, PLUM2_PCC_IO_START(map));
    io->stop = plum2pcc_reg_read(socket, PLUM2_PCC_IO_STOP(map));

    ctrl = plum2pcc_reg_read(socket, PLUM2_PCC_IO_CTRL);
    enable = plum2pcc_reg_read(socket, PLUM2_PCC_WINENABLE);

    if (ctrl & PLUM2_PCC_IO_WS(map)) io->speed = 1;
    else io->speed = 0;

    io->flags = 0;
    if (enable & PLUM2_PCC_IO_ENABLE(map)) io->flags |= MAP_ACTIVE;
    if (ctrl & PLUM2_PCC_IO_0WS(map)) io->flags |= MAP_0WS;
    if (ctrl & PLUM2_PCC_IO_16BIT(map)) io->flags |= MAP_16BIT;
    if (ctrl & PLUM2_PCC_IO_IOIS16S(map)) io->flags |= MAP_AUTOSZ;

    DEBUG(2, "plum2pcc: get_io_map socket %d map %d (flags %#2.2x, %d ns, "
          "%#4.4x-%#4.4x)\n", socket, io->map, io->flags, io->speed,
          io->start, io->stop);

    ISA_UNLOCK(socket, flags);

    return 0;
}

static int plum2pcc_set_io_map(unsigned int socket, struct pccard_io_map *io)
{
    unsigned long flags;
    u_char map;
    u32 val;

    if (socket >= PLUM2_PCMCIA_SOCKETS)
	return -EINVAL;

    ISA_LOCK(socket, flags);

    DEBUG(2, "plum2pcc: set_io_map socket %d map %d (flags %#2.2x, %d ns, "
          "%#4.4x-%#4.4x)\n", socket, io->map, io->flags, io->speed,
          io->start, io->stop);

    map = io->map;
    if (map > PLUM2_PCC_IO_MAX ||
        io->start > PLUM2_PCC_IO_START_MAX ||
        io->stop > PLUM2_PCC_IO_STOP_MAX ||
        io->start > io->stop) return -EINVAL;

    /* Disable I/O Window before change anything */
    plum2pcc_reg_clear(socket, PLUM2_PCC_WINENABLE, PLUM2_PCC_IO_ENABLE(map));

    /* set I/O area 2(default) */
    plum2pcc_reg_write(socket, PLUM2_PCC_IO_ADDRCTRL(map), PLUM2_PCC_IO_AREA2);

    plum2pcc_reg_write(socket, PLUM2_PCC_IO_START(map), io->start);
    plum2pcc_reg_write(socket, PLUM2_PCC_IO_STOP(map), io->stop);

    val = plum2pcc_reg_read(socket, PLUM2_PCC_IO_CTRL) & ~PLUM2_PCC_IO_MASK(map);
    if (io->speed) val |= PLUM2_PCC_IO_WS(map);
    if (io->flags & MAP_0WS) val |= PLUM2_PCC_IO_0WS(map);
    if (io->flags & MAP_16BIT) val |= PLUM2_PCC_IO_16BIT(map);
    if (io->flags & MAP_AUTOSZ) val |= PLUM2_PCC_IO_IOIS16S(map);
    plum2pcc_reg_write(socket, PLUM2_PCC_IO_CTRL, val);

    /* Enable I/O Window */
    if (io->flags & MAP_ACTIVE)
	plum2pcc_reg_set(socket, PLUM2_PCC_WINENABLE, PLUM2_PCC_IO_ENABLE(map));

    ISA_UNLOCK(socket, flags);

    return 0;
}

static int plum2pcc_get_mem_map(unsigned int socket, struct pccard_mem_map *mem)
{
    unsigned long flags;
    u_char map;
    u32 val;

    if (socket >= PLUM2_PCMCIA_SOCKETS)
	return -EINVAL;

    ISA_LOCK(socket, flags);

    map = mem->map;
    if (map > PLUM2_PCC_MEM_MAX)
	return -EINVAL;

    /* Window Enable */
    val = plum2pcc_reg_read(socket, PLUM2_PCC_WINENABLE);
    mem->flags = 0;
    if (val & PLUM2_PCC_MEM_ENABLE(map)) mem->flags |= MAP_ACTIVE;

    /* Memory Window Start Address */
    val = plum2pcc_reg_read(socket, PLUM2_PCC_MEM_START(map));
    mem->sys_start += (val << PLUM2_PCC_MEM_SHIFT);
   
    /* Memory Window Stop Address */
    val = plum2pcc_reg_read(socket, PLUM2_PCC_MEM_STOP(map));
    mem->sys_stop = (val << PLUM2_PCC_MEM_SHIFT) +
                    (((u32)1 << PLUM2_PCC_MEM_SHIFT) - 1);
   
    /* Memory Window Offset Address */
    val = plum2pcc_reg_read(socket, PLUM2_PCC_MEM_OFFSET(map));
    val = (val << PLUM2_PCC_MEM_SHIFT) + mem->sys_start;
    mem->card_start = val & 0x3ffffff;

    /* Memory Window Control */
    val = plum2pcc_reg_read(socket, PLUM2_PCC_MEM_CTRL(map));
    if (val & PLUM2_PCC_MEM_16BIT) mem->flags |= MAP_16BIT;
    if (val & PLUM2_PCC_MEM_0WS) mem->flags |= MAP_0WS;
    if (val & PLUM2_PCC_MEM_WRPROT) mem->flags |= MAP_WRPROT;
    if (val & PLUM2_PCC_MEM_REGACT) mem->flags |= MAP_ATTRIB;
    mem->speed = cycle_time * ((val & PLUM2_PCC_MEM_WS) >> 14);

    DEBUG(2, "plum2pcc: get_mem_map socket %d map %d (flags %#2.2x, %d ns, "
          "%#5.5lx-%#5.5lx, %#5.5x)\n", socket, mem->map, mem->flags,
          mem->speed, mem->sys_start, mem->sys_stop, mem->card_start);

    ISA_UNLOCK(socket, flags);

    return 0;
}

static int plum2pcc_set_mem_map(unsigned int socket, struct pccard_mem_map *mem)
{
    unsigned long flags;
    u_char map;
    u_long size;
    u32 val;

    if (socket >= PLUM2_PCMCIA_SOCKETS)
	return -EINVAL;

    ISA_LOCK(socket, flags);

    DEBUG(2, "plum2pcc: set_mem_map socket %d map %d (flags %#2.2x, %d ns, "
          "%#5.5lx-%#5.5lx, %#5.5x)\n", socket, mem->map, mem->flags,
          mem->speed, mem->sys_start, mem->sys_stop, mem->card_start);

    map = mem->map;
    size = mem->sys_stop - mem->sys_start;
    if (map > PLUM2_PCC_MEM_MAX ||
        mem->card_start > PLUM2_PCC_CARDSTART_MAX ||
        mem->sys_start > PLUM2_PCC_MEM_START_MAX ||
        mem->sys_stop > PLUM2_PCC_MEM_STOP_MAX ||
        mem->sys_start > mem->sys_stop ||
        size > PLUM2_PCC_MEM_SIZE_MAX ||
        mem->speed > PLUM2_PCC_MEM_SPEED_MAX) return -EINVAL;

    /* Disable Memory Window before change anything */
    plum2pcc_reg_clear(socket, PLUM2_PCC_WINENABLE, PLUM2_PCC_MEM_ENABLE(map));

    /* Memory Window Start Address */
    val = (mem->sys_start >> PLUM2_PCC_MEM_SHIFT) & PLUM2_PCC_MEM_MASK;
    plum2pcc_reg_write(socket, PLUM2_PCC_MEM_START(map), val);

    /* Memory Window Stop Address */
    val = (mem->sys_stop >> PLUM2_PCC_MEM_SHIFT) & PLUM2_PCC_MEM_MASK;
    plum2pcc_reg_write(socket, PLUM2_PCC_MEM_STOP(map), val);

    /* Memory Window Offset Address */
    val = ((mem->card_start - mem->sys_start) >> PLUM2_PCC_MEM_SHIFT) &
          PLUM2_PCC_MEM_MASK;
    plum2pcc_reg_write(socket, PLUM2_PCC_MEM_OFFSET(map), val);

    /* Memory Window Control */
    val = 0;
    if (mem->flags & MAP_16BIT) val |= PLUM2_PCC_MEM_16BIT;
//    if (mem->flags & MAP_0WS) val |= PLUM2_PCC_MEM_0WS;
//    val |= ((mem->speed / cycle_time) << 14) & PLUM2_PCC_MEM_WS;
    val |= PLUM2_PCC_MEM_0WS;
    if (mem->flags & MAP_WRPROT) val |= PLUM2_PCC_MEM_WRPROT;
    if (mem->flags & MAP_ATTRIB) val |= PLUM2_PCC_MEM_REGACT;
    val |= PLUM2_PCC_MEM_MAP3;
    plum2pcc_reg_write(socket, PLUM2_PCC_MEM_CTRL(map), val);

    /* Enable Memory Window */
    if (mem->flags & MAP_ACTIVE)
	plum2pcc_reg_set(socket, PLUM2_PCC_WINENABLE, PLUM2_PCC_MEM_ENABLE(map));

    ISA_UNLOCK(socket, flags);

    return 0;
}

static int plum2pcc_get_irq(unsigned int socket, int *irq)
{
    if (socket >= PLUM2_PCMCIA_SOCKETS)
	return -EINVAL;

    switch (socket) {
	case 0:
	    *irq = PLUM2_PCC1_IRQ;
	    break;
	case 1:
	    *irq = PLUM2_PCC2_IRQ;
	    break;
	default:
	    return -EINVAL;
    }

    DEBUG(2, "plum2pcc: get_irq socket %d irq=%d\n", socket, *irq);

    return 0;
}

static int plum2pcc_init(unsigned int socket)
{
    u_char i;
    pccard_io_map io = { 0, 0, 0, 0, 0 };
    pccard_mem_map mem = { 0, 0, 0, 0, 0, 0 };

    plum2pcc_set_socket(socket, &dead_socket);

    for (i = 0; i < PLUM2_PCC_IO_WINDOWS; i++) {
	io.map = i;
	plum2pcc_set_io_map(socket, &io);
    }

    for (i = 0; i < PLUM2_PCC_MEM_WINDOWS; i++) {
	mem.map = i;
	plum2pcc_set_mem_map(socket, &mem);
    }

    DEBUG(2, "plum2pcc: plum2pcc_init\n");

    return 0;
}

static int plum2pcc_suspend(unsigned int socket)
{
    DEBUG(2, "plum2pcc: plum2pcc_suspend socket %d\n", socket);

    return plum2pcc_set_socket(socket, &dead_socket);
}

static int plum2pcc_register_callback(unsigned int socket,
                                      void (*handler)(void *, unsigned int),
                                      void * info)
{
    DEBUG(2, "plum2pcc: plum2pcc_register_callback socket %d, 0x%p\n",
          socket, handler);

    socket_table[socket].handler = handler;
    socket_table[socket].info = info;
    if (handler == NULL)
	MOD_DEC_USE_COUNT;
    else
	MOD_INC_USE_COUNT;

    return 0;
}

static int plum2pcc_inquire_socket(unsigned int socket, socket_cap_t *cap)
{
    *cap = plum2pcc_cap[socket];
    cap->bus = &plum2pcc_bus_operations;

    DEBUG(2, "plum2pcc: plum2pcc_inquire_socket socket %d\n", socket);

    return 0;
}

static void plum2pcc_proc_setup(unsigned int socket,
                                struct proc_dir_entry *base)
{
}

static struct pccard_operations plum2pcc_operations = {
        plum2pcc_init,
        plum2pcc_suspend,
        plum2pcc_register_callback,
        plum2pcc_inquire_socket,
        plum2pcc_get_status,
        plum2pcc_get_socket,
        plum2pcc_set_socket,
        plum2pcc_get_io_map,
        plum2pcc_set_io_map,
        plum2pcc_get_mem_map,
        plum2pcc_set_mem_map,
	plum2pcc_get_irq,
        plum2pcc_proc_setup
};

/*====================================================================*/

static int reg_offset[45] = {
	0x000, 0x004, 0x008, 0x00c, 0x010, 0x014, 0x018, 0x01c, 0x020,
	0x024, 0x028, 0x030, 0x034, 0x038, 0x040, 0x044, 0x048, 0x04c,
	0x058, 0x060, 0x064, 0x068, 0x06c, 0x078, 0x080, 0x084, 0x088,
	0x08c, 0x0a0, 0x0a4, 0x0a8, 0x0ac, 0x0c0, 0x0c4, 0x0c8, 0x0cc,
	0x0ec, 0x0f0, 0x0f4, 0x0f8, 0x0fc, 0x100, 0x104, 0x108, 0x10c

};

void plum2pcc_debug_print(void)
{
	int socket, i;

	printk("ss: Windows CE configuration\n");
	
	for (socket = 0; socket < 2; socket++) {
		printk("ss: socket %d\n", socket);
		for (i = 0; i < 45; i++) {
			printk("ss: 0x%03x=0x%08x\n",
			       reg_offset[i], plum2pcc_reg_read(socket, reg_offset[i]));
		}
		printk("\n");
	}
	
}

/*
 * Initial Toshiba PLUM2 PCMCIA Controller
 */

static int __init init_plum2pcc(void)
{
    servinfo_t serv;
    adjust_t adj;
    plum2pcc_map_t *map;
    u8 i;

    pcmcia_get_card_services_info(&serv);
    if (serv.Revision != CS_RELEASE_CODE) {
        printk(KERN_NOTICE "plum2pcc: Card Services release "
               "does not match!\n");
        return -1;
    }

    printk(KERN_INFO "Toshiba PLUM2 PCMCIA Controller, 2 sockets\n");
    printk(KERN_INFO "  PCMCIA Controller irq  %d\n", PLUM2_CSC_IRQ);
    printk(KERN_INFO "  PC-Card           irq  %d, %d\n",
                                     PLUM2_PCC1_IRQ, PLUM2_PCC2_IRQ);

    sockets = PLUM2_PCMCIA_SOCKETS;
    socket_table[0].reg_base = PLUM2_PCC1REG_BASE;
    socket_table[1].reg_base = PLUM2_PCC2REG_BASE;
    for (i = 0; i < sockets; i++) {
        socket_table[i].psock = i;
        socket_table[i].handler = NULL;
        socket_table[i].info = NULL;
    }

//  plum2pcc_debug_print();

    /* Supply power to slots and cards */
    plum2pcc_pccard_startup();

    for (i = 0; i < sockets; i++) {
	/* Slot enable */
	plum2pcc_reg_write(i, PLUM2_PCC_SLOTCTRL, PLUM2_PCC_SLOT_ON);

	/* Support 3.3V PC-Card */
	plum2pcc_reg_write(i, PLUM2_PCC_FUNCCTRL,
	               PLUM2_PCC_VSS_ENABLE | PLUM2_PCC_3VPOWER_EN);
    }

    plum2pcc_reg_clear(0, PLUM2_PCC_GCTRL, PLUM2_PCC_EWB_CSC_INT);

    request_irq(PLUM2_CSC_IRQ, plum2pcc_interrupt, 0, "PLUM2 PCMCIA",
                plum2pcc_interrupt);

    if (register_ss_entry(sockets, &plum2pcc_operations) != 0)
        printk(KERN_NOTICE "plum2pcc: register_ss_entry() failed\n");

    map = plum2pcc_io_map + 1;
    if (kseg2_page_alloc(map->base, map->size, KSEG2_FLAGS_ALL,
                         &(map->offset)) != 0)
	return -1;
    plum2pcc_io_base = map->offset - mips_io_port_base;
    DEBUG(1, "plum2pcc: io window base 0x%08x\n", plum2pcc_io_base);

    adj.Action = ADD_MANAGED_RESOURCE;
    adj.Resource = RES_IO_RANGE;
    adj.resource.io.BasePort = 0x100;
    adj.resource.io.NumPorts = 0x500;
    if (pcmcia_adjust_resource_info(NULL, &adj) != 0) {
	DEBUG(1, "plum2pcc: Error! added io resource\n");
	return -1;
    }

    map = plum2pcc_mem_map + 2;
    if (kseg2_page_alloc(map->base, map->size, KSEG2_FLAGS_ALL,
                         &(map->offset)) != 0)
	return -1;
    plum2pcc_mem_base = map->offset;
    DEBUG(1, "plum2pcc: memory window base 0x%08x\n", plum2pcc_mem_base);

    adj.Resource = RES_MEMORY_RANGE;
    adj.resource.memory.Base = 0UL; 
    adj.resource.memory.Size = map->size;
    if (pcmcia_adjust_resource_info(NULL, &adj) != 0) {
	DEBUG(1, "plum2pcc: Error! added memory resource\n");
	return -1;
    }

    return 0;
}

static void __exit exit_plum2pcc(void)
{
    u8 i;

    unregister_ss_entry(&plum2pcc_operations);
        
    free_irq(PLUM2_CSC_IRQ, plum2pcc_interrupt);
    for (i = 0; i < sockets; i++) {
        /* Turn off all interrupt sources! */
        plum2pcc_reg_write(i, PLUM2_PCC_CSCINTCTRL, 0);

	/* I/O and Memory Windows disable */
	plum2pcc_reg_write(i, PLUM2_PCC_WINENABLE, 0);
    }
}

module_init(init_plum2pcc);
module_exit(exit_plum2pcc);
