/*
 *  linux/drivers/video/cfb4.c -- Low level frame buffer operations for 4 bpp
 *				  packed pixels
 *
 *	Created 26 Dec 1997 by Michael Schmitz
 *	Based on the old macfb.c 4bpp code by Alan Cox
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file COPYING in the main directory of this archive for
 *  more details.
 */

#include <linux/module.h>
#include <linux/tty.h>
#include <linux/console.h>
#include <linux/string.h>
#include <linux/fb.h>

#include <video/fbcon.h>
#include <video/fbcon-cfb4.h>


    /*
     *  4 bpp packed pixels
     */

    /*
     *  IFF the font is even pixel aligned (that is to say each
     *  character start is a byte start in the pixel pairs). That
     *  avoids us having to mask bytes and means we won't be here
     *  all week. On a MacII that matters _lots_
     */

static u16 nibbletab_cfb4[] = {
#if defined(__BIG_ENDIAN)
    0x0000,0x000f,0x00f0,0x00ff,
    0x0f00,0x0f0f,0x0ff0,0x0fff,
    0xf000,0xf00f,0xf0f0,0xf0ff,
    0xff00,0xff0f,0xfff0,0xffff
#elif defined(__LITTLE_ENDIAN)
    0x0000,0x0f00,0xf000,0xff00,
    0x000f,0x0f0f,0xf00f,0xff0f,
    0x00f0,0x0ff0,0xf0f0,0xfff0,
    0x00ff,0x0fff,0xf0ff,0xffff
#else
#error FIXME: No endianness??
#endif
};


#ifdef CONFIG_FBCON_FONTWIDTH8_ONLY

/* this is Font Width in Bytes */
#define FWB 4
#define CFB4_FONTMASK FONTWIDTH(8)

#else

#define FWB (fontwidth(p) / 2)
#define CFB4_FONTMASK FONTWIDTH(2) | FONTWIDTH(2) | FONTWIDTH(4) \
	| FONTWIDTH(6) | FONTWIDTH(8) | FONTWIDTH(10) | FONTWIDTH(12) \
	| FONTWIDTH(14) | FONTWIDTH(16)
static u8 hntab_cfb4[] = { 0x00,0x0f,0xf0,0xff };

#endif

void fbcon_cfb4_setup(struct display *p)
{
    p->next_line = p->line_length ? p->line_length : p->var.xres_virtual>>1;
    p->next_plane = 0;
}

void fbcon_cfb4_bmove(struct display *p, int sy, int sx, int dy, int dx,
		      int height, int width)
{
	int bytes = p->next_line, linesize = bytes * fontheight(p), rows;
	u8 *src,*dst;

	if (sx == 0 && dx == 0 && width * FWB == bytes) {
		fb_memmove(p->screen_base + dy * linesize,
			  p->screen_base + sy * linesize,
			  height * linesize);
	}
	else {
		if (dy < sy || (dy == sy && dx < sx)) {
			src = p->screen_base + sy * linesize + sx * FWB;
			dst = p->screen_base + dy * linesize + dx * FWB;
			for (rows = height * fontheight(p) ; rows-- ;) {
				fb_memmove(dst, src, width * 4);
				src += bytes;
				dst += bytes;
			}
		}
		else {
			src = p->screen_base + (sy+height) * linesize + sx * FWB
				- bytes;
			dst = p->screen_base + (dy+height) * linesize + dx * FWB
				- bytes;
			for (rows = height * fontheight(p) ; rows-- ;) {
				fb_memmove(dst, src, width * 4);
				src -= bytes;
				dst -= bytes;
			}
		}
	}
}

void fbcon_cfb4_clear(struct vc_data *conp, struct display *p, int sy, int sx,
		      int height, int width)
{
	u8 *dest0,*dest;
	int bytes=p->next_line,lines=height * fontheight(p), rows;
	u32 bgx;

/*	if(p->screen_base!=0xFDD00020)
		mac_boom(1);*/
	dest = p->screen_base + sy * fontheight(p) * bytes + sx * FWB;

	bgx=attr_bgcol_ec(p,conp);
	bgx |= (bgx << 4);	/* expand the colour to 32bits */
	bgx |= (bgx << 8);
	bgx |= (bgx << 16);

	if (sx == 0 && width * FWB == bytes) {
		fb_memset(dest, bgx, lines * width * FWB);
	} else {
		dest0=dest;
		for (rows = lines; rows-- ; dest0 += bytes) {
			dest=dest0;
			fb_memset(dest, bgx, width * FWB);
		}
	}
}

#ifndef CONFIG_FBCON_FONTWIDTH8_ONLY
static void fbcon_cfb4_putc_guts(int fw, u8 **pcdat, u16 eorx, u16 bgx, u8 *dest)
{
	u8 pixl, pixr = 0;

	pixl = *(*pcdat)++;
	if (fw > 8)
		pixr = *(*pcdat)++;

	if (!((long)dest & 1)) {
		/* sorry about the gotos, but without, this code bloats massively */
		switch(fw) {
		case 14:
			fb_writeb((hntab_cfb4[(pixr >> 2) & 0x3] & eorx) ^ bgx, dest+6);
			goto finish_12bits;
		case 10:
			fb_writeb((hntab_cfb4[pixr >> 6] & eorx) ^ bgx, dest+4);
			goto finish_8bits;
		case 6:
			fb_writeb((hntab_cfb4[(pixl >> 2) & 0x3] & eorx) ^ bgx, dest+2);
			goto finish_4bits;
		case 2:
			fb_writeb((hntab_cfb4[pixl >> 6] & eorx) ^ bgx, dest+0);
			break;
		case 16:
			fb_writew((nibbletab_cfb4[pixr & 0xf] & eorx) ^ bgx, dest+6);
		case 12:
finish_12bits:
			fb_writew((nibbletab_cfb4[pixr >> 4] & eorx) ^ bgx, dest+4);
		case 8:
finish_8bits:
			fb_writew((nibbletab_cfb4[pixl & 0xf] & eorx) ^ bgx, dest+2);
		case 4:
finish_4bits:
			fb_writew((nibbletab_cfb4[pixl >> 4] & eorx) ^ bgx, dest+0);
			break;
		}
	} else {
		/* fontwidths 4, 8, 12, or 16 never get here */
		u8 pixm = (pixl << 2) | (pixr >> 2);

		switch(fw) {
		case 14:
			fb_writew((nibbletab_cfb4[(pixr >> 2) & 0xf] & eorx) ^ bgx, dest+5);
		case 10:
			fb_writew((nibbletab_cfb4[pixm & 0xf] & eorx) ^ bgx, dest+3);
		case 6:
			fb_writew((nibbletab_cfb4[(pixl >> 2) & 0xf] & eorx) ^ bgx, dest+1);
		case 2:
			fb_writeb((hntab_cfb4[pixl >> 6] & eorx) ^ bgx, dest+0);
			break;
		}
	}
}
#endif

void fbcon_cfb4_putc(struct vc_data *conp, struct display *p, int c, int yy,
		     int xx)
{
	u8 *dest,*cdat;
	int bytes=p->next_line,rows;
	u16 eorx,fgx,bgx;

	dest = p->screen_base + yy * fontheight(p) * bytes + xx * FWB;
	cdat = p->fontdata + (c & p->charmask) * fontheight(p);

	fgx=attr_fgcol(p,c);
	bgx=attr_bgcol(p,c);
	fgx |= (fgx << 4);
	fgx |= (fgx << 8);
	bgx |= (bgx << 4);
	bgx |= (bgx << 8);
	eorx = fgx ^ bgx;

	for (rows = fontheight(p) ; rows-- ; dest += bytes) {
#ifdef CONFIG_FBCON_FONTWIDTH8_ONLY
		fb_writew((nibbletab_cfb4[*cdat >> 4] & eorx) ^ bgx, dest+0);
		fb_writew((nibbletab_cfb4[*cdat++ & 0xf] & eorx) ^ bgx, dest+2);
#else
		fbcon_cfb4_putc_guts(fontwidth(p), &cdat, eorx, bgx, dest);
#endif
	}
}

void fbcon_cfb4_putcs(struct vc_data *conp, struct display *p, 
		      const unsigned short *s, int count, int yy, int xx)
{
	u8 *cdat, *dest, *dest0;
	u16 c;
	int rows,bytes=p->next_line;
	u16 eorx, fgx, bgx;

	dest0 = p->screen_base + yy * fontheight(p) * bytes + xx * FWB;
	fgx=attr_fgcol(p,scr_readw(s));
	bgx=attr_bgcol(p,scr_readw(s));
	fgx |= (fgx << 4);
	fgx |= (fgx << 8);
	bgx |= (bgx << 4);
	bgx |= (bgx << 8);
	eorx = fgx ^ bgx;
	while (count--) {
		c = scr_readw(s++) & p->charmask;
		cdat = p->fontdata + c * fontheight(p);

		for (rows = fontheight(p), dest = dest0; rows-- ; dest += bytes) {
#ifdef CONFIG_FBCON_FONTWIDTH8_ONLY
			fb_writew((nibbletab_cfb4[*cdat >> 4] & eorx) ^ bgx, dest+0);
			fb_writew((nibbletab_cfb4[*cdat++ & 0xf] & eorx) ^ bgx, dest+2);
#else
			fbcon_cfb4_putc_guts(fontwidth(p), &cdat, eorx, bgx, dest);
#endif
		}
		dest0+=FWB;
	}
}

void fbcon_cfb4_revc(struct display *p, int xx, int yy)
{
	u8 *dest;
	int bytes=p->next_line, rows;

	dest = p->screen_base + yy * fontheight(p) * bytes + xx * FWB;
	for (rows = fontheight(p) ; rows-- ; dest += bytes) {
#ifdef CONFIG_FBCON_FONTWIDTH8_ONLY
		fb_writel(fb_readl(dest+0) ^ 0xffffffff, dest+0);
#else
		switch(fontwidth(p)) {
		case 14:
			fb_writeb(fb_readb(dest+6) ^ 0xff, dest+6);
			fb_writeb(fb_readb(dest+5) ^ 0xff, dest+5);
		case 10:
			fb_writeb(fb_readb(dest+4) ^ 0xff, dest+4);
			fb_writeb(fb_readb(dest+3) ^ 0xff, dest+3);
		case 6:
			fb_writeb(fb_readb(dest+2) ^ 0xff, dest+2);
			fb_writeb(fb_readb(dest+1) ^ 0xff, dest+1);
		case 2:
			fb_writeb(fb_readb(dest+0) ^ 0xff, dest+0);
			break;
		case 12:
			fb_writew(fb_readw(dest+4) ^ 0xffff, dest+4);
			fb_writew(fb_readw(dest+2) ^ 0xffff, dest+2);
		case 4:
			fb_writew(fb_readw(dest+0) ^ 0xffff, dest+0);
			break;
		case 16:
			fb_writel(fb_readl(dest+4) ^ 0xffffffff, dest+4);
		case 8:
			fb_writel(fb_readl(dest+0) ^ 0xffffffff, dest+0);
			break;
		}
#endif
	}
}


    /*
     *  `switch' for the low level operations
     */

struct display_switch fbcon_cfb4 = {
    fbcon_cfb4_setup, fbcon_cfb4_bmove, fbcon_cfb4_clear, fbcon_cfb4_putc,
    fbcon_cfb4_putcs, fbcon_cfb4_revc, NULL, NULL, NULL, CFB4_FONTMASK
};


#ifdef MODULE
int init_module(void)
{
    return 0;
}

void cleanup_module(void)
{}
#endif /* MODULE */


    /*
     *  Visible symbols for modules
     */

EXPORT_SYMBOL(fbcon_cfb4);
EXPORT_SYMBOL(fbcon_cfb4_setup);
EXPORT_SYMBOL(fbcon_cfb4_bmove);
EXPORT_SYMBOL(fbcon_cfb4_clear);
EXPORT_SYMBOL(fbcon_cfb4_putc);
EXPORT_SYMBOL(fbcon_cfb4_putcs);
EXPORT_SYMBOL(fbcon_cfb4_revc);
