/*
 *  linux/drivers/video/r3912fb.c -- R3912 embedded frame buffer device
 *
 *	 Based on simple frame buffer which was
 *	 based on virtual frame buffer by Geert Uytterhoeven
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/malloc.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <asm/bootinfo.h>
#include <asm/r39xx.h>
#include <asm/uaccess.h>

#include <video/fbcon.h>
#include <video/fbcon-mfb.h>
#include <video/fbcon-cfb2.h>
#include <video/fbcon-cfb4.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb24.h>
#include <video/fbcon-cfb32.h>
#include <video/fbcon-mac.h>

/*
 * This can go back into a header file if/when more platforms are defined.
 * For now, it's just Mobilon parameters.
 *
 * VIDEORAM_SIZE     Normally x_res * y_res * (bpp/8) 
 * FB_X_RES          Visible horizontal resolution in pixels
 * FB_X_VIRTUAL_RES  Horizontal resolution of framebuffer memory in pixels
 *                   (default is FB_X_RES)
 * FB_Y_RES          Visible vertical resolution in pixels
 * FB_Y_VIRTUAL_RES  Vertical resolution of framebuffer memory in pixels
 *                   (default FB_Y_RES)
 * FB_BPP            Bits per pixel
 * FB_IS_GREY        1 means greyscale device, otherwise 0
 * FB_IS_INVERSE     1 means pallette 0 is white instead of black
 *
 * FIXME: this has to be set up depending on size and color depth
 */

#ifdef CONFIG_PHILIPS_VELO
#define FB_X_RES       480
#define FB_Y_RES       240
#ifdef CONFIG_PHILIPS_VELO1_4GRAY
#define FB_BPP         2
#else
#ifdef CONFIG_PHILIPS_VELO1_16GRAY
#define FB_BPP	       4
#else
#error velo500 owner, please fix this
#endif
#endif
#define FB_IS_GREY     1
#define FB_IS_INVERSE  0
#define VIDEORAM_SIZE  (FB_X_RES * FB_Y_RES * FB_BPP / 8)

#elif defined(CONFIG_VTECH_HELIO)
#define FB_X_RES       160
#define FB_Y_RES       160
#ifdef CONFIG_FB_R3912_FORCE_MONO
#define FB_BPP 1
#else
#define FB_BPP         4
#endif
#define FB_IS_GREY     1
#define FB_IS_INVERSE  0 /* it is inverse, but that gives poor results */
#define VIDEORAM_SIZE  (FB_X_RES * FB_Y_RES * FB_BPP / 8)

 #define LCD_CLK_36M    0
 #define LCD_CLK_18M    1
 #define LCD_CLK_9M     2
 #define LCD_CLK_4M     3

 #define VIDEO_RF       LCD_CLK_9M
 #define LCD_CLK        (36864000 / (1 << VIDEO_RF))
 #define LCD_BAUD       0x6               /* for 4.5MHz Lcd Clock only */
 #define LCD_CP         (LCD_CLK / (LCD_BAUD * 2 + 2))

 #define LCD_FRAMERATE  70
 #define LCD_LINERATE   (70 * (FB_X_RES + 1))
 #define LCD_VIDRATE    ((LCD_CP / LCD_LINERATE) - 1)

#else
#define FB_X_RES       640
#define FB_Y_RES       240
#define FB_BPP         8
#define FB_IS_GREY     0
#define FB_IS_INVERSE  0
#define VIDEORAM_SIZE  (FB_X_RES * FB_Y_RES * FB_BPP / 8)
#endif

#define arraysize(x)	(sizeof(x)/sizeof(*(x)))

u_long videomemory, videomemorysize = VIDEORAM_SIZE;

static int currcon = 0;
static struct display disp;
static struct fb_info fb_info;
static struct { u_char red, green, blue, pad; } palette[256];

#if defined(FBCON_HAS_CFB8)
static union {
#ifdef FBCON_HAS_CFB8
    u8 cfb8[16];
#endif
} fbcon_cmap;
#endif

static char r3912fb_name[16] = "R3912 FB";

#ifndef FB_X_VIRTUAL_RES
#define FB_X_VIRTUAL_RES FB_X_RES
#endif

#ifndef FB_Y_VIRTUAL_RES
#define FB_Y_VIRTUAL_RES FB_Y_RES
#endif

static struct fb_var_screeninfo r3912fb_default = {
    FB_X_RES, FB_Y_RES, FB_X_VIRTUAL_RES, FB_Y_VIRTUAL_RES, 0, 0, FB_BPP, FB_IS_GREY,
    {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0},
    0, FB_ACTIVATE_NOW, -1, -1, 0, 20000, 64, 64, 32, 32, 64, 2,
    0, FB_VMODE_NONINTERLACED, {0,0,0,0,0,0}
};


    /*
     *  Interface used by the world
     */

int r3912fb_setup(char*);

static int r3912fb_open(struct fb_info *info, int user);
static int r3912fb_release(struct fb_info *info, int user);
static int r3912fb_get_fix(struct fb_fix_screeninfo *fix, int con,
		       struct fb_info *info);
static int r3912fb_get_var(struct fb_var_screeninfo *var, int con,
		       struct fb_info *info);
static int r3912fb_set_var(struct fb_var_screeninfo *var, int con,
		       struct fb_info *info);
static int r3912fb_pan_display(struct fb_var_screeninfo *var, int con,
			   struct fb_info *info);
static int r3912fb_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			struct fb_info *info);
static int r3912fb_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			struct fb_info *info);
static int r3912fb_ioctl(struct inode *inode, struct file *file, u_int cmd,
		     u_long arg, int con, struct fb_info *info);


    /*
     *  Interface to the low level console driver
     */

int r3912fb_init(void);
static int r3912fbcon_switch(int con, struct fb_info *info);
static int r3912fbcon_updatevar(int con, struct fb_info *info);
static void r3912fbcon_blank(int blank, struct fb_info *info);


    /*
     *  Internal routines
     */

static u_long get_line_length(int xres_virtual, int bpp);
static void r3912fb_encode_fix(struct fb_fix_screeninfo *fix,
			   struct fb_var_screeninfo *var);
static void set_color_bitfields(struct fb_var_screeninfo *var);
static int r3912fb_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
                         u_int *transp, struct fb_info *info);
static int r3912fb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
                         u_int transp, struct fb_info *info);
static void do_install_cmap(int con, struct fb_info *info);

static struct fb_ops r3912fb_ops = {
	owner: THIS_MODULE,
	fb_open: r3912fb_open,
	fb_release: r3912fb_release,
	fb_get_fix: r3912fb_get_fix,
	fb_get_var: r3912fb_get_var,
	fb_set_var: r3912fb_set_var,
	fb_get_cmap: r3912fb_get_cmap,
	fb_set_cmap: r3912fb_set_cmap,
	fb_pan_display: r3912fb_pan_display,
	fb_ioctl: r3912fb_ioctl,
	fb_mmap: 0,
	fb_rasterimg: 0
};

    /*
     *  Open/Release the frame buffer device
     */

static int r3912fb_open(struct fb_info *info, int user)
{
    /*                                                                     
     *  Nothing, only a usage count for the moment                          
     */                                                                    

    MOD_INC_USE_COUNT;
    return(0);                              
}
        
static int r3912fb_release(struct fb_info *info, int user)
{
    MOD_DEC_USE_COUNT;
    return(0);                                                    
}


    /*
     *  Get the Fixed Part of the Display
     */

static int r3912fb_get_fix(struct fb_fix_screeninfo *fix, int con,
		       struct fb_info *info)
{
    struct fb_var_screeninfo *var;

    if (con == -1)
	var = &r3912fb_default;
    else
	var = &fb_display[con].var;
    r3912fb_encode_fix(fix, var);
    return 0;
}


    /*
     *  Get the User Defined Part of the Display
     */

static int r3912fb_get_var(struct fb_var_screeninfo *var, int con,
		       struct fb_info *info)
{
    if (con == -1)
	*var = r3912fb_default;
    else
	*var = fb_display[con].var;
    set_color_bitfields(var);
    return 0;
}


    /*
     *  Set the User Defined Part of the Display
     */

static int r3912fb_set_var(struct fb_var_screeninfo *var, int con,
		       struct fb_info *info)
{
    int err, activate = var->activate;
    int oldxres, oldyres, oldvxres, oldvyres, oldbpp;
    u_long line_length;
    struct display *display;

    if (con >= 0)
	display = &fb_display[con];
    else
	display = &disp;	/* used during initialization */

    /*
     *  FB_VMODE_CONUPDATE and FB_VMODE_SMOOTH_XPAN are equal!
     *  as FB_VMODE_SMOOTH_XPAN is only used internally
     */

    if (var->vmode & FB_VMODE_CONUPDATE) {
	var->vmode |= FB_VMODE_YWRAP;
	var->xoffset = display->var.xoffset;
	var->yoffset = display->var.yoffset;
    }

    /*
     *  Some very basic checks
     */
    if (!var->xres)
	var->xres = 1;
    if (!var->yres)
	var->yres = 1;
    if (var->xres > var->xres_virtual)
	var->xres_virtual = var->xres;
    if (var->yres > var->yres_virtual)
	var->yres_virtual = var->yres;
    if (var->bits_per_pixel <= 1)
	var->bits_per_pixel = 1;
    else if (var->bits_per_pixel <= 2)
	var->bits_per_pixel = 2;
    else if (var->bits_per_pixel <= 4)
	var->bits_per_pixel = 4;
    else if (var->bits_per_pixel <= 8)
	var->bits_per_pixel = 8;
    else if (var->bits_per_pixel <= 16)
	var->bits_per_pixel = 16;
    /* fbcon doesn't support this (yet) */
    else if (var->bits_per_pixel <= 24)
	var->bits_per_pixel = 24;
    else if (var->bits_per_pixel <= 32)
	var->bits_per_pixel = 32;
    else
	return -EINVAL;

    /*
     *  Memory limit
     */
    line_length = get_line_length(var->xres_virtual, var->bits_per_pixel);
    if (line_length*var->yres_virtual > videomemorysize)
	return -ENOMEM;

    set_color_bitfields(var);

    if ((activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_NOW) {
	oldxres = display->var.xres;
	oldyres = display->var.yres;
	oldvxres = display->var.xres_virtual;
	oldvyres = display->var.yres_virtual;
	oldbpp = display->var.bits_per_pixel;
	display->var = *var;
	if (oldxres != var->xres || oldyres != var->yres ||
	    oldvxres != var->xres_virtual || oldvyres != var->yres_virtual ||
	    oldbpp != var->bits_per_pixel) {
	    struct fb_fix_screeninfo fix;

	    r3912fb_encode_fix(&fix, var);
	    display->screen_base = (char *)videomemory;
	    display->visual = fix.visual;
	    display->type = fix.type;
	    display->type_aux = fix.type_aux;
	    display->ypanstep = fix.ypanstep;
	    display->ywrapstep = fix.ywrapstep;
	    display->line_length = fix.line_length;
	    display->can_soft_blank = 0;
	    display->inverse = FB_IS_INVERSE;
	    display->scrollmode = SCROLL_YREDRAW;
#ifdef FBCON_HAS_MAC
            display->dispsw = &fbcon_mac; /* it can handle everything */
#else
	    switch (var->bits_per_pixel) {
#ifdef FBCON_HAS_MFB
		case 1:
		    display->dispsw = &fbcon_mfb;
		    break;
#endif
#ifdef FBCON_HAS_CFB2
		case 2:
		    display->dispsw = &fbcon_cfb2;
		    break;
#endif
#ifdef FBCON_HAS_CFB4
		case 4:
		    display->dispsw = &fbcon_cfb4;
		    break;
#endif
#ifdef FBCON_HAS_CFB8
		case 8:
		    display->dispsw = &fbcon_cfb8;
		    display->dispsw_data = fbcon_cmap.cfb8;
		    break;
#endif
		default:
		    display->dispsw = &fbcon_dummy;
		    break;
	    }
#endif
	    if (fb_info.changevar)
		(*fb_info.changevar)(con);
	}
	if (oldbpp != var->bits_per_pixel) {
	    if ((err = fb_alloc_cmap(&display->cmap, 0, 0)))
		return err;
	    do_install_cmap(con, info);
	}
    }
    return 0;
}


    /*
     *  Pan or Wrap the Display
     *
     *  This call looks only at xoffset, yoffset and the FB_VMODE_YWRAP flag
     */

static int r3912fb_pan_display(struct fb_var_screeninfo *var, int con,
			   struct fb_info *info)
{
#if 1
    return -EINVAL;
#else
    if (var->vmode & FB_VMODE_YWRAP) {
	if (var->yoffset < 0 ||
	    var->yoffset >= fb_display[con].var.yres_virtual ||
	    var->xoffset)
	    return -EINVAL;
    } else {
	if (var->xoffset+fb_display[con].var.xres >
	    fb_display[con].var.xres_virtual ||
	    var->yoffset+fb_display[con].var.yres >
	    fb_display[con].var.yres_virtual)
	    return -EINVAL;
    }
    fb_display[con].var.xoffset = var->xoffset;
    fb_display[con].var.yoffset = var->yoffset;
    if (var->vmode & FB_VMODE_YWRAP)
	fb_display[con].var.vmode |= FB_VMODE_YWRAP;
    else
	fb_display[con].var.vmode &= ~FB_VMODE_YWRAP;
    return 0;
#endif
}

    /*
     *  Get the Colormap
     */

static int r3912fb_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			struct fb_info *info)
{
    if (con == currcon) /* current console? */
	return fb_get_cmap(cmap, kspc, r3912fb_getcolreg, info);
    else if (fb_display[con].cmap.len) /* non default colormap? */
	fb_copy_cmap(&fb_display[con].cmap, cmap, kspc ? 0 : 2);
    else
	fb_copy_cmap(fb_default_cmap(1<<fb_display[con].var.bits_per_pixel),
		     cmap, kspc ? 0 : 2);
    return 0;
}

    /*
     *  Set the Colormap
     */

static int r3912fb_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			struct fb_info *info)
{
    int err;

    if (!fb_display[con].cmap.len) {	/* no colormap allocated? */
	if ((err = fb_alloc_cmap(&fb_display[con].cmap,
			      1<<fb_display[con].var.bits_per_pixel, 0)))
	    return err;
    }
    if (con == currcon)			/* current console? */
	return fb_set_cmap(cmap, kspc, r3912fb_setcolreg, info);
    else
	fb_copy_cmap(cmap, &fb_display[con].cmap, kspc ? 0 : 1);
    return 0;
}


    /*
     *  R3912 Frame Buffer Specific ioctls
     */

static int r3912fb_ioctl(struct inode *inode, struct file *file, u_int cmd,
		     u_long arg, int con, struct fb_info *info)
{
    return -EINVAL;
}

int __init r3912fb_setup(char *options)
{
    char *this_opt;

    fb_info.fontname[0] = '\0';

    if (!options || !*options)
	return 0;

    for (this_opt = strtok(options, ","); this_opt;
	 this_opt = strtok(NULL, ",")) {
	if (!strncmp(this_opt, "font:", 5))
	    strcpy(fb_info.fontname, this_opt+5);
    }
    return 0;
}

/*
 * Helio specific LCD control
 */

#if defined(CONFIG_VTECH_HELIO)

void turn_on_helio_lcd(void)
{
    VidCtrl2 = ((LCD_VIDRATE << VIDRATE_SHIFT) & VIDRATE_MASK) |
	    (((FB_X_RES / 4 - 1) << HORZVAL_SHIFT) & HORZVAL_MASK) |
	    ((FB_Y_RES - 1) & LINEVAL_MASK);

    VidCtrl1 |= ENVID;         /* Enable Video Logic */
    VidCtrl1 |= DISPON;             /* Display on, but disable video logic */

    ClockControl |= CLK_ENVIDCLK;

    MFIOSelect |= MFIO_PIN_LCD_POWER;
    MFIODirection |= MFIO_PIN_LCD_POWER;
    MFIOOutput |= MFIO_PIN_LCD_POWER;

}

void turn_off_helio_lcd(void)
{
    VidCtrl2 = ((LCD_VIDRATE << VIDRATE_SHIFT) & VIDRATE_MASK) |
	    (((FB_X_RES / 4 - 1) << HORZVAL_SHIFT) & HORZVAL_MASK) |
	    ((FB_Y_RES + 30) & LINEVAL_MASK);

    /* Freeze after end of frame */
    while( (VidCtrl1 >> LINECNT_SHIFT) < 161 );
    VidCtrl1 |= ENFREEZEFRAME;
    while( (VidCtrl1 >> LINECNT_SHIFT) != 0 );

    VidCtrl1 &= ~ENVID;         /* Disable Video Logic */
    ClockControl &= ~CLK_ENVIDCLK;
    VidCtrl1 &= ~DISPON;             /* Display off */

    MFIOSelect |= MFIO_PIN_LCD_POWER;
    MFIODirection |= MFIO_PIN_LCD_POWER;
    MFIOOutput &= ~MFIO_PIN_LCD_POWER;

    VidCtrl2 = ((LCD_VIDRATE << VIDRATE_SHIFT) & VIDRATE_MASK) |
	    (((FB_X_RES / 4 - 1) << HORZVAL_SHIFT) & HORZVAL_MASK) |
	    ((FB_Y_RES - 1) & LINEVAL_MASK);

    VidCtrl1 &= ~ENFREEZEFRAME;
}

#endif /* CONFIG_VTECH_HELIO */


    /*
     *  Initialisation
     */

int __init r3912fb_init(void)
{
#if defined(CONFIG_SHARP_MOBILON) || defined(CONFIG_COMPAQ) || defined(CONFIG_PHILIPS_VELO)
    /*
     * TODO: check VidCtrl1 for monochrome/greyscale LCDs
     * and set up fb_info and friends accordingly
     */

#if 0
    if (!(VidCtrl1 & DISP8))
	return -EINVAL;
#endif

    /*
     * videomemory has been set in arch/mips/r39xx/prom/memory.c
     */
    /*
     * Stop video at end of frame
     */
#ifndef CONFIG_VTECH_HELIO_EMULATOR
    VidCtrl1 |= ENFREEZEFRAME;
    IntClear1 |= INT1_LCDINT;
    /*
     * Wait for end of frame: Poll Interrupt Status Register 1
     */
    while (!(IntStatus1 & INT1_LCDINT))
	;
#endif

    VidCtrl1 &= ~(DISPON | ENVID);
    udelay(100);
    VidCtrl3 = PHYSADDR(videomemory);
    /*
     * As a consequence of this DFVAL and FRAMEMASKVAL are set to 0.
     * This may not work on all R3912/Poseidons.
     */
    VidCtrl4 = PHYSADDR(videomemory + videomemorysize + 1) & VIDBASELO_MASK;
    VidCtrl1 &= ~(DFMODE | ENFREEZEFRAME);
    udelay(100);
    VidCtrl1 |= DISPON;
    VidCtrl1 |= ENVID;

#elif defined(CONFIG_VTECH_HELIO)

    VidCtrl1 = (0x6 << BAUDVAL_SHIFT) |   /* 4.5MHz LCD Clock */
	    (0x4 << VIDDONEVAL_SHIFT) |   /* Delay for VIDDONE signal */
#ifdef CONFIG_FB_R3912_FORCE_MONO
	    (0x0 << BITSEL_SHIFT) |       /* 1bpp monochrome mode */
#else
	    (0x2 << BITSEL_SHIFT) |       /* 4bpp greyscale mode */
#endif
	    DISPON;             /* Display on, but disable video logic */

    VidCtrl2 = ((LCD_VIDRATE << VIDRATE_SHIFT) & VIDRATE_MASK) |
	    (((FB_X_RES / 4 - 1) << HORZVAL_SHIFT) & HORZVAL_MASK) |
	    ((FB_Y_RES - 1) & LINEVAL_MASK);

    VidCtrl3 = PHYSADDR(videomemory);
    /*
     * As a consequence of this DFVAL and FRAMEMASKVAL are set to 0.
     * This may not work on all R3912/Poseidons.
     */
    VidCtrl4 = PHYSADDR(videomemory + videomemorysize + 1) & VIDBASELO_MASK;

    /* Setup LUT tables */
    VidCtrl7 = 0xFA50;
    VidCtrl8 = 0x7DA;
    VidCtrl9 = 0x7DBEA5A5;
    VidCtrl10 = 0x7DFBE;
    VidCtrl11 = 0x7A5AD;
    VidCtrl12 = 0xFBFDFE7;
    VidCtrl13 = 0x7B5ADEF;
    VidCtrl14 = 0xB9DC663;

    turn_on_helio_lcd();

#endif 

    strcpy(fb_info.modename, r3912fb_name);
    fb_info.changevar = NULL;
    fb_info.node = -1;
    fb_info.fbops = &r3912fb_ops;
    fb_info.disp = &disp;
    fb_info.switch_con = &r3912fbcon_switch;
    fb_info.updatevar = &r3912fbcon_updatevar;
    fb_info.blank = &r3912fbcon_blank;
    fb_info.flags = FBINFO_FLAG_DEFAULT;

    r3912fb_set_var(&r3912fb_default, -1, &fb_info);

    if (register_framebuffer(&fb_info) < 0)
	return -EINVAL;

    printk(KERN_INFO "fb%d: R3912 frame buffer device, using %ldK of video memory\n",
	   GET_FB_IDX(fb_info.node), videomemorysize>>10);
    return 0;
}


static int r3912fbcon_switch(int con, struct fb_info *info)
{
    /* Do we have to save the colormap? */
    if (fb_display[currcon].cmap.len)
	fb_get_cmap(&fb_display[currcon].cmap, 1, r3912fb_getcolreg, info);

    currcon = con;
    /* Install new colormap */
    do_install_cmap(con, info);
    return 0;
}

    /*
     *  Update the `var' structure (called by fbcon.c)
     */

static int r3912fbcon_updatevar(int con, struct fb_info *info)
{
    /* Nothing */
    return 0;
}

    /*
     *  Blank the display.
     */

static void r3912fbcon_blank(int blank, struct fb_info *info)
{
    /* Nothing */
}

static u_long get_line_length(int xres_virtual, int bpp)
{
    u_long length;
    
    length = (xres_virtual * bpp + 7) >> 3;
    return(length);
}

static void r3912fb_encode_fix(struct fb_fix_screeninfo *fix,
			   struct fb_var_screeninfo *var)
{
    memset(fix, 0, sizeof(struct fb_fix_screeninfo));
    strcpy(fix->id, r3912fb_name);
    fix->smem_start = videomemory;
    fix->smem_len = videomemorysize;
    fix->type = FB_TYPE_PACKED_PIXELS;
    fix->type_aux = 0;
    switch (var->bits_per_pixel) {
    case 1:
	    fix->visual = FB_VISUAL_MONO01;
	    break;
    case 2:
    case 4:
            fix->visual = FB_VISUAL_STATIC_PSEUDOCOLOR;
    case 8:
	    fix->visual = FB_VISUAL_TRUECOLOR;
	    break;
    }
    fix->ywrapstep = 1;
    fix->xpanstep = 1;
    fix->ypanstep = 1;
    fix->line_length = get_line_length(var->xres_virtual, var->bits_per_pixel);
}

static void set_color_bitfields(struct fb_var_screeninfo *var)
{
    switch (var->bits_per_pixel) {
/* FIXME
	case 1:
	case 2:
	case 4:
*/	case 8:
	    /* RGB 332 */
	    var->red.offset = 5;
	    var->red.length = 3;
	    var->green.offset = 2;
	    var->green.length = 3;
	    var->blue.offset = 0;
	    var->blue.length = 2;
	    var->transp.offset = 0;
	    var->transp.length = 0;
	    break;
    }
    var->red.msb_right = 0;
    var->green.msb_right = 0;
    var->blue.msb_right = 0;
    var->transp.msb_right = 0;
}


    /*
     *  Read a single color register and split it into
     *  colors/transparent. Return != 0 for invalid regno.
     */

static int r3912fb_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
                         u_int *transp, struct fb_info *info)
{
    if (regno > 255)
	return 1;
#if FB_IS_GREY
    {
            u_int grey;
            
            grey = regno * 255 / 15;
#if FB_IS_INVERSE
            grey ^= 255;
#endif
            grey |= grey << 8;
            *red = grey;
            *green = grey;
            *blue = grey;
    }
#else
    *red = (palette[regno].red<<8) | palette[regno].red;
    *green = (palette[regno].green<<8) | palette[regno].green;
    *blue = (palette[regno].blue<<8) | palette[regno].blue;
#endif
    *transp = 0;
    return 0;
}


    /*
     *  Set a single color register. The values supplied are already
     *  rounded down to the hardware's capabilities (according to the
     *  entries in the var structure). Return != 0 for invalid regno.
     */

static int r3912fb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
                         u_int transp, struct fb_info *info)
{
    if (regno > 255)
	return 1;

#ifdef FBCON_HAS_CFB8
    if( regno < 16 ) {
	// maintain index to rgb map for fbcon (accessed through dispsw_data)
	fbcon_cmap.cfb8[regno] =
	 ((red & 0xe000) >> 8)
	 | ((green & 0xe000) >> 11)
	 | ((blue & 0xc000) >> 14);
    }
#endif 

    red >>= 8;
    green >>= 8;
    blue >>= 8;
    palette[regno].red = red;
    palette[regno].green = green;
    palette[regno].blue = blue;
    return 0;
}


static void do_install_cmap(int con, struct fb_info *info)
{
    if (con != currcon)
	return;
    if (fb_display[con].cmap.len)
	fb_set_cmap(&fb_display[con].cmap, 1, r3912fb_setcolreg, info);
    else
	fb_set_cmap(fb_default_cmap(1<<fb_display[con].var.bits_per_pixel), 1,
		    r3912fb_setcolreg, info);
}


#ifdef MODULE
int init_module(void)
{
    return r3912fb_init();
}

void cleanup_module(void)
{
    unregister_framebuffer(&fb_info);
}

#endif /* MODULE */
