/*
 * kernel/axmon.c  -- AXE Process Scheduling Monitor
 *
 * Copyright (C) 2001 AXE, Inc.
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#ifdef CONFIG_CPU_TMPR3922
#include <asm/tx3920.h>
#endif
#include "axmon.h"

static int atoi(const char *p)
{
	int n = 0;
	while (*p && '0' <= *p && *p <= '9') {
		n *= 10;
		n += *p - '0';
		p++;
	}
	return (n);
}

#ifdef CONFIG_AXE_MONITOR_SCHED_LOG
#define AXMON_TMPBUF_SIZE	0x60000
static char axmon_tmpbuf[AXMON_TMPBUF_SIZE+256];

#define AXMON_USE_TRACE		0
#define AXMON_USE_MISC		1
#define AXMON_USE_TRACE_DELAY	2
#define AXMON_USE_PSWITCH	3
static int axmon_tmpbuf_last_used = -1;
static int axmon_tmpbuf_len = 0;

struct axmon_pf_hist axmon_pf_hist[AXMON_HIST_NUM];
struct axmon_pf_hist axmon_pf_hist_snap[AXMON_HIST_NUM];
int axmon_pf_hist_ptr;
unsigned long axmon_pf_last_rotate;

unsigned long axmon_pf_counter_max, axmon_pf_counter_pid;

int axmon_misc_enable = 0;

void axmon_pf_snap(void)
{
    int i, p;
    unsigned long tmp;

    do{
	p = axmon_pf_hist_ptr;
	for( i=0; i<AXMON_HIST_NUM; i++ )
	    axmon_pf_hist_snap[i] = axmon_pf_hist[(p+i+1)%AXMON_HIST_NUM];
    }while(p != axmon_pf_hist_ptr);
    tmp = jiffies;
    axmon_pf_hist_snap[AXMON_HIST_NUM - 1].tick_count =
	tmp - axmon_pf_last_rotate;
    axmon_pf_last_rotate = tmp;

    for( i=0; i<AXMON_HIST_NUM; i++ )
	axmon_pf_rotate(0);
}

void axmon_pf_print(void)
{
    int i, j;

    printk("axmon_pf_print:\n");
    axmon_pf_snap();
    for( i=0; i<AXMON_HIST_NUM; i++ ){
	printk("[n=%3d,tick=%4ld]", i, axmon_pf_hist_snap[i].tick_count);
	for( j=0; j<AXMON_TICK_RANGE_MAX+1; j++ ){
	    if( j != 0 && j % 16 == 0 )
		printk("\n                 ");
	    printk(" %4ld", axmon_pf_hist_snap[i].num[j]);
	}
	printk("\n");
    }
    printk(" counter(max) = %ld, pid = %ld\n", axmon_pf_counter_max,
	   axmon_pf_counter_pid);
}

/* ******************************************************************** */
#define AXMON_TRACE_PROC_MAX 2
#define AXMON_TRACE_ELEM_LEN 0x4000
#ifdef CONFIG_CPU_TMPR3922
#define AXMON_TRACE_HIST_STEP 1000	/* 1msec */
#define AXMON_TRACE_HIST_MAX 320
#else
#define AXMON_TRACE_HIST_MAX 32
#endif

struct ax_task_trace_elem{
    unsigned long jiffies;
#ifdef CONFIG_CPU_TMPR3922
    unsigned rtclo;
#endif
    int state;
    int counter;
    pid_t running_pid;
};

struct ax_task_trace_delay{
    unsigned long jiffies;
    int val;
#ifdef CONFIG_CPU_TMPR3922
    unsigned rtclo;
    unsigned val_rtclo;
#endif
};

struct ax_delay_pack{
    struct ax_task_trace_delay delay[AXMON_TRACE_ELEM_LEN];
    int l_head;
    int l_tail;
    int hist[AXMON_TRACE_HIST_MAX+1];
};

struct ax_task_trace{
    struct task_struct *task;
    struct ax_task_trace_elem elems[AXMON_TRACE_ELEM_LEN];
    int head;
    int tail;
    struct ax_delay_pack dp;
};

struct ax_sched_trace{
    struct ax_delay_pack dp;
};

static struct ax_task_trace axmon_trace[AXMON_TRACE_PROC_MAX];
static struct ax_sched_trace axmon_sched_trace;
static int axmon_sched_trace_flag=0;
static int axmon_trace_ptr;

void axmon_set_trace(struct task_struct *tgt, int ev)
{
    int i, ptr;
    struct task_struct *p;
    unsigned flags;
#ifdef CONFIG_CPU_TMPR3922
    unsigned rtclo = RTClow;
#endif

    axmon_sched_trace_flag = 1;
    local_irq_save(flags);
    for( i=0; i<AXMON_TRACE_PROC_MAX; i++ ){
	p = axmon_trace[i].task;
	if( p != NULL && p == tgt ){
	    ptr = axmon_trace[i].tail;
	    axmon_trace[i].elems[ptr].jiffies = jiffies;
#ifdef CONFIG_CPU_TMPR3922
	    axmon_trace[i].elems[ptr].rtclo = rtclo;
#endif
	    axmon_trace[i].elems[ptr].state   = ev;
	    axmon_trace[i].elems[ptr].counter = p->counter;
	    axmon_trace[i].elems[ptr].running_pid
		= current->pid;
	    axmon_trace[i].tail = (axmon_trace[i].tail + 1) %
		AXMON_TRACE_ELEM_LEN;
	    if( axmon_trace[i].head == axmon_trace[i].tail )
		axmon_trace[i].head = (axmon_trace[i].head + 1) %
		    AXMON_TRACE_ELEM_LEN;
	}
    }
    local_irq_restore(flags);
}

#ifdef CONFIG_CPU_TMPR3922
static void
axmon_check_trace_torun(int i, int ptr, unsigned jif, unsigned rtclo)
#else
static void
axmon_check_trace_torun(int i, int ptr, unsigned jif)
#endif
{
    int cat;
    if( axmon_trace[i].head != ptr ){
	int pptr = (ptr + AXMON_TRACE_ELEM_LEN - 1)
	    % AXMON_TRACE_ELEM_LEN;
	int l_tail = axmon_trace[i].dp.l_tail;

	if( axmon_trace[i].elems[pptr].state == AXMON_TRACE_EV_TO_READY ||
	    axmon_trace[i].elems[pptr].state == AXMON_TRACE_EV_RUN2READY ){
	    axmon_trace[i].dp.delay[l_tail].jiffies = jif;
	    axmon_trace[i].dp.delay[l_tail].val = jif -
		axmon_trace[i].elems[pptr].jiffies;
#ifdef CONFIG_CPU_TMPR3922
	    axmon_trace[i].dp.delay[l_tail].rtclo = rtclo;
	    axmon_trace[i].dp.delay[l_tail].val_rtclo = rtclo - 
		axmon_trace[i].elems[pptr].rtclo;

	    cat = (axmon_trace[i].dp.delay[l_tail].val_rtclo *
		   AXMON_TRACE_HIST_STEP) / RTC_PERSEC_COUNT;
	    if( cat >= AXMON_TRACE_HIST_MAX )
		cat = AXMON_TRACE_HIST_MAX;
#else
	    cat = axmon_trace[i].dp.delay[l_tail].val;
	    if( cat >= AXMON_TRACE_HIST_MAX )
		cat = AXMON_TRACE_HIST_MAX;
#endif
	    axmon_trace[i].dp.hist[cat]++;

	    axmon_trace[i].dp.l_tail = (l_tail + 1) % AXMON_TRACE_ELEM_LEN;
	    if( axmon_trace[i].dp.l_head == axmon_trace[i].dp.l_tail )
		axmon_trace[i].dp.l_head = (axmon_trace[i].dp.l_head + 1) % AXMON_TRACE_ELEM_LEN;
	}
    }
}

#ifdef CONFIG_CPU_TMPR3922
static void
axmon_check_trace_sched(unsigned jif, unsigned rtclo)
{
    static unsigned last_rtclo=0;
#else
static void
axmon_check_trace_sched(unsigned jif)
{
#endif
    static unsigned last_jif=0;
    int cat, l_tail;

    if( last_jif == 0 && last_rtclo == 0 ){
	last_jif = jif;
	last_rtclo = rtclo;
	axmon_sched_trace.dp.l_tail = 0;
	axmon_sched_trace.dp.l_head = 0;
    }

    l_tail = axmon_sched_trace.dp.l_tail;

    axmon_sched_trace.dp.delay[l_tail].jiffies = jif;
    axmon_sched_trace.dp.delay[l_tail].val = jif - last_jif;
#ifdef CONFIG_CPU_TMPR3922
    axmon_sched_trace.dp.delay[l_tail].rtclo = rtclo;
    axmon_sched_trace.dp.delay[l_tail].val_rtclo = rtclo - last_rtclo;

    cat = (axmon_sched_trace.dp.delay[l_tail].val_rtclo *
		   AXMON_TRACE_HIST_STEP) / RTC_PERSEC_COUNT;
    if( cat >= AXMON_TRACE_HIST_MAX )
	cat = AXMON_TRACE_HIST_MAX;
#else
    cat = axmon_sched_trace.dp.delay[l_tail].val;
    if( cat >= AXMON_TRACE_HIST_MAX )
	cat = AXMON_TRACE_HIST_MAX;
#endif
    axmon_sched_trace.dp.hist[cat]++;

    axmon_sched_trace.dp.l_tail = (l_tail + 1) % AXMON_TRACE_ELEM_LEN;
    if( axmon_sched_trace.dp.l_head == axmon_sched_trace.dp.l_tail )
	axmon_sched_trace.dp.l_head = (axmon_sched_trace.dp.l_head + 1) % AXMON_TRACE_ELEM_LEN;
    last_jif = jif;
    last_rtclo = rtclo;
}

void axmon_check_trace(struct task_struct *prev, struct task_struct *next)
{
    int i, ptr;
    struct task_struct *p;
    unsigned jif=jiffies;
#ifdef CONFIG_CPU_TMPR3922
    unsigned rtclo = RTClow;
#endif

    for( i=0; i<AXMON_TRACE_PROC_MAX; i++ ){
	p = axmon_trace[i].task;
	if( p ){
	    ptr = axmon_trace[i].tail;
	    if( prev == p && prev != next ){
		axmon_trace[i].elems[ptr].jiffies = jif;
#ifdef CONFIG_CPU_TMPR3922
		axmon_trace[i].elems[ptr].rtclo = rtclo;
#endif
		if( p->state == TASK_RUNNING )
		    axmon_trace[i].elems[ptr].state   = AXMON_TRACE_EV_RUN2READY;
		else
		    axmon_trace[i].elems[ptr].state   = AXMON_TRACE_EV_RUN2WAIT;
		axmon_trace[i].elems[ptr].counter = p->counter;
		axmon_trace[i].elems[ptr].running_pid
		    = next->pid;
		axmon_trace[i].tail = (axmon_trace[i].tail + 1) %
		    AXMON_TRACE_ELEM_LEN;
		if( axmon_trace[i].head == axmon_trace[i].tail )
		    axmon_trace[i].head = (axmon_trace[i].head + 1) %
			AXMON_TRACE_ELEM_LEN;
	    }else if( next == p && prev != next ){
#ifdef CONFIG_CPU_TMPR3922
		axmon_check_trace_torun(i, ptr, jif, rtclo);
		axmon_trace[i].elems[ptr].rtclo = rtclo;
#else
		axmon_check_trace_torun(i, ptr, jif);
#endif
		axmon_trace[i].elems[ptr].jiffies = jif;
		axmon_trace[i].elems[ptr].state   = AXMON_TRACE_EV_TO_RUN;
		axmon_trace[i].elems[ptr].counter = p->counter;
		axmon_trace[i].elems[ptr].running_pid = next->pid;
		axmon_trace[i].tail = (axmon_trace[i].tail + 1) %
		    AXMON_TRACE_ELEM_LEN;
		if( axmon_trace[i].head == axmon_trace[i].tail )
		    axmon_trace[i].head = (axmon_trace[i].head + 1) %
			AXMON_TRACE_ELEM_LEN;
	    }
	}
    }

    if( axmon_sched_trace_flag ){
#ifdef CONFIG_CPU_TMPR3922
	axmon_check_trace_sched(jif, rtclo);
#else
	axmon_check_trace_sched(jif);
#endif
    }
}

/*************************************************************************/
#define AXMON_PSWITCH_VECSIZ 128
static int pswitch_ptr = 0;
static int pswitch_len = 0;
static struct code_trace_box{
    unsigned long time;
    struct task_struct *prev;
    struct task_struct *next;
}pswitch_vec[AXMON_PSWITCH_VECSIZ];

void axmon_pswitch(struct task_struct *prev, struct task_struct *next)
{
    unsigned flags;

    local_irq_save(flags);
    pswitch_vec[pswitch_ptr].time = jiffies;
    pswitch_vec[pswitch_ptr].prev = prev;
    pswitch_vec[pswitch_ptr].next = next;
    pswitch_ptr = (pswitch_ptr + 1) % AXMON_PSWITCH_VECSIZ;
    if( pswitch_len < AXMON_PSWITCH_VECSIZ )
	pswitch_len++;
    local_irq_restore(flags);
}

/*************************************************************************/

static int __init axmon_sched_log_init(void)
{
    int i;
#ifdef CONFIG_AXE_MONITOR_SCHED_LOG
    for( i=0; i<AXMON_TRACE_PROC_MAX; i++ ){
	axmon_trace[i].task = NULL;
	axmon_trace[i].head = 0;
	axmon_trace[i].tail= 0;
	axmon_trace[i].dp.l_head = 0;
	axmon_trace[i].dp.l_tail= 0;
    }
#endif /* CONFIG_AXE_MONITOR_SCHED_LOG */
    return 0;
}
__initcall(axmon_sched_log_init);
#endif /* CONFIG_AXE_MONITOR_SCHED_LOG */

#ifdef CONFIG_PROC_FS
#ifdef CONFIG_AXE_MONITOR_SCHED_LOG
static inline struct task_struct *find_process_by_pid(pid_t pid)
{
	struct task_struct *tsk = current;

	if (pid)
		tsk = find_task_by_pid(pid);
	return tsk;
}

#if 0
static int axmon_setstrbuf(char *dist, int distn, char *src, int srcn)
{
    int n;

    if( distn < srcn + 1)
	n = distn - 1;
    else
	n = srcn;
    if( n == 0 ){
	if( distn > 0 )
	    dist[0] = '\0';
	return 0;
    }
    memcpy(dist, src, n);
    dist[n] = '\0';
    return n;
}
#endif

static int misc_setbuf(char *buf, int bufsiz)
{
    int i, j, tn;

    tn = 0;

    axmon_pf_snap();
    tn += sprintf(buf+tn, "               ");
    for( j=0; j<AXMON_TICK_RANGE_MAX; j++ )
	tn += sprintf(buf+tn, " %4d", j);
    tn += sprintf(buf+tn, " >=%d\n", j);
	
    for( i=0; i<AXMON_HIST_NUM; i++ ){
	tn += sprintf(buf+tn, "%3d[tick=%4ld]|", i, axmon_pf_hist_snap[i].tick_count);
	for( j=0; j<AXMON_TICK_RANGE_MAX+1; j++ ){
	    if( j != 0 && j % 16 == 0 )
		tn += sprintf(buf+tn, "\n                 ");
	    tn += sprintf(buf+tn, " %4ld", axmon_pf_hist_snap[i].num[j]);
	}
	tn += sprintf(buf+tn, "\n");
    }
    if( axmon_misc_enable )
	tn += sprintf(buf+tn, " period=%d(sec), counter= (max=%ld,pid=%ld), enable=%d\n",
		      AXMON_PERIOD_SEC, axmon_pf_counter_max,
		      axmon_pf_counter_pid,
		      axmon_misc_enable);
    else
	tn += sprintf(buf+tn, " period=%d(sec), counter= (max=%d,pid=%d), enable=%d\n",
		      AXMON_PERIOD_SEC, -1, -1,
		      axmon_misc_enable);

    axmon_pf_counter_max = -1;
    axmon_pf_counter_pid = -1;
    if( tn + 1 > bufsiz ){
	printk("axmon: Critical error - buffer overflow!!\n");
	buf[bufsiz-1] = '\0';
	return bufsiz - 1;
    }
    return tn;
}

static ssize_t
axmon_misc_read(struct file *file, char *buf, size_t len,
			 loff_t *ppos)
{
    size_t rlen;

    if (*ppos != 0)
	return 0;

    rlen = misc_setbuf(axmon_tmpbuf, AXMON_TMPBUF_SIZE);

    axmon_tmpbuf_last_used = AXMON_USE_MISC;
    axmon_tmpbuf_len = rlen;
    
    if (len < rlen)
	return -EIO;
    if (copy_to_user(buf, axmon_tmpbuf, rlen))
	return -EFAULT;
    *ppos += rlen;
    return rlen;
}

static ssize_t
axmon_misc_write(struct file *file, const char *buf, size_t len, loff_t *ppos)
{
    static char tmpbuf[64];
    int i;

    if (*ppos != 0)
	return -EIO;

    if (len > 63)
	return -EIO;

    if (copy_from_user(tmpbuf, buf, len))
	return -EFAULT;
    tmpbuf[len] = '\0';

    i = 0;
    if (tmpbuf[i] >= '0' && tmpbuf[i] <= '9')	
	axmon_misc_enable = atoi(tmpbuf);
    else
	return -EIO;

    if( !axmon_misc_enable ){
	for( i=0; i<AXMON_HIST_NUM; i++ )
	    axmon_pf_rotate(0);
    }

    *ppos += len;
    return len;
}

static int axmon_trace_setbuf(int n, char *buf, int bufsiz)
{
    int tn, ptr, tail, i;
    unsigned flags;

    tn = 0;

    local_irq_save(flags);
    if( axmon_trace[n].task )
	tn += sprintf(buf+tn, "((pid %d)", axmon_trace[n].task->pid);
    else{
	tn += sprintf(buf+tn, "((pid nil))\n");
	local_irq_restore(flags);
	return tn;
    }

    tail = axmon_trace[n].tail;
    for( ptr = axmon_trace[n].head, i = 0;
	 ptr != tail;
	 ptr = (ptr + 1) % AXMON_TRACE_ELEM_LEN, i++ ){
	if( i % 5 == 0 )
	    tn += sprintf(buf+tn, "\n");
#ifdef CONFIG_CPU_TMPR3922
	tn += sprintf(buf+tn,
		      " (%5d %d %2d %4d)",
		      axmon_trace[n].elems[ptr].rtclo,
		      axmon_trace[n].elems[ptr].state,
		      axmon_trace[n].elems[ptr].counter,
		      (int)axmon_trace[n].elems[ptr].running_pid);
#else
	tn += sprintf(buf+tn,
		      " (%5ld %d %2d %4d)",
		      axmon_trace[n].elems[ptr].jiffies,
		      axmon_trace[n].elems[ptr].state,
		      axmon_trace[n].elems[ptr].counter,
		      axmon_trace[n].elems[ptr].running_pid);
#endif
	if( tn + 1 > bufsiz ){
	    printk("axmon: Critical error - buffer overflow!!\n");
	    axmon_trace[n].head = ptr;
	    local_irq_restore(flags);
	    buf[bufsiz-1] = '\0';
	    return bufsiz - 1;
	}
    }
    axmon_trace[n].head = ptr;
    local_irq_restore(flags);
    
    tn += sprintf(buf+tn, ")\n");

    if( tn + 1 > bufsiz ){
	printk("axmon: Critical error - buffer overflow!!\n");
	buf[bufsiz-1] = '\0';
	return bufsiz - 1;
    }
    return tn;
}

static ssize_t
axmon_trace_read(struct file *file, char *buf, size_t len, loff_t *ppos)
{
	size_t rlen;
	int i;

	if (*ppos != 0){
	    if( axmon_tmpbuf_last_used != AXMON_USE_TRACE ||
		axmon_tmpbuf_len <= *ppos )
		return 0;
	    rlen = axmon_tmpbuf_len - *ppos;
	}else{
	    rlen = 0;
	    for( i=0; i<AXMON_TRACE_PROC_MAX; i++ )
		rlen += axmon_trace_setbuf(i, axmon_tmpbuf+rlen,
					   AXMON_TMPBUF_SIZE-rlen);

	    axmon_tmpbuf_last_used = AXMON_USE_TRACE;
	    axmon_tmpbuf_len = rlen;
	}
	if (len < rlen)
	    rlen = len;
	if (copy_to_user(buf, axmon_tmpbuf + *ppos, rlen))
	    return -EFAULT;
	*ppos += rlen;
	return rlen;
}

static ssize_t
axmon_trace_write(struct file *file, const char *buf, size_t len, loff_t *ppos)
{
    static char tmpbuf[16];
    pid_t pid;
    struct task_struct *p;

    if (*ppos != 0)
	return -EIO;

    if (len > 15)
	return -EIO;

    if (copy_from_user(tmpbuf, buf, len))
	return -EFAULT;
    tmpbuf[len] = '\0';

    if (tmpbuf[0] >= '0' && tmpbuf[0] <= '9')	
	pid = atoi(tmpbuf);
    else
	return -EIO;

    if( pid < 0 )
	return -EIO;

    if( pid == 0 ){
	axmon_trace[axmon_trace_ptr].task = NULL;
    }else{
	p = find_process_by_pid(pid);
	if( p == NULL )
	    return -EIO;
	axmon_trace[axmon_trace_ptr].task = p;
    }
    axmon_trace[axmon_trace_ptr].head = 0;
    axmon_trace[axmon_trace_ptr].tail = 0;
    axmon_trace_ptr = (axmon_trace_ptr+1) % AXMON_TRACE_PROC_MAX;

    *ppos += len;
    return len;
}

static int
axmon_trace_delay_setbuf_0(struct ax_delay_pack *dpp, char *buf, int bufsiz)
{
    int tn, ptr, tail, i;

    tn = 0;
    tail = dpp->l_tail;
    for( ptr = dpp->l_head, i = 0; ptr != tail;
	 ptr = (ptr + 1) % AXMON_TRACE_ELEM_LEN, i++ ){
	if( i % 7 == 0 )
	    tn += sprintf(buf+tn, "\n");
#ifdef CONFIG_CPU_TMPR3922
	tn += sprintf(buf+tn,
		      " (%5d %5d)",
		      dpp->delay[ptr].rtclo,
		      dpp->delay[ptr].val_rtclo);
#else
	tn += sprintf(buf+tn,
		      " (%5ld %2d)",
		      dpp->delay[ptr].jiffies,
		      dpp->delay[ptr].val);
#endif
	if( tn + 1 > bufsiz ){
	    printk("axmon: Critical error - buffer overflow!!\n");
	    dpp->l_head = ptr;
	    buf[bufsiz-1] = '\0';
	    return bufsiz - 1;
	    return tn;
	}
    }
    dpp->l_head = ptr;
    return tn;
}

static int
axmon_trace_hist_setbuf_0(struct ax_delay_pack *dpp, char *buf, int bufsiz)
{
    int tn, tail, i;

    tn = 0;
    tail = dpp->l_tail;
    for( i = 0; i<=AXMON_TRACE_HIST_MAX; i++ ){
	tn += sprintf(buf+tn, " (%2d %2d)", i, dpp->hist[i]);
	dpp->hist[i] = 0;
	if( tn + 1 > bufsiz ){
	    printk("axmon: Critical error - buffer overflow!!\n");
	    buf[bufsiz-1] = '\0';
	    return bufsiz - 1;
	}
    }
    return tn;
}

static int axmon_trace_delay_setbuf(int n, char *buf, int bufsiz)
{
    int tn;
    unsigned flags;

    tn = 0;

    local_irq_save(flags);
    if( axmon_trace[n].task )
	tn += sprintf(buf+tn, "(delay (pid %d)", axmon_trace[n].task->pid);
    else{
	tn += sprintf(buf+tn, "(delay (pid nil))\n");
	local_irq_restore(flags);
	return tn;
    }

    tn += axmon_trace_delay_setbuf_0(&axmon_trace[n].dp, buf+tn, bufsiz-tn);

    local_irq_restore(flags);
    
    tn += sprintf(buf+tn, ")\n");

    if( tn + 1 > bufsiz ){
	printk("axmon: Critical error - buffer overflow!!\n");
	buf[bufsiz-1] = '\0';
	return bufsiz - 1;
    }
    return tn;
}

static int axmon_trace_hist_setbuf(int n, char *buf, int bufsiz)
{
    int tn;
    unsigned flags;

    tn = 0;

    local_irq_save(flags);
    if( axmon_trace[n].task )
	tn += sprintf(buf+tn, "(histgram (pid %d)", axmon_trace[n].task->pid);
    else{
	tn += sprintf(buf+tn, "(histgram (pid nil))\n");
	local_irq_restore(flags);
	return tn;
    }

    tn += axmon_trace_hist_setbuf_0(&axmon_trace[n].dp, buf+tn, bufsiz-tn);

    local_irq_restore(flags);
    
    tn += sprintf(buf+tn, ")\n");

    if( tn + 1 > bufsiz ){
	printk("axmon: Critical error - buffer overflow!!\n");
	return tn;
    }
    return tn;
}

static int axmon_sched_trace_setbuf(char *buf, int bufsiz)
{
    int tn;
    unsigned flags;

    tn = 0;

    local_irq_save(flags);
    if( axmon_sched_trace_flag )
	tn += sprintf(buf+tn, "(delay (sched t)");
    else{
	tn += sprintf(buf+tn, "(delay (sched nil))\n");
	local_irq_restore(flags);
	return tn;
    }
    tn += axmon_trace_delay_setbuf_0(&axmon_sched_trace.dp, buf+tn, bufsiz-tn);
    tn += sprintf(buf+tn, ")\n");

    if( axmon_sched_trace_flag )
	tn += sprintf(buf+tn, "(histgram (sched t)");
    else{
	tn += sprintf(buf+tn, "(histgram (sched nil))\n");
	local_irq_restore(flags);
	return tn;
    }
    tn += axmon_trace_hist_setbuf_0(&axmon_sched_trace.dp, buf+tn, bufsiz-tn);
    tn += sprintf(buf+tn, ")\n");
    local_irq_restore(flags);

    if( tn + 1 > bufsiz ){
	printk("axmon: Critical error - buffer overflow!!\n");
	return tn;
    }
    return tn;
}

static ssize_t
axmon_trace_delay_read(struct file *file, char *buf, size_t len,
			 loff_t *ppos)
{
	size_t rlen;
	int i;

	if (*ppos != 0){
	    if( axmon_tmpbuf_last_used != AXMON_USE_TRACE_DELAY ||
		axmon_tmpbuf_len <= *ppos )
		return 0;
	    rlen = axmon_tmpbuf_len - *ppos;
	}else{
#ifdef CONFIG_CPU_TMPR3922
	    rlen = sprintf(axmon_tmpbuf,"period: %dHz\nhist: %dmsec\n",
			   RTC_PERSEC_COUNT, 1);
#else
	    rlen = sprintf(axmon_tmpbuf,"period: %dHz\nhist: %dmsec\n",
			   HZ, 10);
#endif
	    for( i=0; i<AXMON_TRACE_PROC_MAX; i++ ){
		rlen += axmon_trace_delay_setbuf(i, axmon_tmpbuf+rlen,
						   AXMON_TMPBUF_SIZE - rlen);
		rlen += axmon_trace_hist_setbuf(i, axmon_tmpbuf+rlen,
						   AXMON_TMPBUF_SIZE - rlen);
	    }

	    rlen += axmon_sched_trace_setbuf(axmon_tmpbuf+rlen,
					     AXMON_TMPBUF_SIZE - rlen);

	    axmon_tmpbuf_last_used = AXMON_USE_TRACE_DELAY;
	    axmon_tmpbuf_len = rlen;
	}
	if (len < rlen)
	    rlen = len;
	if (copy_to_user(buf, axmon_tmpbuf + *ppos, rlen))
	    return -EFAULT;
	*ppos += rlen;
	return rlen;
}


static int axmon_pswitch_setbuf(int n, char *buf, int bufsiz)
{
    int tn, head, p, prev, i;
    unsigned flags;

    if( pswitch_len <= 1 )
	return 0;

    tn = 0;
    local_irq_save(flags);
    tn += sprintf(buf+tn, "; pid alive-time jiffies command\n");

    head = pswitch_ptr + AXMON_PSWITCH_VECSIZ - pswitch_len;
    prev = -1;
    for( i=0; i<pswitch_len; i++ ){
	p = (head + i) % AXMON_PSWITCH_VECSIZ;
	if( pswitch_vec[p].next == NULL ){
	    tn += sprintf(buf+tn, "; NULL %6ld\n", 
			  pswitch_vec[p].time);
	    if( tn + 1 > bufsiz ){
		printk("axmon: Critical error - buffer overflow!!\n");
		local_irq_restore(flags);
		buf[bufsiz-1] = '\0';
		return bufsiz - 1;
	    }
	    continue;
	}
	if( prev != -1 ){
	    tn += sprintf(buf+tn, "%-5d %2ld %6ld %s\n", 
			  pswitch_vec[p].prev->pid, 
			  pswitch_vec[p].time - pswitch_vec[prev].time,
			  pswitch_vec[p].time,
			  pswitch_vec[p].prev->comm);
	    pswitch_vec[prev].prev = 0;
	    pswitch_vec[prev].next = 0;
	    pswitch_vec[prev].time = 0;
	    if( tn + 1 > bufsiz ){
		printk("axmon: Critical error - buffer overflow!!\n");
		local_irq_restore(flags);
		buf[bufsiz-1] = '\0';
		return bufsiz - 1;
	    }
	}
	prev = p;
    }
    pswitch_len = 1;
    local_irq_restore(flags);
    
    tn += sprintf(buf+tn, "\n");

    if( tn + 1 > bufsiz ){
	printk("axmon: Critical error - buffer overflow!!\n");
	buf[bufsiz-1] = '\0';
	return bufsiz - 1;
    }
    return tn;
}

static ssize_t
axmon_pswitch_read(struct file *file, char *buf, size_t len, loff_t *ppos)
{
	size_t rlen;
	int i;

	if (*ppos != 0){
	    if( axmon_tmpbuf_last_used != AXMON_USE_PSWITCH ||
		axmon_tmpbuf_len <= *ppos )
		return 0;
	    rlen = axmon_tmpbuf_len - *ppos;
	}else{
	    rlen = 0;
	    for( i=0; i<AXMON_TRACE_PROC_MAX; i++ )
		rlen += axmon_pswitch_setbuf(i, axmon_tmpbuf+rlen,
					   AXMON_TMPBUF_SIZE-rlen);

	    axmon_tmpbuf_last_used = AXMON_USE_PSWITCH;
	    axmon_tmpbuf_len = rlen;
	}
	if (len < rlen)
	    rlen = len;
	if (copy_to_user(buf, axmon_tmpbuf + *ppos, rlen))
	    return -EFAULT;
	*ppos += rlen;
	return rlen;
}
#endif /* CONFIG_AXE_MONITOR_SCHED_LOG */

#ifdef CONFIG_AXE_SCHED
extern int axsched_nice_weight, axsched_counter_shift, axsched_counter_max;
extern int axsched_add_tail, axsched_task_run_max, axsched_last_sw_time;
extern int axsched_ignore_running_task, axsched_ignore_current_mm;

#define USE_PARAM_READ  1
#define USE_PARAM_WRITE 2

#define TMPBUF_SIZE 256
static char tmpbuf[TMPBUF_SIZE];
static int tmpbuf_len=0;
static int tmpbuf_last_used=0;

static ssize_t
axmon_param_read(struct file *file, char *buf, size_t len, loff_t *ppos)
{
	size_t rlen;

	if (*ppos != 0){
	    if( tmpbuf_last_used != USE_PARAM_READ || tmpbuf_len <= *ppos )
		return 0;
	    rlen = tmpbuf_len - *ppos;
	}else{
	    rlen = 0;
	    rlen += sprintf(tmpbuf+rlen, "%d %d %d %d %d %d %d\n",
			    axsched_nice_weight, axsched_counter_shift,
			    axsched_counter_max, axsched_add_tail,
			    axsched_task_run_max,
			    axsched_ignore_running_task,
			    axsched_ignore_current_mm);
	    rlen += sprintf(tmpbuf+rlen,
			    "; weight counter_shift counter_max add_tail runtick_max ign_runtask ign_curmm\n");
	    tmpbuf_last_used = USE_PARAM_READ;
	    tmpbuf_len = rlen;
	}
	if (len < rlen)
		rlen = len;
	if (copy_to_user(buf, tmpbuf + *ppos, rlen))
		return -EFAULT;
	*ppos += rlen;
	return rlen;
}

static int next_tokn(char *str, int p)
{
    while(str[p] >= '0' && str[p] <= '9')
	p++;
    while(!(str[p] >= '0' && str[p] <= '9'))
	p++;
    return p;
}

static ssize_t
axmon_param_write(struct file *file, const char *buf, size_t len, loff_t *ppos)
{
    int i;

    if (*ppos != 0)
	return -EIO;

    if (len > TMPBUF_SIZE-1)
	return -EIO;

    tmpbuf_len = 0;
    tmpbuf_last_used = USE_PARAM_WRITE;

    if (copy_from_user(tmpbuf, buf, len))
	return -EFAULT;
    tmpbuf[len] = '\0';

    i = 0;
    if (tmpbuf[i] >= '0' && tmpbuf[i] <= '9')	
	axsched_nice_weight = atoi(tmpbuf);
    else
	return -EIO;

    i = next_tokn(tmpbuf, i);

    if (tmpbuf[i] >= '0' && tmpbuf[i] <= '9')	
	axsched_counter_shift = atoi(tmpbuf+i);
    else
	return -EIO;

    i = next_tokn(tmpbuf, i);

    if (tmpbuf[i] >= '0' && tmpbuf[i] <= '9')	
	axsched_counter_max = atoi(tmpbuf+i);
    else
	return -EIO;

    i = next_tokn(tmpbuf, i);

    if (tmpbuf[i] >= '0' && tmpbuf[i] <= '9')	
	axsched_add_tail = atoi(tmpbuf+i);
    else
	return -EIO;

    i = next_tokn(tmpbuf, i);

    if (tmpbuf[i] >= '0' && tmpbuf[i] <= '9'){
	int x;
	x = atoi(tmpbuf+i);
	if( axsched_task_run_max != x ){
	    axsched_task_run_max = x;
	    axsched_last_sw_time = jiffies;
	}
    }else
	return -EIO;

    i = next_tokn(tmpbuf, i);

    if (tmpbuf[i] >= '0' && tmpbuf[i] <= '9')	
	axsched_ignore_running_task = atoi(tmpbuf+i);
    else
	return -EIO;

    i = next_tokn(tmpbuf, i);

    if (tmpbuf[i] >= '0' && tmpbuf[i] <= '9')	
	axsched_ignore_current_mm = atoi(tmpbuf+i);
    else
	return -EIO;

   *ppos += len;
    return len;
}
#endif /* CONFIG_AXE_SCHED */

#ifdef CONFIG_AXE_MONITOR_SCHED_LOG
static struct file_operations misc_fops = 
{
	read:	axmon_misc_read,
	write:	axmon_misc_write,
};

static struct file_operations trace_fops = 
{
	read:	axmon_trace_read,
	write:	axmon_trace_write,
};

static struct file_operations trace_delay_fops = 
{
	read:	axmon_trace_delay_read,
	//	write:	axmon_trace_write,
};

static struct file_operations pswitch_fops = 
{
	read:	axmon_pswitch_read,
};
#endif /* CONFIG_AXE_MONITOR_SCHED_LOG */

#ifdef CONFIG_AXE_SCHED
static struct file_operations param_fops = 
{
	read:	axmon_param_read,
	write:	axmon_param_write,
};
#endif

static struct proc_dir_entry *proc_axmon_root = NULL;

#ifdef CONFIG_AXE_MONITOR_SCHED_LOG
static struct proc_dir_entry *proc_misc;
static struct proc_dir_entry *proc_trace;
static struct proc_dir_entry *proc_trace_delay;
static struct proc_dir_entry *proc_pswitch;
#endif /* CONFIG_AXE_MONITOR_SCHED_LOG */

#ifdef CONFIG_AXE_SCHED
static struct proc_dir_entry *proc_param;
#endif

static int __init axmon_proc_init(void)
{
    printk(KERN_INFO "AXE Kernel Monitor.\n");
    proc_axmon_root = proc_mkdir("axmon", 0);
    if( proc_axmon_root ){
#ifdef CONFIG_AXE_MONITOR_SCHED_LOG
	proc_misc = create_proc_entry("debug0", S_IWUSR | S_IRUGO,
					      proc_axmon_root);
	if( proc_misc )
	    proc_misc->proc_fops = &misc_fops;
	proc_trace = create_proc_entry("trace", S_IWUSR | S_IRUGO,
				       proc_axmon_root);
	if( proc_trace )
	    proc_trace->proc_fops = &trace_fops;
	proc_trace_delay = create_proc_entry("trace_delay", S_IWUSR | S_IRUGO,
					     proc_axmon_root);
	if( proc_trace_delay )
	    proc_trace_delay->proc_fops = &trace_delay_fops;

	proc_pswitch = create_proc_entry("pswitch", S_IWUSR | S_IRUGO,
					       proc_axmon_root);
	if( proc_pswitch )
	    proc_pswitch->proc_fops = &pswitch_fops;
#endif /* CONFIG_AXE_MONITOR_SCHED_LOG */

#ifdef CONFIG_AXE_SCHED
	proc_param = create_proc_entry("param", S_IWUSR | S_IRUGO,
				       proc_axmon_root);
	if( proc_param )
	    proc_param->proc_fops = &param_fops;
#endif
    }

#ifdef CONFIG_AXE_MONITOR_SCHED_LOG
    {
	int i;
	for( i=0; i<AXMON_TRACE_PROC_MAX; i++ ){
	    axmon_trace[i].task = NULL;
	    axmon_trace[i].head = 0;
	    axmon_trace[i].tail= 0;
	    axmon_trace[i].dp.l_head = 0;
	    axmon_trace[i].dp.l_tail= 0;
	}
    }
#endif /* CONFIG_AXE_MONITOR_SCHED_LOG */
    return 0;
}
__initcall(axmon_proc_init);
#endif /* CONFIG_PROC_FS */
